package edu.wpi.appjudicator.ssl;

import android.content.Intent;
import android.security.KeyChain;

import org.bouncycastle.asn1.ASN1ObjectIdentifier;
import org.bouncycastle.asn1.x500.X500Name;
import org.bouncycastle.asn1.x509.BasicConstraints;
import org.bouncycastle.cert.CertIOException;
import org.bouncycastle.cert.jcajce.JcaX509CertificateConverter;
import org.bouncycastle.cert.jcajce.JcaX509v3CertificateBuilder;

import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.operator.ContentSigner;
import org.bouncycastle.operator.OperatorCreationException;
import org.bouncycastle.operator.jcajce.JcaContentSignerBuilder;
import org.bouncycastle.util.io.pem.PemObject;
import org.bouncycastle.util.io.pem.PemReader;


import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.math.BigInteger;


import java.nio.charset.Charset;
import java.security.KeyFactory;
import java.security.KeyPair;
import java.security.KeyPairGenerator;

import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.Provider;
import java.security.Security;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;

import java.security.interfaces.RSAPrivateCrtKey;
import java.security.interfaces.RSAPrivateKey;
import java.security.spec.PKCS8EncodedKeySpec;
import java.util.Base64;
import java.util.Calendar;
import java.util.Date;
import java.util.concurrent.ConcurrentHashMap;

public class SSL_Certificate_Manager {
    final String CERT_FILE = "/system/etc/security/cacerts/d2a13c41.0";
    final String PRIVATE_KEY_FILE = "/storage/emulated/0/Download/cakey.pem";

    private static long serialNum = 0;

    private static SSL_Certificate_Manager singleton_instance = null;
    private static PrivateKey cakey;

    private static ConcurrentHashMap<String, X509Certificate> certificates_map;

    private SSL_Certificate_Manager() {
        certificates_map = new ConcurrentHashMap<>();
    }

    public static SSL_Certificate_Manager getInstance() {
        if (singleton_instance == null) {
            singleton_instance = new SSL_Certificate_Manager();

            String private_key = new String(cakey_raw, Charset.defaultCharset());

            String privateKeyPEM = private_key
                    .replace("-----BEGIN PRIVATE KEY-----", "")
                    .replaceAll(System.lineSeparator(), "")
                    .replace("-----END PRIVATE KEY-----", "");

            byte[] encoded = Base64.getDecoder().decode(privateKeyPEM);

            try {
                KeyFactory keyFactory = KeyFactory.getInstance("RSA");
                PKCS8EncodedKeySpec keySpec = new PKCS8EncodedKeySpec(encoded);
                cakey = (RSAPrivateKey) keyFactory.generatePrivate(keySpec);
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

        return singleton_instance;
    }

    public X509Certificate createAndSignCertificate(String server_name, KeyPair keyPairServer) {
        X509Certificate cert = certificates_map.get(server_name);
        if (cert != null) {
            return cert;    //already produced a cert like this one, no need for another one
        }
        try {
            Provider bcProvider = new BouncyCastleProvider();
            Security.addProvider(bcProvider);

            long now = System.currentTimeMillis();
            Date startDate = new Date(now);

            X500Name dnName = new X500Name("CN="+server_name);
            X500Name ca_dnName = new X500Name("CN=mycadomain.com");

            BigInteger certSerialNumber = new BigInteger(Long.toString(serialNum)); // <-- Using the current timestamp as the certificate serial number

            Calendar calendar = Calendar.getInstance();
            calendar.setTime(startDate);
            calendar.add(Calendar.YEAR, 1); // <-- 1 Yr validity

            Date endDate = calendar.getTime();

            String signatureAlgorithm = "SHA256WithRSA"; // <-- Use appropriate signature algorithm based on your keyPair algorithm.

            ContentSigner contentSigner = new JcaContentSignerBuilder(signatureAlgorithm).build(cakey);

            JcaX509v3CertificateBuilder certBuilder = new JcaX509v3CertificateBuilder(ca_dnName, certSerialNumber, startDate, endDate, dnName, keyPairServer.getPublic());

            // Extensions --------------------------

            // Basic Constraints
            BasicConstraints basicConstraints = new BasicConstraints(false); // <-- true for CA, false for EndEntity

            certBuilder.addExtension(new ASN1ObjectIdentifier("2.5.29.19"), true, basicConstraints); // Basic Constraints is usually marked as critical.

            // -------------------------------------
            cert = new JcaX509CertificateConverter().setProvider(bcProvider).getCertificate(certBuilder.build(contentSigner));
            serialNum += 1;
            //certificates_map.put(server_name, cert); TODO: This wasn't implemented properly
            return cert;
        } catch (Exception e) {
            return null;
        }
    }

    //TODO: When one figures out how to generate and store the root CA effectively, change this
//    public RSAPrivateCrtKey getCAPrivateKey() throws Exception {
//        KeyFactory factory = KeyFactory.getInstance("RSA");
//
//        try (FileReader keyReader = new FileReader(PRIVATE_KEY_FILE);
//             PemReader pemReader = new PemReader(keyReader)) {
//
//            PemObject pemObject = pemReader.readPemObject();
//            byte[] content = pemObject.getContent();
//            PKCS8EncodedKeySpec privKeySpec = new PKCS8EncodedKeySpec(content);
//            return (RSAPrivateCrtKey) factory.generatePrivate(privKeySpec);
//        }
//    }
//
//    public X509Certificate getSelfSignedCA() throws CertificateException {
//        try {
//            FileInputStream certIs =  new FileInputStream(CERT_FILE);
//            X509Certificate x509 = (X509Certificate) CertificateFactory.getInstance("X.509").generateCertificate(certIs);
//            return x509;
//        } catch (FileNotFoundException e) {
//            return null;
//        } catch (CertificateException e) {
//            e.printStackTrace();
//            throw e;
//        }
//    }
//
//    public X509Certificate storeNewRootCA(Intent intent, X509Certificate x509) throws CertificateException{
//        try {
//                intent.putExtra(KeyChain.EXTRA_CERTIFICATE, x509.getEncoded());
//                intent.putExtra(KeyChain.EXTRA_NAME, "AppJudicator Cert");
//        } catch (CertificateEncodingException e) {
//            e.printStackTrace();
//        }
//        return x509;
//    }
//
//    private X509Certificate generateRootCA(){
//        KeyPairGenerator keyGen = null;
//        try {
//            keyGen = KeyPairGenerator.getInstance("RSA");
//        } catch (NoSuchAlgorithmException e) {
//            e.printStackTrace();
//        }
//        keyGen.initialize(1024);
//        KeyPair keyPair = keyGen.generateKeyPair();
//
//        Provider bcProvider = new BouncyCastleProvider();
//        Security.addProvider(bcProvider);
//
//        long now = System.currentTimeMillis();
//        Date startDate = new Date(now);
//
//        X500Name dnName = new X500Name("CN=" + "appJudicator");
//        BigInteger certSerialNumber = new BigInteger(Long.toString(now)); // <-- Using the current timestamp as the certificate serial number
//
//        Calendar calendar = Calendar.getInstance();
//        calendar.setTime(startDate);
//        calendar.add(Calendar.YEAR, 1); // <-- 1 Yr validity
//
//        Date endDate = calendar.getTime();
//
//        String signatureAlgorithm = "SHA256WithRSA"; // <-- Use appropriate signature algorithm based on your keyPair algorithm.
//
//        ContentSigner contentSigner = null;
//        try {
//            contentSigner = new JcaContentSignerBuilder(signatureAlgorithm).build(keyPair.getPrivate());
//        } catch (OperatorCreationException e) {
//            e.printStackTrace();
//        }
//
//        JcaX509v3CertificateBuilder certBuilder = new JcaX509v3CertificateBuilder(dnName, certSerialNumber, startDate, endDate, dnName, keyPair.getPublic());
//
//        // Extensions --------------------------
//
//        // Basic Constraints
//        BasicConstraints basicConstraints = new BasicConstraints(true); // <-- true for CA, false for EndEntity
//
//        try {
//            certBuilder.addExtension(new ASN1ObjectIdentifier("2.5.29.19"), true, basicConstraints); // Basic Constraints is usually marked as critical.
//        } catch (CertIOException e) {
//            e.printStackTrace();
//        }
//
//        // -------------------------------------
//
//        try {
//            return new JcaX509CertificateConverter().setProvider(bcProvider).getCertificate(certBuilder.build(contentSigner));
//        } catch (CertificateException e) {
//            e.printStackTrace();
//        }
//
//        return null;
//    }

//    private static class UnifiedTrustManager implements X509TrustManager {
//        private X509TrustManager defaultTrustManager;
//        private X509TrustManager localTrustManager;
//        public UnifiedTrustManager(KeyStore localKeyStore) throws KeyStoreException {
//            try {
//                this.defaultTrustManager = createTrustManager(null);
//                this.localTrustManager = createTrustManager(localKeyStore);
//            } catch (NoSuchAlgorithmException e) {
//                e.printStackTrace();
//            }
//        }
//        private X509TrustManager createTrustManager(KeyStore store) throws NoSuchAlgorithmException, KeyStoreException {
//            String tmfAlgorithm = TrustManagerFactory.getDefaultAlgorithm();
//            TrustManagerFactory tmf = TrustManagerFactory.getInstance(tmfAlgorithm);
//            tmf.init((KeyStore) store);
//            TrustManager[] trustManagers = tmf.getTrustManagers();
//            return (X509TrustManager) trustManagers[0];
//        }
//        public void checkServerTrusted(X509Certificate[] chain, String authType) throws CertificateException {
//            try {
//                defaultTrustManager.checkServerTrusted(chain, authType);
//            } catch (CertificateException ce) {
//                localTrustManager.checkServerTrusted(chain, authType);
//            }
//        }
//        @Override
//        public void checkClientTrusted(X509Certificate[] chain, String authType) throws CertificateException {
//            try {
//                defaultTrustManager.checkClientTrusted(chain, authType);
//            } catch (CertificateException ce) {
//                localTrustManager.checkClientTrusted(chain, authType);
//            }
//        }
//        @Override
//        public X509Certificate[] getAcceptedIssuers() {
//            X509Certificate[] first = defaultTrustManager.getAcceptedIssuers();
//            X509Certificate[] second = localTrustManager.getAcceptedIssuers();
//            X509Certificate[] result = Arrays.copyOf(first, first.length + second.length);
//            System.arraycopy(second, 0, result, first.length, second.length);
//            return result;
//        }
//    }

//    public static void setCustomCertificateAuthority(InputStream inputStream) {
//
//        try {
//            // Load CAs from an InputStream
//            // (could be from a resource or ByteArrayInputStream or ...)
//            CertificateFactory cf = CertificateFactory.getInstance("X.509");
//            InputStream caInput = new BufferedInputStream(inputStream);
//            Certificate ca;
//            try {
//                ca = cf.generateCertificate(caInput);
//                System.out.println("ca=" + ((X509Certificate) ca).getSubjectDN());
//            } finally {
//                caInput.close();
//            }
//
//            // Create a KeyStore containing our trusted CAs
//            String keyStoreType = KeyStore.getDefaultType();
//            KeyStore keyStore = KeyStore.getInstance(keyStoreType);
//            keyStore.load(null, null);
//            keyStore.setCertificateEntry("ca", ca);
//
//            // Create a TrustManager that trusts the CAs in our KeyStore and system CA
//            UnifiedTrustManager trustManager = new UnifiedTrustManager(keyStore);
//
//            // Create an SSLContext that uses our TrustManager
//            SSLContext context = SSLContext.getInstance("TLS");
//            context.init(null, new TrustManager[]{trustManager}, null);
//
//            // Tell the URLConnection to use a SocketFactory from our SSLContext
//            HttpsURLConnection.setDefaultSSLSocketFactory(context.getSocketFactory());
//
//        } catch (CertificateException e) {
//            e.printStackTrace();
//        } catch (NoSuchAlgorithmException e) {
//            e.printStackTrace();
//        } catch (KeyStoreException e) {
//            e.printStackTrace();
//        } catch (KeyManagementException e) {
//            e.printStackTrace();
//        } catch (MalformedURLException e) {
//            e.printStackTrace();
//        } catch (IOException e) {
//            e.printStackTrace();
//        }
//    }

    public static final byte[] ca_cert_raw_bytes = new byte[] {
            (byte)0x30, (byte)0x82, (byte)0x03, (byte)0x1f, (byte)0x30, (byte)0x82, (byte)0x02, (byte)0x07, (byte)0xa0, (byte)0x03, (byte)0x02, (byte)0x01, (byte)0x02, (byte)0x02, (byte)0x13, (byte)0x18,
            (byte)0x04, (byte)0x15, (byte)0xd6, (byte)0xef, (byte)0xa3, (byte)0x18, (byte)0x07, (byte)0xf8, (byte)0x66, (byte)0x76, (byte)0x34, (byte)0xb4, (byte)0x9b, (byte)0x42, (byte)0x6a, (byte)0x94,
            (byte)0xea, (byte)0xac, (byte)0x30, (byte)0x0d, (byte)0x06, (byte)0x09, (byte)0x2a, (byte)0x86, (byte)0x48, (byte)0x86, (byte)0xf7, (byte)0x0d, (byte)0x01, (byte)0x01, (byte)0x0b, (byte)0x05,
            (byte)0x00, (byte)0x30, (byte)0x19, (byte)0x31, (byte)0x17, (byte)0x30, (byte)0x15, (byte)0x06, (byte)0x03, (byte)0x55, (byte)0x04, (byte)0x03, (byte)0x0c, (byte)0x0e, (byte)0x6d, (byte)0x79,
            (byte)0x63, (byte)0x61, (byte)0x64, (byte)0x6f, (byte)0x6d, (byte)0x61, (byte)0x69, (byte)0x6e, (byte)0x2e, (byte)0x63, (byte)0x6f, (byte)0x6d, (byte)0x30, (byte)0x1e, (byte)0x17, (byte)0x0d,
            (byte)0x32, (byte)0x33, (byte)0x30, (byte)0x32, (byte)0x31, (byte)0x33, (byte)0x31, (byte)0x37, (byte)0x30, (byte)0x36, (byte)0x31, (byte)0x39, (byte)0x5a, (byte)0x17, (byte)0x0d, (byte)0x32,
            (byte)0x34, (byte)0x30, (byte)0x32, (byte)0x31, (byte)0x33, (byte)0x31, (byte)0x37, (byte)0x30, (byte)0x36, (byte)0x31, (byte)0x39, (byte)0x5a, (byte)0x30, (byte)0x19, (byte)0x31, (byte)0x17,
            (byte)0x30, (byte)0x15, (byte)0x06, (byte)0x03, (byte)0x55, (byte)0x04, (byte)0x03, (byte)0x0c, (byte)0x0e, (byte)0x6d, (byte)0x79, (byte)0x63, (byte)0x61, (byte)0x64, (byte)0x6f, (byte)0x6d,
            (byte)0x61, (byte)0x69, (byte)0x6e, (byte)0x2e, (byte)0x63, (byte)0x6f, (byte)0x6d, (byte)0x30, (byte)0x82, (byte)0x01, (byte)0x22, (byte)0x30, (byte)0x0d, (byte)0x06, (byte)0x09, (byte)0x2a,
            (byte)0x86, (byte)0x48, (byte)0x86, (byte)0xf7, (byte)0x0d, (byte)0x01, (byte)0x01, (byte)0x01, (byte)0x05, (byte)0x00, (byte)0x03, (byte)0x82, (byte)0x01, (byte)0x0f, (byte)0x00, (byte)0x30,
            (byte)0x82, (byte)0x01, (byte)0x0a, (byte)0x02, (byte)0x82, (byte)0x01, (byte)0x01, (byte)0x00, (byte)0xd1, (byte)0x45, (byte)0x45, (byte)0xdb, (byte)0x8c, (byte)0x19, (byte)0x07, (byte)0x05,
            (byte)0xf5, (byte)0x71, (byte)0xe3, (byte)0x8f, (byte)0x1c, (byte)0xfa, (byte)0x27, (byte)0x70, (byte)0x77, (byte)0xab, (byte)0x05, (byte)0x64, (byte)0x67, (byte)0x4a, (byte)0x5f, (byte)0xaa,
            (byte)0x20, (byte)0xd0, (byte)0xae, (byte)0xc4, (byte)0x41, (byte)0xb1, (byte)0xa5, (byte)0xd6, (byte)0xf3, (byte)0x90, (byte)0x35, (byte)0x1b, (byte)0xd7, (byte)0x1d, (byte)0x6f, (byte)0x5d,
            (byte)0x10, (byte)0xe2, (byte)0xbd, (byte)0x9f, (byte)0x83, (byte)0x06, (byte)0x96, (byte)0x71, (byte)0x67, (byte)0xff, (byte)0xbf, (byte)0x0f, (byte)0x9a, (byte)0x2f, (byte)0x5c, (byte)0x95,
            (byte)0x19, (byte)0x28, (byte)0x95, (byte)0xc0, (byte)0xe0, (byte)0x9b, (byte)0x59, (byte)0x72, (byte)0xb1, (byte)0x1a, (byte)0x00, (byte)0xd9, (byte)0x96, (byte)0x85, (byte)0x53, (byte)0x82,
            (byte)0x6b, (byte)0x4b, (byte)0xb7, (byte)0x23, (byte)0x8d, (byte)0x8f, (byte)0xd5, (byte)0x08, (byte)0xc2, (byte)0xff, (byte)0x17, (byte)0x25, (byte)0xc9, (byte)0x6e, (byte)0x0a, (byte)0x2a,
            (byte)0xca, (byte)0x18, (byte)0x87, (byte)0x8b, (byte)0x78, (byte)0xf9, (byte)0x8c, (byte)0x89, (byte)0x05, (byte)0x97, (byte)0x5c, (byte)0x5d, (byte)0xdf, (byte)0xd4, (byte)0xe6, (byte)0x88,
            (byte)0x3b, (byte)0xae, (byte)0xac, (byte)0xd6, (byte)0x7b, (byte)0x69, (byte)0x7a, (byte)0xc9, (byte)0x85, (byte)0x59, (byte)0x94, (byte)0xf1, (byte)0x08, (byte)0xcb, (byte)0xb0, (byte)0xb3,
            (byte)0xa7, (byte)0xfe, (byte)0x8b, (byte)0x96, (byte)0x35, (byte)0x0e, (byte)0xdb, (byte)0xa3, (byte)0x0a, (byte)0x71, (byte)0x70, (byte)0x15, (byte)0x63, (byte)0xd9, (byte)0xa5, (byte)0xf8,
            (byte)0xd5, (byte)0x4e, (byte)0x47, (byte)0xd9, (byte)0xb6, (byte)0x8b, (byte)0xcf, (byte)0x56, (byte)0x82, (byte)0x8a, (byte)0xee, (byte)0x50, (byte)0x45, (byte)0x30, (byte)0x09, (byte)0x26,
            (byte)0x37, (byte)0xda, (byte)0x73, (byte)0x37, (byte)0xf0, (byte)0xd4, (byte)0x3f, (byte)0x26, (byte)0x83, (byte)0x45, (byte)0xf5, (byte)0x7d, (byte)0xbe, (byte)0xe5, (byte)0x2f, (byte)0xbe,
            (byte)0x6b, (byte)0x60, (byte)0x8a, (byte)0x8e, (byte)0x0b, (byte)0x0e, (byte)0xe2, (byte)0x82, (byte)0x37, (byte)0x2e, (byte)0x3c, (byte)0x22, (byte)0x67, (byte)0x49, (byte)0x63, (byte)0x03,
            (byte)0x0b, (byte)0xda, (byte)0xf5, (byte)0x2d, (byte)0x78, (byte)0xbc, (byte)0x4c, (byte)0x8c, (byte)0x3c, (byte)0x3a, (byte)0xdd, (byte)0xf0, (byte)0x42, (byte)0x6b, (byte)0x78, (byte)0x87,
            (byte)0xdc, (byte)0xda, (byte)0xd8, (byte)0xb4, (byte)0x01, (byte)0xec, (byte)0x5d, (byte)0x5a, (byte)0x88, (byte)0x2f, (byte)0x82, (byte)0x02, (byte)0xfb, (byte)0x0f, (byte)0x21, (byte)0xe5,
            (byte)0xfd, (byte)0x36, (byte)0x10, (byte)0xf1, (byte)0x16, (byte)0xe9, (byte)0xee, (byte)0x6f, (byte)0x35, (byte)0xca, (byte)0x4e, (byte)0xc8, (byte)0xdf, (byte)0xa7, (byte)0x3b, (byte)0x24,
            (byte)0x45, (byte)0xe1, (byte)0xdf, (byte)0x0c, (byte)0xab, (byte)0x8b, (byte)0x26, (byte)0x94, (byte)0xc8, (byte)0x0a, (byte)0x6f, (byte)0xff, (byte)0x03, (byte)0xb7, (byte)0x77, (byte)0x7c,
            (byte)0x44, (byte)0x47, (byte)0xaf, (byte)0x4d, (byte)0x5c, (byte)0x95, (byte)0xfb, (byte)0x77, (byte)0x02, (byte)0x03, (byte)0x01, (byte)0x00, (byte)0x01, (byte)0xa3, (byte)0x60, (byte)0x30,
            (byte)0x5e, (byte)0x30, (byte)0x1d, (byte)0x06, (byte)0x03, (byte)0x55, (byte)0x1d, (byte)0x0e, (byte)0x04, (byte)0x16, (byte)0x04, (byte)0x14, (byte)0x5a, (byte)0x4f, (byte)0xd7, (byte)0x6a,
            (byte)0x3c, (byte)0x84, (byte)0xda, (byte)0x79, (byte)0x53, (byte)0x4d, (byte)0x27, (byte)0xa5, (byte)0x56, (byte)0x83, (byte)0x63, (byte)0x4f, (byte)0x4e, (byte)0x1a, (byte)0xad, (byte)0x6f,
            (byte)0x30, (byte)0x1f, (byte)0x06, (byte)0x03, (byte)0x55, (byte)0x1d, (byte)0x23, (byte)0x04, (byte)0x18, (byte)0x30, (byte)0x16, (byte)0x80, (byte)0x14, (byte)0x5a, (byte)0x4f, (byte)0xd7,
            (byte)0x6a, (byte)0x3c, (byte)0x84, (byte)0xda, (byte)0x79, (byte)0x53, (byte)0x4d, (byte)0x27, (byte)0xa5, (byte)0x56, (byte)0x83, (byte)0x63, (byte)0x4f, (byte)0x4e, (byte)0x1a, (byte)0xad,
            (byte)0x6f, (byte)0x30, (byte)0x0f, (byte)0x06, (byte)0x03, (byte)0x55, (byte)0x1d, (byte)0x13, (byte)0x01, (byte)0x01, (byte)0xff, (byte)0x04, (byte)0x05, (byte)0x30, (byte)0x03, (byte)0x01,
            (byte)0x01, (byte)0xff, (byte)0x30, (byte)0x0b, (byte)0x06, (byte)0x03, (byte)0x55, (byte)0x1d, (byte)0x0f, (byte)0x04, (byte)0x04, (byte)0x03, (byte)0x02, (byte)0x01, (byte)0x06, (byte)0x30,
            (byte)0x0d, (byte)0x06, (byte)0x09, (byte)0x2a, (byte)0x86, (byte)0x48, (byte)0x86, (byte)0xf7, (byte)0x0d, (byte)0x01, (byte)0x01, (byte)0x0b, (byte)0x05, (byte)0x00, (byte)0x03, (byte)0x82,
            (byte)0x01, (byte)0x01, (byte)0x00, (byte)0x1d, (byte)0x32, (byte)0xb0, (byte)0xbc, (byte)0x95, (byte)0xda, (byte)0x20, (byte)0xa1, (byte)0x8f, (byte)0x24, (byte)0x0c, (byte)0xbb, (byte)0x15,
            (byte)0x38, (byte)0x2a, (byte)0xca, (byte)0xe3, (byte)0xe3, (byte)0x42, (byte)0xe2, (byte)0xde, (byte)0x85, (byte)0x89, (byte)0xa5, (byte)0x8f, (byte)0x8e, (byte)0x00, (byte)0x0a, (byte)0x2e,
            (byte)0x80, (byte)0xfe, (byte)0xac, (byte)0xf4, (byte)0x27, (byte)0x74, (byte)0x11, (byte)0x09, (byte)0x43, (byte)0x4f, (byte)0x62, (byte)0xdf, (byte)0xf9, (byte)0x71, (byte)0xea, (byte)0x48,
            (byte)0xd6, (byte)0x13, (byte)0x9e, (byte)0x8b, (byte)0xc2, (byte)0xde, (byte)0x56, (byte)0x6a, (byte)0x6a, (byte)0x54, (byte)0xb5, (byte)0x6a, (byte)0x21, (byte)0xfc, (byte)0x60, (byte)0xfd,
            (byte)0xbc, (byte)0x23, (byte)0x0c, (byte)0x99, (byte)0xbc, (byte)0x13, (byte)0xd3, (byte)0xd1, (byte)0x92, (byte)0xe7, (byte)0x3a, (byte)0x17, (byte)0x2f, (byte)0x0e, (byte)0x41, (byte)0x15,
            (byte)0xcd, (byte)0x50, (byte)0x69, (byte)0xae, (byte)0x13, (byte)0xa0, (byte)0xee, (byte)0xca, (byte)0x73, (byte)0xfa, (byte)0x80, (byte)0xd0, (byte)0xc6, (byte)0x8f, (byte)0x14, (byte)0x04,
            (byte)0x73, (byte)0x1d, (byte)0x3a, (byte)0x51, (byte)0xcd, (byte)0x20, (byte)0x67, (byte)0xc7, (byte)0x54, (byte)0xb1, (byte)0x48, (byte)0x80, (byte)0xa4, (byte)0x65, (byte)0xf1, (byte)0x4c,
            (byte)0x86, (byte)0x4e, (byte)0x92, (byte)0x6e, (byte)0xfd, (byte)0xb4, (byte)0xa2, (byte)0xab, (byte)0x7d, (byte)0xc4, (byte)0xab, (byte)0x27, (byte)0xa6, (byte)0x2d, (byte)0xb7, (byte)0xae,
            (byte)0x31, (byte)0x4f, (byte)0x4f, (byte)0x58, (byte)0x00, (byte)0x5f, (byte)0x88, (byte)0xff, (byte)0xd8, (byte)0x74, (byte)0xd9, (byte)0xd3, (byte)0x1c, (byte)0x88, (byte)0x30, (byte)0x27,
            (byte)0x4e, (byte)0xdb, (byte)0x3d, (byte)0x00, (byte)0xa3, (byte)0x74, (byte)0x81, (byte)0xf8, (byte)0x2c, (byte)0x7c, (byte)0x09, (byte)0x2d, (byte)0x95, (byte)0xfe, (byte)0x77, (byte)0x3f,
            (byte)0x62, (byte)0x6c, (byte)0x00, (byte)0xc8, (byte)0xec, (byte)0xea, (byte)0x89, (byte)0x87, (byte)0x89, (byte)0x22, (byte)0xc1, (byte)0x94, (byte)0xe4, (byte)0x7c, (byte)0xa5, (byte)0x0e,
            (byte)0x87, (byte)0x93, (byte)0xd0, (byte)0x1a, (byte)0xbb, (byte)0xfb, (byte)0x7e, (byte)0xef, (byte)0xc8, (byte)0x33, (byte)0x4b, (byte)0x15, (byte)0x2e, (byte)0x6a, (byte)0xdf, (byte)0x39,
            (byte)0x49, (byte)0xbf, (byte)0xed, (byte)0xc3, (byte)0x09, (byte)0x7a, (byte)0x96, (byte)0x83, (byte)0x53, (byte)0x0d, (byte)0xdf, (byte)0x5b, (byte)0xa6, (byte)0x33, (byte)0x3c, (byte)0x8c,
            (byte)0xf5, (byte)0x67, (byte)0x40, (byte)0x1f, (byte)0x30, (byte)0xad, (byte)0x2c, (byte)0xab, (byte)0xd5, (byte)0x14, (byte)0x48, (byte)0xbf, (byte)0x4c, (byte)0x21, (byte)0xec, (byte)0x44,
            (byte)0xd6, (byte)0xaf, (byte)0xaa, (byte)0x04, (byte)0xed, (byte)0x40, (byte)0xc4, (byte)0x33, (byte)0xa9, (byte)0xcc, (byte)0x48, (byte)0xe0, (byte)0x63, (byte)0x13, (byte)0x7f, (byte)0x61,
            (byte)0x1f, (byte)0xd8, (byte)0xd7, (byte)0xc5, (byte)0xd6, (byte)0xae, (byte)0xd8, (byte)0x18, (byte)0xf6, (byte)0x44, (byte)0xe1, (byte)0x96, (byte)0x38, (byte)0x0d, (byte)0x45, (byte)0x2a,
            (byte)0xc9, (byte)0xeb, (byte)0xad
    };

    private static final byte[] cakey_raw = {
            (byte)0x2d, (byte)0x2d, (byte)0x2d, (byte)0x2d, (byte)0x2d, (byte)0x42, (byte)0x45, (byte)0x47, (byte)0x49, (byte)0x4e, (byte)0x20, (byte)0x50, (byte)0x52, (byte)0x49, (byte)0x56,
            (byte)0x41, (byte)0x54, (byte)0x45, (byte)0x20, (byte)0x4b, (byte)0x45, (byte)0x59, (byte)0x2d, (byte)0x2d, (byte)0x2d, (byte)0x2d, (byte)0x2d, (byte)0x0a, (byte)0x4d, (byte)0x49, (byte)0x49,
            (byte)0x45, (byte)0x76, (byte)0x77, (byte)0x49, (byte)0x42, (byte)0x41, (byte)0x44, (byte)0x41, (byte)0x4e, (byte)0x42, (byte)0x67, (byte)0x6b, (byte)0x71, (byte)0x68, (byte)0x6b, (byte)0x69,
            (byte)0x47, (byte)0x39, (byte)0x77, (byte)0x30, (byte)0x42, (byte)0x41, (byte)0x51, (byte)0x45, (byte)0x46, (byte)0x41, (byte)0x41, (byte)0x53, (byte)0x43, (byte)0x42, (byte)0x4b, (byte)0x6b,
            (byte)0x77, (byte)0x67, (byte)0x67, (byte)0x53, (byte)0x6c, (byte)0x41, (byte)0x67, (byte)0x45, (byte)0x41, (byte)0x41, (byte)0x6f, (byte)0x49, (byte)0x42, (byte)0x41, (byte)0x51, (byte)0x44,
            (byte)0x52, (byte)0x52, (byte)0x55, (byte)0x58, (byte)0x62, (byte)0x6a, (byte)0x42, (byte)0x6b, (byte)0x48, (byte)0x42, (byte)0x66, (byte)0x56, (byte)0x78, (byte)0x0a, (byte)0x34, (byte)0x34,
            (byte)0x38, (byte)0x63, (byte)0x2b, (byte)0x69, (byte)0x64, (byte)0x77, (byte)0x64, (byte)0x36, (byte)0x73, (byte)0x46, (byte)0x5a, (byte)0x47, (byte)0x64, (byte)0x4b, (byte)0x58, (byte)0x36,
            (byte)0x6f, (byte)0x67, (byte)0x30, (byte)0x4b, (byte)0x37, (byte)0x45, (byte)0x51, (byte)0x62, (byte)0x47, (byte)0x6c, (byte)0x31, (byte)0x76, (byte)0x4f, (byte)0x51, (byte)0x4e, (byte)0x52,
            (byte)0x76, (byte)0x58, (byte)0x48, (byte)0x57, (byte)0x39, (byte)0x64, (byte)0x45, (byte)0x4f, (byte)0x4b, (byte)0x39, (byte)0x6e, (byte)0x34, (byte)0x4d, (byte)0x47, (byte)0x6c, (byte)0x6e,
            (byte)0x46, (byte)0x6e, (byte)0x2f, (byte)0x37, (byte)0x38, (byte)0x50, (byte)0x6d, (byte)0x69, (byte)0x39, (byte)0x63, (byte)0x6c, (byte)0x52, (byte)0x6b, (byte)0x6f, (byte)0x0a, (byte)0x6c,
            (byte)0x63, (byte)0x44, (byte)0x67, (byte)0x6d, (byte)0x31, (byte)0x6c, (byte)0x79, (byte)0x73, (byte)0x52, (byte)0x6f, (byte)0x41, (byte)0x32, (byte)0x5a, (byte)0x61, (byte)0x46, (byte)0x55,
            (byte)0x34, (byte)0x4a, (byte)0x72, (byte)0x53, (byte)0x37, (byte)0x63, (byte)0x6a, (byte)0x6a, (byte)0x59, (byte)0x2f, (byte)0x56, (byte)0x43, (byte)0x4d, (byte)0x4c, (byte)0x2f, (byte)0x46,
            (byte)0x79, (byte)0x58, (byte)0x4a, (byte)0x62, (byte)0x67, (byte)0x6f, (byte)0x71, (byte)0x79, (byte)0x68, (byte)0x69, (byte)0x48, (byte)0x69, (byte)0x33, (byte)0x6a, (byte)0x35, (byte)0x6a,
            (byte)0x49, (byte)0x6b, (byte)0x46, (byte)0x6c, (byte)0x31, (byte)0x78, (byte)0x64, (byte)0x33, (byte)0x39, (byte)0x54, (byte)0x6d, (byte)0x69, (byte)0x44, (byte)0x75, (byte)0x75, (byte)0x0a,
            (byte)0x72, (byte)0x4e, (byte)0x5a, (byte)0x37, (byte)0x61, (byte)0x58, (byte)0x72, (byte)0x4a, (byte)0x68, (byte)0x56, (byte)0x6d, (byte)0x55, (byte)0x38, (byte)0x51, (byte)0x6a, (byte)0x4c,
            (byte)0x73, (byte)0x4c, (byte)0x4f, (byte)0x6e, (byte)0x2f, (byte)0x6f, (byte)0x75, (byte)0x57, (byte)0x4e, (byte)0x51, (byte)0x37, (byte)0x62, (byte)0x6f, (byte)0x77, (byte)0x70, (byte)0x78,
            (byte)0x63, (byte)0x42, (byte)0x56, (byte)0x6a, (byte)0x32, (byte)0x61, (byte)0x58, (byte)0x34, (byte)0x31, (byte)0x55, (byte)0x35, (byte)0x48, (byte)0x32, (byte)0x62, (byte)0x61, (byte)0x4c,
            (byte)0x7a, (byte)0x31, (byte)0x61, (byte)0x43, (byte)0x69, (byte)0x75, (byte)0x35, (byte)0x51, (byte)0x52, (byte)0x54, (byte)0x41, (byte)0x4a, (byte)0x4a, (byte)0x6a, (byte)0x66, (byte)0x61,
            (byte)0x0a, (byte)0x63, (byte)0x7a, (byte)0x66, (byte)0x77, (byte)0x31, (byte)0x44, (byte)0x38, (byte)0x6d, (byte)0x67, (byte)0x30, (byte)0x58, (byte)0x31, (byte)0x66, (byte)0x62, (byte)0x37,
            (byte)0x6c, (byte)0x4c, (byte)0x37, (byte)0x35, (byte)0x72, (byte)0x59, (byte)0x49, (byte)0x71, (byte)0x4f, (byte)0x43, (byte)0x77, (byte)0x37, (byte)0x69, (byte)0x67, (byte)0x6a, (byte)0x63,
            (byte)0x75, (byte)0x50, (byte)0x43, (byte)0x4a, (byte)0x6e, (byte)0x53, (byte)0x57, (byte)0x4d, (byte)0x44, (byte)0x43, (byte)0x39, (byte)0x72, (byte)0x31, (byte)0x4c, (byte)0x58, (byte)0x69,
            (byte)0x38, (byte)0x54, (byte)0x49, (byte)0x77, (byte)0x38, (byte)0x4f, (byte)0x74, (byte)0x33, (byte)0x77, (byte)0x51, (byte)0x6d, (byte)0x74, (byte)0x34, (byte)0x68, (byte)0x39, (byte)0x7a,
            (byte)0x61, (byte)0x0a, (byte)0x32, (byte)0x4c, (byte)0x51, (byte)0x42, (byte)0x37, (byte)0x46, (byte)0x31, (byte)0x61, (byte)0x69, (byte)0x43, (byte)0x2b, (byte)0x43, (byte)0x41, (byte)0x76,
            (byte)0x73, (byte)0x50, (byte)0x49, (byte)0x65, (byte)0x58, (byte)0x39, (byte)0x4e, (byte)0x68, (byte)0x44, (byte)0x78, (byte)0x46, (byte)0x75, (byte)0x6e, (byte)0x75, (byte)0x62, (byte)0x7a,
            (byte)0x58, (byte)0x4b, (byte)0x54, (byte)0x73, (byte)0x6a, (byte)0x66, (byte)0x70, (byte)0x7a, (byte)0x73, (byte)0x6b, (byte)0x52, (byte)0x65, (byte)0x48, (byte)0x66, (byte)0x44, (byte)0x4b,
            (byte)0x75, (byte)0x4c, (byte)0x4a, (byte)0x70, (byte)0x54, (byte)0x49, (byte)0x43, (byte)0x6d, (byte)0x2f, (byte)0x2f, (byte)0x41, (byte)0x37, (byte)0x64, (byte)0x33, (byte)0x66, (byte)0x45,
            (byte)0x52, (byte)0x48, (byte)0x0a, (byte)0x72, (byte)0x30, (byte)0x31, (byte)0x63, (byte)0x6c, (byte)0x66, (byte)0x74, (byte)0x33, (byte)0x41, (byte)0x67, (byte)0x4d, (byte)0x42, (byte)0x41,
            (byte)0x41, (byte)0x45, (byte)0x43, (byte)0x67, (byte)0x67, (byte)0x45, (byte)0x41, (byte)0x45, (byte)0x66, (byte)0x66, (byte)0x4c, (byte)0x72, (byte)0x72, (byte)0x55, (byte)0x52, (byte)0x74,
            (byte)0x51, (byte)0x75, (byte)0x71, (byte)0x37, (byte)0x6c, (byte)0x4e, (byte)0x6d, (byte)0x68, (byte)0x30, (byte)0x74, (byte)0x6e, (byte)0x4e, (byte)0x58, (byte)0x33, (byte)0x67, (byte)0x33,
            (byte)0x6a, (byte)0x36, (byte)0x6e, (byte)0x71, (byte)0x37, (byte)0x67, (byte)0x5a, (byte)0x43, (byte)0x47, (byte)0x5a, (byte)0x43, (byte)0x69, (byte)0x7a, (byte)0x46, (byte)0x45, (byte)0x79,
            (byte)0x6f, (byte)0x4e, (byte)0x78, (byte)0x0a, (byte)0x4f, (byte)0x44, (byte)0x47, (byte)0x48, (byte)0x4b, (byte)0x57, (byte)0x6f, (byte)0x71, (byte)0x6c, (byte)0x4b, (byte)0x67, (byte)0x4e,
            (byte)0x66, (byte)0x39, (byte)0x43, (byte)0x42, (byte)0x56, (byte)0x49, (byte)0x42, (byte)0x34, (byte)0x54, (byte)0x6b, (byte)0x43, (byte)0x6b, (byte)0x4b, (byte)0x2f, (byte)0x66, (byte)0x47,
            (byte)0x46, (byte)0x30, (byte)0x48, (byte)0x70, (byte)0x2f, (byte)0x68, (byte)0x55, (byte)0x56, (byte)0x6b, (byte)0x58, (byte)0x4e, (byte)0x6c, (byte)0x4d, (byte)0x37, (byte)0x62, (byte)0x48,
            (byte)0x44, (byte)0x38, (byte)0x6d, (byte)0x65, (byte)0x35, (byte)0x65, (byte)0x57, (byte)0x73, (byte)0x72, (byte)0x66, (byte)0x2b, (byte)0x37, (byte)0x51, (byte)0x72, (byte)0x38, (byte)0x31,
            (byte)0x32, (byte)0x55, (byte)0x6d, (byte)0x2b, (byte)0x0a, (byte)0x54, (byte)0x46, (byte)0x36, (byte)0x31, (byte)0x62, (byte)0x4e, (byte)0x30, (byte)0x59, (byte)0x33, (byte)0x47, (byte)0x47,
            (byte)0x72, (byte)0x78, (byte)0x33, (byte)0x2b, (byte)0x50, (byte)0x4c, (byte)0x49, (byte)0x59, (byte)0x48, (byte)0x67, (byte)0x63, (byte)0x5a, (byte)0x6b, (byte)0x4e, (byte)0x4e, (byte)0x38,
            (byte)0x79, (byte)0x44, (byte)0x58, (byte)0x55, (byte)0x6e, (byte)0x41, (byte)0x72, (byte)0x46, (byte)0x75, (byte)0x58, (byte)0x50, (byte)0x78, (byte)0x4b, (byte)0x67, (byte)0x61, (byte)0x45,
            (byte)0x45, (byte)0x64, (byte)0x77, (byte)0x68, (byte)0x58, (byte)0x41, (byte)0x2f, (byte)0x34, (byte)0x77, (byte)0x2f, (byte)0x72, (byte)0x79, (byte)0x5a, (byte)0x53, (byte)0x5a, (byte)0x56,
            (byte)0x57, (byte)0x71, (byte)0x44, (byte)0x37, (byte)0x6f, (byte)0x0a, (byte)0x69, (byte)0x6e, (byte)0x67, (byte)0x46, (byte)0x66, (byte)0x71, (byte)0x6e, (byte)0x37, (byte)0x42, (byte)0x63,
            (byte)0x4b, (byte)0x4a, (byte)0x6b, (byte)0x2f, (byte)0x53, (byte)0x4c, (byte)0x6c, (byte)0x32, (byte)0x34, (byte)0x68, (byte)0x4e, (byte)0x50, (byte)0x30, (byte)0x78, (byte)0x79, (byte)0x2b,
            (byte)0x46, (byte)0x75, (byte)0x2f, (byte)0x75, (byte)0x37, (byte)0x4b, (byte)0x31, (byte)0x75, (byte)0x6e, (byte)0x4e, (byte)0x50, (byte)0x45, (byte)0x2f, (byte)0x4b, (byte)0x73, (byte)0x50,
            (byte)0x77, (byte)0x52, (byte)0x34, (byte)0x35, (byte)0x6f, (byte)0x6e, (byte)0x71, (byte)0x72, (byte)0x4c, (byte)0x77, (byte)0x66, (byte)0x6f, (byte)0x33, (byte)0x55, (byte)0x30, (byte)0x4c,
            (byte)0x5a, (byte)0x55, (byte)0x7a, (byte)0x63, (byte)0x66, (byte)0x61, (byte)0x0a, (byte)0x46, (byte)0x53, (byte)0x4f, (byte)0x63, (byte)0x45, (byte)0x70, (byte)0x62, (byte)0x6e, (byte)0x38,
            (byte)0x37, (byte)0x4b, (byte)0x75, (byte)0x6e, (byte)0x61, (byte)0x6d, (byte)0x4a, (byte)0x52, (byte)0x53, (byte)0x67, (byte)0x58, (byte)0x50, (byte)0x44, (byte)0x75, (byte)0x45, (byte)0x44,
            (byte)0x42, (byte)0x4f, (byte)0x4a, (byte)0x68, (byte)0x6b, (byte)0x2f, (byte)0x6d, (byte)0x6b, (byte)0x61, (byte)0x64, (byte)0x56, (byte)0x43, (byte)0x62, (byte)0x51, (byte)0x35, (byte)0x5a,
            (byte)0x49, (byte)0x41, (byte)0x78, (byte)0x76, (byte)0x72, (byte)0x75, (byte)0x52, (byte)0x4f, (byte)0x77, (byte)0x64, (byte)0x69, (byte)0x73, (byte)0x48, (byte)0x4d, (byte)0x33, (byte)0x6e,
            (byte)0x79, (byte)0x37, (byte)0x38, (byte)0x66, (byte)0x38, (byte)0x41, (byte)0x79, (byte)0x0a, (byte)0x42, (byte)0x4d, (byte)0x37, (byte)0x62, (byte)0x48, (byte)0x38, (byte)0x74, (byte)0x50,
            (byte)0x43, (byte)0x50, (byte)0x49, (byte)0x55, (byte)0x2f, (byte)0x32, (byte)0x76, (byte)0x79, (byte)0x6b, (byte)0x4d, (byte)0x43, (byte)0x76, (byte)0x42, (byte)0x63, (byte)0x74, (byte)0x51,
            (byte)0x6b, (byte)0x6e, (byte)0x41, (byte)0x42, (byte)0x5a, (byte)0x46, (byte)0x73, (byte)0x4b, (byte)0x35, (byte)0x6b, (byte)0x4f, (byte)0x5a, (byte)0x57, (byte)0x4f, (byte)0x53, (byte)0x30,
            (byte)0x59, (byte)0x51, (byte)0x4b, (byte)0x42, (byte)0x67, (byte)0x51, (byte)0x44, (byte)0x2b, (byte)0x69, (byte)0x74, (byte)0x67, (byte)0x57, (byte)0x4f, (byte)0x79, (byte)0x6c, (byte)0x6a,
            (byte)0x2f, (byte)0x70, (byte)0x77, (byte)0x36, (byte)0x39, (byte)0x64, (byte)0x4d, (byte)0x53, (byte)0x0a, (byte)0x38, (byte)0x47, (byte)0x37, (byte)0x2b, (byte)0x66, (byte)0x75, (byte)0x38,
            (byte)0x41, (byte)0x66, (byte)0x51, (byte)0x74, (byte)0x50, (byte)0x4b, (byte)0x45, (byte)0x74, (byte)0x6d, (byte)0x45, (byte)0x70, (byte)0x67, (byte)0x51, (byte)0x4e, (byte)0x79, (byte)0x2b,
            (byte)0x32, (byte)0x45, (byte)0x76, (byte)0x6f, (byte)0x42, (byte)0x6f, (byte)0x34, (byte)0x54, (byte)0x5a, (byte)0x6d, (byte)0x43, (byte)0x62, (byte)0x71, (byte)0x4b, (byte)0x55, (byte)0x4e,
            (byte)0x2b, (byte)0x72, (byte)0x4e, (byte)0x78, (byte)0x52, (byte)0x6b, (byte)0x49, (byte)0x7a, (byte)0x68, (byte)0x2f, (byte)0x4e, (byte)0x36, (byte)0x54, (byte)0x68, (byte)0x59, (byte)0x31,
            (byte)0x2f, (byte)0x4b, (byte)0x75, (byte)0x6b, (byte)0x68, (byte)0x62, (byte)0x4e, (byte)0x4f, (byte)0x51, (byte)0x0a, (byte)0x79, (byte)0x34, (byte)0x7a, (byte)0x49, (byte)0x51, (byte)0x32,
            (byte)0x31, (byte)0x70, (byte)0x43, (byte)0x50, (byte)0x76, (byte)0x53, (byte)0x55, (byte)0x47, (byte)0x2f, (byte)0x2f, (byte)0x4d, (byte)0x42, (byte)0x37, (byte)0x75, (byte)0x55, (byte)0x34,
            (byte)0x45, (byte)0x6c, (byte)0x33, (byte)0x78, (byte)0x57, (byte)0x7a, (byte)0x53, (byte)0x6f, (byte)0x68, (byte)0x6a, (byte)0x68, (byte)0x55, (byte)0x41, (byte)0x7a, (byte)0x6b, (byte)0x2b,
            (byte)0x67, (byte)0x72, (byte)0x37, (byte)0x67, (byte)0x46, (byte)0x50, (byte)0x75, (byte)0x38, (byte)0x6c, (byte)0x6e, (byte)0x64, (byte)0x52, (byte)0x42, (byte)0x68, (byte)0x64, (byte)0x71,
            (byte)0x31, (byte)0x73, (byte)0x6a, (byte)0x41, (byte)0x34, (byte)0x47, (byte)0x54, (byte)0x74, (byte)0x43, (byte)0x41, (byte)0x0a, (byte)0x6d, (byte)0x4b, (byte)0x37, (byte)0x46, (byte)0x71,
            (byte)0x47, (byte)0x52, (byte)0x4f, (byte)0x42, (byte)0x7a, (byte)0x2f, (byte)0x54, (byte)0x61, (byte)0x6c, (byte)0x73, (byte)0x6c, (byte)0x58, (byte)0x4d, (byte)0x55, (byte)0x61, (byte)0x46,
            (byte)0x79, (byte)0x41, (byte)0x4d, (byte)0x4d, (byte)0x51, (byte)0x4b, (byte)0x42, (byte)0x67, (byte)0x51, (byte)0x44, (byte)0x53, (byte)0x65, (byte)0x41, (byte)0x2b, (byte)0x61, (byte)0x78,
            (byte)0x42, (byte)0x50, (byte)0x63, (byte)0x51, (byte)0x7a, (byte)0x65, (byte)0x44, (byte)0x5a, (byte)0x63, (byte)0x50, (byte)0x66, (byte)0x68, (byte)0x6c, (byte)0x37, (byte)0x51, (byte)0x51,
            (byte)0x48, (byte)0x61, (byte)0x4b, (byte)0x75, (byte)0x39, (byte)0x54, (byte)0x4c, (byte)0x56, (byte)0x37, (byte)0x4a, (byte)0x35, (byte)0x0a, (byte)0x7a, (byte)0x6f, (byte)0x39, (byte)0x6d,
            (byte)0x4f, (byte)0x75, (byte)0x53, (byte)0x41, (byte)0x54, (byte)0x6c, (byte)0x4f, (byte)0x57, (byte)0x72, (byte)0x6c, (byte)0x4d, (byte)0x77, (byte)0x33, (byte)0x2b, (byte)0x6a, (byte)0x35,
            (byte)0x55, (byte)0x47, (byte)0x33, (byte)0x33, (byte)0x45, (byte)0x45, (byte)0x38, (byte)0x45, (byte)0x65, (byte)0x42, (byte)0x6c, (byte)0x48, (byte)0x72, (byte)0x6d, (byte)0x5a, (byte)0x49,
            (byte)0x6c, (byte)0x55, (byte)0x49, (byte)0x48, (byte)0x71, (byte)0x72, (byte)0x50, (byte)0x75, (byte)0x4a, (byte)0x35, (byte)0x4f, (byte)0x59, (byte)0x76, (byte)0x32, (byte)0x71, (byte)0x4a,
            (byte)0x49, (byte)0x64, (byte)0x6d, (byte)0x75, (byte)0x6e, (byte)0x41, (byte)0x76, (byte)0x4b, (byte)0x6e, (byte)0x64, (byte)0x43, (byte)0x55, (byte)0x0a, (byte)0x71, (byte)0x76, (byte)0x55,
            (byte)0x6f, (byte)0x4f, (byte)0x4d, (byte)0x48, (byte)0x4b, (byte)0x43, (byte)0x4b, (byte)0x36, (byte)0x70, (byte)0x38, (byte)0x55, (byte)0x51, (byte)0x58, (byte)0x4e, (byte)0x7a, (byte)0x77,
            (byte)0x62, (byte)0x73, (byte)0x62, (byte)0x44, (byte)0x6a, (byte)0x5a, (byte)0x73, (byte)0x67, (byte)0x58, (byte)0x36, (byte)0x59, (byte)0x30, (byte)0x63, (byte)0x32, (byte)0x32, (byte)0x6e,
            (byte)0x39, (byte)0x76, (byte)0x63, (byte)0x49, (byte)0x34, (byte)0x55, (byte)0x44, (byte)0x48, (byte)0x44, (byte)0x6b, (byte)0x49, (byte)0x35, (byte)0x56, (byte)0x43, (byte)0x79, (byte)0x6b,
            (byte)0x52, (byte)0x54, (byte)0x41, (byte)0x6c, (byte)0x69, (byte)0x7a, (byte)0x7a, (byte)0x50, (byte)0x7a, (byte)0x33, (byte)0x66, (byte)0x6f, (byte)0x54, (byte)0x0a, (byte)0x45, (byte)0x7a,
            (byte)0x4e, (byte)0x66, (byte)0x73, (byte)0x6f, (byte)0x34, (byte)0x67, (byte)0x4a, (byte)0x77, (byte)0x4b, (byte)0x42, (byte)0x67, (byte)0x51, (byte)0x44, (byte)0x78, (byte)0x43, (byte)0x72,
            (byte)0x42, (byte)0x59, (byte)0x56, (byte)0x51, (byte)0x76, (byte)0x38, (byte)0x47, (byte)0x6b, (byte)0x62, (byte)0x39, (byte)0x2b, (byte)0x6e, (byte)0x34, (byte)0x34, (byte)0x53, (byte)0x43,
            (byte)0x45, (byte)0x44, (byte)0x74, (byte)0x56, (byte)0x5a, (byte)0x45, (byte)0x34, (byte)0x79, (byte)0x55, (byte)0x4c, (byte)0x66, (byte)0x37, (byte)0x2f, (byte)0x75, (byte)0x32, (byte)0x42,
            (byte)0x72, (byte)0x52, (byte)0x2b, (byte)0x45, (byte)0x43, (byte)0x68, (byte)0x4c, (byte)0x64, (byte)0x55, (byte)0x38, (byte)0x4b, (byte)0x67, (byte)0x6f, (byte)0x4c, (byte)0x0a, (byte)0x4f,
            (byte)0x47, (byte)0x39, (byte)0x5a, (byte)0x66, (byte)0x56, (byte)0x76, (byte)0x69, (byte)0x49, (byte)0x48, (byte)0x57, (byte)0x2b, (byte)0x63, (byte)0x36, (byte)0x46, (byte)0x46, (byte)0x2b,
            (byte)0x2f, (byte)0x30, (byte)0x57, (byte)0x4f, (byte)0x2f, (byte)0x6c, (byte)0x6a, (byte)0x66, (byte)0x46, (byte)0x49, (byte)0x71, (byte)0x69, (byte)0x66, (byte)0x44, (byte)0x4c, (byte)0x42,
            (byte)0x45, (byte)0x61, (byte)0x37, (byte)0x2b, (byte)0x6b, (byte)0x59, (byte)0x6f, (byte)0x6d, (byte)0x67, (byte)0x35, (byte)0x46, (byte)0x2b, (byte)0x6c, (byte)0x41, (byte)0x6b, (byte)0x6d,
            (byte)0x77, (byte)0x41, (byte)0x66, (byte)0x64, (byte)0x6b, (byte)0x6f, (byte)0x74, (byte)0x66, (byte)0x39, (byte)0x54, (byte)0x2f, (byte)0x35, (byte)0x4d, (byte)0x76, (byte)0x46, (byte)0x0a,
            (byte)0x57, (byte)0x70, (byte)0x45, (byte)0x53, (byte)0x68, (byte)0x72, (byte)0x54, (byte)0x73, (byte)0x72, (byte)0x4d, (byte)0x54, (byte)0x6e, (byte)0x4b, (byte)0x5a, (byte)0x41, (byte)0x75,
            (byte)0x2f, (byte)0x59, (byte)0x64, (byte)0x57, (byte)0x6c, (byte)0x46, (byte)0x6a, (byte)0x6b, (byte)0x47, (byte)0x72, (byte)0x6a, (byte)0x4a, (byte)0x4e, (byte)0x6d, (byte)0x64, (byte)0x78,
            (byte)0x37, (byte)0x36, (byte)0x52, (byte)0x4a, (byte)0x6e, (byte)0x77, (byte)0x33, (byte)0x6c, (byte)0x61, (byte)0x69, (byte)0x34, (byte)0x42, (byte)0x32, (byte)0x59, (byte)0x79, (byte)0x59,
            (byte)0x6c, (byte)0x41, (byte)0x49, (byte)0x2b, (byte)0x6b, (byte)0x37, (byte)0x6d, (byte)0x51, (byte)0x63, (byte)0x51, (byte)0x4b, (byte)0x42, (byte)0x67, (byte)0x51, (byte)0x43, (byte)0x4f,
            (byte)0x0a, (byte)0x50, (byte)0x44, (byte)0x6b, (byte)0x2f, (byte)0x66, (byte)0x48, (byte)0x64, (byte)0x33, (byte)0x69, (byte)0x61, (byte)0x32, (byte)0x41, (byte)0x6a, (byte)0x55, (byte)0x75,
            (byte)0x63, (byte)0x53, (byte)0x66, (byte)0x44, (byte)0x56, (byte)0x69, (byte)0x46, (byte)0x73, (byte)0x55, (byte)0x55, (byte)0x75, (byte)0x2f, (byte)0x79, (byte)0x49, (byte)0x2b, (byte)0x67,
            (byte)0x6e, (byte)0x7a, (byte)0x57, (byte)0x4b, (byte)0x33, (byte)0x30, (byte)0x65, (byte)0x45, (byte)0x4f, (byte)0x6e, (byte)0x4f, (byte)0x42, (byte)0x38, (byte)0x6d, (byte)0x6e, (byte)0x37,
            (byte)0x36, (byte)0x4c, (byte)0x72, (byte)0x67, (byte)0x42, (byte)0x6c, (byte)0x55, (byte)0x74, (byte)0x62, (byte)0x68, (byte)0x70, (byte)0x67, (byte)0x41, (byte)0x43, (byte)0x7a, (byte)0x48,
            (byte)0x36, (byte)0x0a, (byte)0x63, (byte)0x58, (byte)0x41, (byte)0x4b, (byte)0x61, (byte)0x57, (byte)0x52, (byte)0x67, (byte)0x54, (byte)0x67, (byte)0x71, (byte)0x66, (byte)0x54, (byte)0x66,
            (byte)0x75, (byte)0x2f, (byte)0x5a, (byte)0x4d, (byte)0x30, (byte)0x59, (byte)0x73, (byte)0x39, (byte)0x6d, (byte)0x71, (byte)0x32, (byte)0x53, (byte)0x5a, (byte)0x57, (byte)0x33, (byte)0x65,
            (byte)0x52, (byte)0x53, (byte)0x4e, (byte)0x46, (byte)0x37, (byte)0x4e, (byte)0x75, (byte)0x64, (byte)0x32, (byte)0x74, (byte)0x4c, (byte)0x63, (byte)0x54, (byte)0x73, (byte)0x63, (byte)0x48,
            (byte)0x4a, (byte)0x31, (byte)0x78, (byte)0x67, (byte)0x78, (byte)0x48, (byte)0x49, (byte)0x77, (byte)0x54, (byte)0x4b, (byte)0x43, (byte)0x58, (byte)0x48, (byte)0x38, (byte)0x52, (byte)0x77,
            (byte)0x66, (byte)0x6f, (byte)0x0a, (byte)0x6b, (byte)0x70, (byte)0x66, (byte)0x45, (byte)0x74, (byte)0x58, (byte)0x4d, (byte)0x47, (byte)0x48, (byte)0x35, (byte)0x77, (byte)0x68, (byte)0x38,
            (byte)0x42, (byte)0x56, (byte)0x67, (byte)0x66, (byte)0x39, (byte)0x52, (byte)0x58, (byte)0x4c, (byte)0x69, (byte)0x33, (byte)0x42, (byte)0x71, (byte)0x77, (byte)0x6f, (byte)0x50, (byte)0x4b,
            (byte)0x74, (byte)0x71, (byte)0x50, (byte)0x2f, (byte)0x59, (byte)0x2f, (byte)0x67, (byte)0x67, (byte)0x6f, (byte)0x6a, (byte)0x75, (byte)0x54, (byte)0x51, (byte)0x4b, (byte)0x42, (byte)0x67,
            (byte)0x51, (byte)0x43, (byte)0x66, (byte)0x2b, (byte)0x62, (byte)0x77, (byte)0x6f, (byte)0x62, (byte)0x66, (byte)0x4e, (byte)0x79, (byte)0x2f, (byte)0x45, (byte)0x52, (byte)0x6a, (byte)0x55,
            (byte)0x4a, (byte)0x37, (byte)0x6c, (byte)0x0a, (byte)0x6c, (byte)0x6c, (byte)0x65, (byte)0x62, (byte)0x7a, (byte)0x74, (byte)0x66, (byte)0x63, (byte)0x6a, (byte)0x41, (byte)0x57, (byte)0x41,
            (byte)0x2f, (byte)0x2b, (byte)0x39, (byte)0x62, (byte)0x31, (byte)0x37, (byte)0x6c, (byte)0x71, (byte)0x49, (byte)0x64, (byte)0x61, (byte)0x31, (byte)0x43, (byte)0x71, (byte)0x34, (byte)0x41,
            (byte)0x66, (byte)0x59, (byte)0x77, (byte)0x59, (byte)0x78, (byte)0x72, (byte)0x52, (byte)0x4a, (byte)0x54, (byte)0x67, (byte)0x50, (byte)0x31, (byte)0x32, (byte)0x62, (byte)0x62, (byte)0x6d,
            (byte)0x35, (byte)0x42, (byte)0x63, (byte)0x52, (byte)0x6a, (byte)0x45, (byte)0x78, (byte)0x6d, (byte)0x36, (byte)0x55, (byte)0x4a, (byte)0x42, (byte)0x65, (byte)0x56, (byte)0x63, (byte)0x50,
            (byte)0x4d, (byte)0x44, (byte)0x31, (byte)0x70, (byte)0x0a, (byte)0x35, (byte)0x6c, (byte)0x4a, (byte)0x78, (byte)0x51, (byte)0x34, (byte)0x6a, (byte)0x63, (byte)0x47, (byte)0x78, (byte)0x48,
            (byte)0x4e, (byte)0x79, (byte)0x54, (byte)0x42, (byte)0x4d, (byte)0x50, (byte)0x65, (byte)0x43, (byte)0x68, (byte)0x4f, (byte)0x47, (byte)0x77, (byte)0x31, (byte)0x41, (byte)0x4f, (byte)0x74,
            (byte)0x46, (byte)0x31, (byte)0x66, (byte)0x42, (byte)0x6b, (byte)0x49, (byte)0x6d, (byte)0x38, (byte)0x48, (byte)0x59, (byte)0x62, (byte)0x78, (byte)0x59, (byte)0x30, (byte)0x57, (byte)0x4a,
            (byte)0x50, (byte)0x30, (byte)0x6e, (byte)0x50, (byte)0x48, (byte)0x53, (byte)0x30, (byte)0x6c, (byte)0x45, (byte)0x56, (byte)0x72, (byte)0x34, (byte)0x57, (byte)0x69, (byte)0x6c, (byte)0x43,
            (byte)0x6c, (byte)0x77, (byte)0x7a, (byte)0x6b, (byte)0x44, (byte)0x0a, (byte)0x32, (byte)0x7a, (byte)0x64, (byte)0x30, (byte)0x31, (byte)0x67, (byte)0x6a, (byte)0x31, (byte)0x70, (byte)0x32,
            (byte)0x53, (byte)0x2b, (byte)0x68, (byte)0x41, (byte)0x4f, (byte)0x71, (byte)0x38, (byte)0x47, (byte)0x54, (byte)0x4f, (byte)0x35, (byte)0x78, (byte)0x70, (byte)0x58, (byte)0x36, (byte)0x67,
            (byte)0x3d, (byte)0x3d, (byte)0x0a, (byte)0x2d, (byte)0x2d, (byte)0x2d, (byte)0x2d, (byte)0x2d, (byte)0x45, (byte)0x4e, (byte)0x44, (byte)0x20, (byte)0x50, (byte)0x52, (byte)0x49, (byte)0x56,
            (byte)0x41, (byte)0x54, (byte)0x45, (byte)0x20, (byte)0x4b, (byte)0x45, (byte)0x59, (byte)0x2d, (byte)0x2d, (byte)0x2d, (byte)0x2d, (byte)0x2d, (byte)0x0a
    };
};