package edu.wpi.appjudicator.ssl;


import android.util.Log;

import java.util.*;

import org.bouncycastle.asn1.edec.EdECObjectIdentifiers;
import org.bouncycastle.asn1.x509.AlgorithmIdentifier;
import org.bouncycastle.asn1.x509.SubjectPublicKeyInfo;
import org.bouncycastle.crypto.CipherParameters;
import org.bouncycastle.crypto.DataLengthException;
import org.bouncycastle.crypto.InvalidCipherTextException;
import org.bouncycastle.crypto.digests.SHA256Digest;
import org.bouncycastle.crypto.engines.AESEngine;
import org.bouncycastle.crypto.macs.HMac;
import org.bouncycastle.crypto.modes.GCMBlockCipher;
import org.bouncycastle.crypto.params.AEADParameters;
import org.bouncycastle.crypto.params.KeyParameter;

import org.bouncycastle.jcajce.spec.XDHParameterSpec;

import org.bouncycastle.jce.provider.BouncyCastleProvider;


import org.bouncycastle.util.io.pem.PemObject;
import org.bouncycastle.util.io.pem.PemReader;

import java.io.File;
import java.io.FileReader;
import java.security.KeyFactory;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.PublicKey;
import java.security.SecureRandom;
import java.security.Signature;
import java.security.cert.X509Certificate;
import java.security.interfaces.RSAPrivateKey;
import java.security.interfaces.RSAPublicKey;


import java.security.spec.MGF1ParameterSpec;
import java.security.spec.PKCS8EncodedKeySpec;
import java.security.spec.PSSParameterSpec;
import java.security.spec.X509EncodedKeySpec;
import java.util.Arrays;
import java.util.LinkedList;

import javax.crypto.KeyAgreement;
import javax.crypto.Mac;
import javax.crypto.spec.SecretKeySpec;

import edu.wpi.appjudicator.ssl.tls_packet_parsing.TLS_Builder;
import edu.wpi.appjudicator.ssl.tls_packet_parsing.TLS_Change_Cypher_Spec_Message;
import edu.wpi.appjudicator.ssl.tls_packet_parsing.TLS_Encrypted_Application_Data;
import edu.wpi.appjudicator.ssl.tls_packet_parsing.TLS_Enum_Content_Type;
import edu.wpi.appjudicator.ssl.tls_packet_parsing.TLS_Handshake_Sublayer;
import edu.wpi.appjudicator.ssl.tls_packet_parsing.TLS_List_Builder;
import edu.wpi.appjudicator.ssl.tls_packet_parsing.TLS_Packet;
import edu.wpi.appjudicator.ssl.tls_packet_parsing.TLS_Record_Layer;
import edu.wpi.appjudicator.ssl.tls_packet_parsing.TLS_Handshake_Certificate;
import edu.wpi.appjudicator.ssl.tls_packet_parsing.TLS_Handshake_Client_Hello;
import edu.wpi.appjudicator.ssl.tls_packet_parsing.TLS_Handshake_Client_Key_Exchange;
import edu.wpi.appjudicator.ssl.tls_packet_parsing.TLS_Handshake_Encrypted_Message;
import edu.wpi.appjudicator.ssl.tls_packet_parsing.TLS_Handshake_Extension_Abstract;
import edu.wpi.appjudicator.ssl.tls_packet_parsing.TLS_Handshake_Extension_EC_POINT_FORMATS;
import edu.wpi.appjudicator.ssl.tls_packet_parsing.TLS_Handshake_Extension_SERVER_NAME;
import edu.wpi.appjudicator.ssl.tls_packet_parsing.TLS_Handshake_Extension_Unknown;
import edu.wpi.appjudicator.ssl.tls_packet_parsing.TLS_Handshake_Random_Component_Subclass;
import edu.wpi.appjudicator.ssl.tls_packet_parsing.TLS_Handshake_Server_Hello;
import edu.wpi.appjudicator.ssl.tls_packet_parsing.TLS_Handshake_Server_Key_Exchange;
import edu.wpi.appjudicator.ssl.tls_packet_parsing.TLS_Sublayer;

public class TLS_Server_State_Machine {
    private static String TAG = "TLS SERVER STATE MACHINE";

    static final int VERIFY_DATA_SIZE = 12; //12 octets for verify data in the "Handshake Finished Messages"
    static final int MASTER_SECRET_LEN = 48;
    static final int KEY_BLOCK_SIZE = 40;
    static final int AES_KEY_SIZE = 16;
    static final int GCM_IV_SIZE = 12;
    static final int GCM_IV_IMPLICIT_SIZE = 4;
    static final int GCM_IV_EXPLICIT_SIZE = 8;
    static final int MAC_SIZE = 16;
    static final byte[] FINISHED_HEADER = new byte[] {0x14, 0x00, 0x00, 0x0c};
    static final byte[] GENERIC_ADDITIONAL_DATA = new byte[] {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x03, 0x03, 0x00, 0x00};
    public enum TLS_Server_State {
        STATE_CREATED,
        STATE_RECEIVED_CLIENT_HELLO,
        STATE_SENT_SERVER_HELLO,
        STATE_RECEIVED_CLIENT_KEY_EXCHANGE,
        STATE_ENCRYPTED_COMMUNICATION,
        STATE_CLOSED_COMMUNICATION
    }

    static String ca_cert_str = "/storage/emulated/0/Download/cacert.pem";
    static String server_private_key_str = "/storage/emulated/0/Download/serverkey.key";
    static String server_public_key_str = "/storage/emulated/0/Download/serverkey.pem";


    TLS_Server_State currentState;
    KeyPair serverECDHEKeyPair;
    KeyPair serverRSAKeyPair;
    PublicKey clientDHEPublicKey;

    TLS_Handshake_Random_Component_Subclass server_random;

    TLS_Handshake_Random_Component_Subclass client_random;

    TLS_Handshake_Server_Key_Exchange.ECDHE_Server_Params ecdhe_server_params;



    byte[] master_secret;
    byte[] key_block;
    byte[] client_aes_key;
    byte[] server_aes_key;
    byte[] client_gcm_iv;
    byte[] server_gcm_iv;

    byte[] client_hello_raw; //All the handshake messages concatenated
    byte[] server_hello_raw;
    byte[] server_certs_raw;
    byte[] server_key_exchange_raw;
    byte[] server_hello_done_raw;

    byte[] client_verify_data_raw;
    byte[] server_verify_data_raw;
    long client_seq_num;
    long server_seq_num;

    byte[] buf_client; //for debugging

    Random rand_class;

    GCMBlockCipher client_cipher;
    GCMBlockCipher server_cipher;

    public TLS_Server_State_Machine() {
        currentState = TLS_Server_State.STATE_CREATED;
        rand_class = new Random();
        try {
            this.client_aes_key = new byte[this.AES_KEY_SIZE];
            this.server_aes_key = new byte[this.AES_KEY_SIZE];

            this.client_gcm_iv = new byte[this.GCM_IV_SIZE];
            this.server_gcm_iv = new byte[this.GCM_IV_SIZE];

            this.client_seq_num = 0;
            this.server_seq_num = 0;

            this.client_verify_data_raw = new byte[this.VERIFY_DATA_SIZE];
            this.server_verify_data_raw = new byte[this.VERIFY_DATA_SIZE];



//            TLS_Handshake_Extension_SERVER_NAME server_name_ext = client_hello.getServerNameExtension();
            //String server_name = String(server_name_ext.list.get(0).name_bytes);
            //Certificate c = generateCertificate(server_name);

            //temporary solution
//            File rsa_private_key_file = new File(server_private_key_str);
            //RSAPrivateKey rsa_private_key = readPrivateKey(rsa_private_key_file);
//            File rsa_public_key_file = new File(server_public_key_str);
            //RSAPublicKey rsa_public_key = readPublicKey(rsa_public_key_file);

            //serverRSAKeyPair = new KeyPair(rsa_public_key, rsa_private_key);

//                File ca_cert_file = new File(ca_cert_str);
//            byte[] ca_cert_data = new byte[(int)ca_cert_file.length()];
//            FileInputStream fs = new FileInputStream(ca_cert_file);
//            fs.read(ca_cert_data);
//
//            byte[] server_cert_data = new byte[(int)rsa_public_key_file.length()];
//            FileInputStream fs2 = new FileInputStream(rsa_public_key_file);
//            fs2.read(server_cert_data);


        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    //Takes in a TLS Packet and produces appropriate response bytes
    //In case of error, simply throws an exception, indicating that no bytes should be sent
    //Use the raw data to concatenate
    public byte[] process_TLS_Handshake(TLS_Packet p) throws Exception{
        switch (this.currentState) {
            case STATE_CREATED:
                try {
                    TLS_Handshake_Client_Hello.class.cast(p.record_layers.get(0).payload);


                    TLS_Handshake_Client_Hello client_hello = (TLS_Handshake_Client_Hello) p.record_layers.get(0).payload;
                    client_hello_raw =client_hello.getRawData();  //need to keep track

                    TLS_Handshake_Random_Component_Subclass.TLS_Handshake_Random_Component_Subclass_Builder client_random_builder = new TLS_Handshake_Random_Component_Subclass.TLS_Handshake_Random_Component_Subclass_Builder(client_hello.random_component);
                    client_random = client_random_builder.build();

                    updateSessionID();
                    updateServerRandomID();
                    generate_fake_certificate(client_hello.getServerNameExtension());
                    generate_ecdhe_key_and_sign_server_key_exchange();

                    tlsServerHelloTemplateBuilder.fixSizeFields();


                    currentState = TLS_Server_State.STATE_SENT_SERVER_HELLO;
                    this.server_hello_raw = ((TLS_Record_Layer.TLS_Record_Layer_Builder) this.tlsServerHelloTemplateBuilder.record_layer_list_builder.get(0)).payload_builder.build().getRawData(); //need to keep track
                    this.server_certs_raw = ((TLS_Record_Layer.TLS_Record_Layer_Builder) this.tlsServerHelloTemplateBuilder.record_layer_list_builder.get(1)).payload_builder.build().getRawData(); //need to keep track
                    this.server_key_exchange_raw =((TLS_Record_Layer.TLS_Record_Layer_Builder) this.tlsServerHelloTemplateBuilder.record_layer_list_builder.get(2)).payload_builder.build().getRawData(); //need to keep track
                    this.server_hello_done_raw = ((TLS_Record_Layer.TLS_Record_Layer_Builder) this.tlsServerHelloTemplateBuilder.record_layer_list_builder.get(3)).payload_builder.build().getRawData(); //need to keep track

                    return this.tlsServerHelloTemplateBuilder.build().getRawData();
                } catch (ClassCastException e) {
                    //Received the wrong packet for this stage, simply return
                    throw new Exception("Current State Is STATE_CREATED, but didn't receive Client Hello");
                } catch (Exception e) {
                    e.printStackTrace();
                    throw e;
                }
            case STATE_RECEIVED_CLIENT_HELLO:
                break;  //nothing to do here, just let the previous step complete
            case STATE_SENT_SERVER_HELLO:
                try {
                    TLS_Handshake_Client_Key_Exchange.class.cast(p.record_layers.get(0).payload);
                    TLS_Change_Cypher_Spec_Message.class.cast(p.record_layers.get(1).payload);
                    TLS_Handshake_Encrypted_Message.class.cast(p.record_layers.get(2).payload);

                    TLS_Handshake_Client_Key_Exchange client_key_exchange = (TLS_Handshake_Client_Key_Exchange) p.record_layers.get(0).payload;
                    TLS_Handshake_Encrypted_Message client_finished = (TLS_Handshake_Encrypted_Message) p.record_layers.get(2).payload;

                    //we need to hash all the handshake messages up to the client finished message to verify the client finished message, and also to send our own server finished
                    byte[] client_key_exchange_bytes = client_key_exchange.getRawData();
                    byte[] client_finished_bytes = client_finished.getRawData();

                    do_ECDHE_key_exchange(client_key_exchange);
                    generateEphemeralKeysFromMaster();

                    //do not forget to update the explicit part of the IV for client
                    System.arraycopy(client_finished_bytes, 0, this.client_gcm_iv, this.GCM_IV_IMPLICIT_SIZE, this.GCM_IV_EXPLICIT_SIZE);

                    //this is what we have to verify
                    int totalCopied = 0;
                    byte[] bytesToSignAndVerify_client = new byte[this.client_hello_raw.length + this.server_hello_raw.length + this.server_certs_raw.length + this.server_key_exchange_raw.length + this.server_hello_done_raw.length + client_key_exchange_bytes.length];
                    System.arraycopy(this.client_hello_raw, 0, bytesToSignAndVerify_client, totalCopied, this.client_hello_raw.length);
                    totalCopied += this.client_hello_raw.length;
                    System.arraycopy(this.server_hello_raw, 0, bytesToSignAndVerify_client, totalCopied, this.server_hello_raw.length);
                    totalCopied += this.server_hello_raw.length;
                    System.arraycopy(this.server_certs_raw, 0, bytesToSignAndVerify_client, totalCopied, this.server_certs_raw.length);
                    totalCopied += this.server_certs_raw.length;
                    System.arraycopy(this.server_key_exchange_raw, 0, bytesToSignAndVerify_client, totalCopied, this.server_key_exchange_raw.length);
                    totalCopied += this.server_key_exchange_raw.length;
                    System.arraycopy(this.server_hello_done_raw, 0, bytesToSignAndVerify_client, totalCopied, this.server_hello_done_raw.length);
                    totalCopied += this.server_hello_done_raw.length;
                    System.arraycopy(client_key_exchange_bytes, 0, bytesToSignAndVerify_client,  totalCopied, client_key_exchange_bytes.length);
                    //totalCopied += client_key_exchange_bytes.length;


                    MessageDigest md = null;
                    md = MessageDigest.getInstance("SHA-256");

                    md.reset();
                    md.update(bytesToSignAndVerify_client, 0, bytesToSignAndVerify_client.length);
                    byte[] bytesToSignAndVerify_ClientHashed = md.digest();
                    this.client_verify_data_raw = PRF(this.master_secret, "client finished", bytesToSignAndVerify_ClientHashed, VERIFY_DATA_SIZE);
                    byte[] client_finished_packet_plaintext = Arrays.copyOf( this.FINISHED_HEADER, this.VERIFY_DATA_SIZE + FINISHED_HEADER.length);
                    System.arraycopy(this.client_verify_data_raw, 0, client_finished_packet_plaintext, this.FINISHED_HEADER.length, this.VERIFY_DATA_SIZE);

                    //we need to construct the authentication data
                    byte[] additionalData_client = generateAdditionalData(TLS_Enum_Content_Type.RECORD_LAYER_CONTENT_TYPE_HANDSHAKE.value(), computePlaintextMessageLen(client_finished_bytes), true);

                    this.buf_client = decryptAndVerifyClientBytes(client_finished_bytes, 0, additionalData_client);

                    if (this.buf_client == null) {
                        throw new IllegalStateException("Client message failed with MAC authentication");
                    }

                    if (!SSL_Helper.compareByteArrays(buf_client, client_finished_packet_plaintext)) {
                        Log.e(TAG, "Received unexpected encrypted bytes from client!");
                        throw new IllegalArgumentException("Should have received a handshake finished message");
                    }

                    //send your own change cipher spec and Encrypted Handshake Message                    return null;
                    byte[] bytesToSignAndVerify_server = Arrays.copyOf(bytesToSignAndVerify_client, bytesToSignAndVerify_client.length + client_finished_packet_plaintext.length);
                    System.arraycopy(client_finished_packet_plaintext, 0, bytesToSignAndVerify_server, bytesToSignAndVerify_client.length, client_finished_packet_plaintext.length);
                    md.reset();
                    md.update(bytesToSignAndVerify_server, 0, bytesToSignAndVerify_server.length);
                    byte[] bytesToSignAndVerify_ServerHashed = md.digest();

                    this.server_verify_data_raw = PRF(this.master_secret, "server finished", bytesToSignAndVerify_ServerHashed, VERIFY_DATA_SIZE);
                    byte[] server_finished_packet_plaintext = Arrays.copyOf(this.FINISHED_HEADER, this.FINISHED_HEADER.length + this.server_verify_data_raw.length);
                    System.arraycopy(this.server_verify_data_raw, 0, server_finished_packet_plaintext, this.FINISHED_HEADER.length, this.server_verify_data_raw.length);

                    byte[] additionalData_server = generateAdditionalData(TLS_Enum_Content_Type.RECORD_LAYER_CONTENT_TYPE_HANDSHAKE.value(), (short)(server_finished_packet_plaintext.length), false);
                    byte[] server_finished_encrypted_msg_raw = encryptServerBytes(server_finished_packet_plaintext, additionalData_server);

                    //Specific to the handshake, generate the record layers which will contain the messages
                    TLS_Change_Cypher_Spec_Message.TLS_Change_Cypher_Spec_Message_Builder change_cipher_b = new TLS_Change_Cypher_Spec_Message.TLS_Change_Cypher_Spec_Message_Builder();
                    TLS_Record_Layer.TLS_Record_Layer_Builder record_b = new TLS_Record_Layer.TLS_Record_Layer_Builder(TLS_Enum_Content_Type.RECORD_LAYER_CONTENT_TYPE_CHANGE_CYPHER_SPEC.value());
                    record_b.payload_builder = (TLS_Builder) change_cipher_b;

                    TLS_Handshake_Encrypted_Message.TLS_Handshake_Encrypted_Message_Builder finished_b = new TLS_Handshake_Encrypted_Message.TLS_Handshake_Encrypted_Message_Builder(server_finished_encrypted_msg_raw);
                    TLS_Record_Layer.TLS_Record_Layer_Builder record2_b = new TLS_Record_Layer.TLS_Record_Layer_Builder(TLS_Enum_Content_Type.RECORD_LAYER_CONTENT_TYPE_HANDSHAKE.value());
                    record2_b.payload_builder = (TLS_Builder) finished_b;

                    record2_b.fixAndGetTotalSize();
                    TLS_List_Builder<TLS_Record_Layer> list_b = new TLS_List_Builder<>();

                    list_b.add(record_b);
                    list_b.add(record2_b);
                    TLS_Packet.TLS_Packet_Builder packet_b = new TLS_Packet.TLS_Packet_Builder();
                    packet_b.record_layer_list_builder = list_b;
                    packet_b.fixSizeFields();

                    this.currentState = TLS_Server_State.STATE_ENCRYPTED_COMMUNICATION;

                    return packet_b.build().getRawData();

                } catch (DataLengthException e) {
                    e.printStackTrace();
                }
                catch (IllegalStateException e) {
                    e.printStackTrace();

                } catch (InvalidCipherTextException e) {
                    e.printStackTrace();

                }
                catch (ClassCastException e) {
                    //Received the wrong packet for this stage, simply return
                    e.printStackTrace();
                    //throw new Exception("Current State Is STATE_SENT_SERVER_HELLO, but didn't receive Client Key Exchange, Change Cipher Spec and Encrypted Handshake Message");
                } catch (NoSuchAlgorithmException e) {
                    e.printStackTrace();
                }

            default: STATE_ENCRYPTED_COMMUNICATION:
                return null;
        }
        return null;
    }

    public void updateSessionID() {
        TLS_Record_Layer.TLS_Record_Layer_Builder record_layer = (TLS_Record_Layer.TLS_Record_Layer_Builder) this.tlsServerHelloTemplateBuilder.record_layer_list_builder.get(0);
        TLS_Handshake_Server_Hello.TLS_Handshake_Server_Hello_Builder server_hello_builder =  TLS_Handshake_Server_Hello.TLS_Handshake_Server_Hello_Builder.class.cast(record_layer.payload_builder);
        rand_class.nextBytes(server_hello_builder.session_ID_builder.sessionID);
    }

    public void updateServerRandomID() {
        TLS_Record_Layer.TLS_Record_Layer_Builder record_layer = (TLS_Record_Layer.TLS_Record_Layer_Builder) this.tlsServerHelloTemplateBuilder.record_layer_list_builder.get(0);
        TLS_Handshake_Server_Hello.TLS_Handshake_Server_Hello_Builder server_hello_builder =  TLS_Handshake_Server_Hello.TLS_Handshake_Server_Hello_Builder.class.cast(record_layer.payload_builder);
        rand_class.nextBytes(server_hello_builder.random_component_builder.GMT_time);
        rand_class.nextBytes(server_hello_builder.random_component_builder.random_bytes);
    }

    //for each record layer of encrypted application data,
    //returns a linekd list of decrypted plaintext upon success or null upon failure
    public LinkedList<byte[]> handle_Encrypted_Application_Packet(TLS_Packet p) {
        if (this.currentState != TLS_Server_State.STATE_ENCRYPTED_COMMUNICATION) {
            throw new IllegalStateException("Called Decrypt Packet when in Closed or Handshake State");
        }

       LinkedList<byte[]> plaintexts = new LinkedList<>();
        for (TLS_Record_Layer r : p.record_layers) {
            //TODO: Probably here, one would want to handle re-handshaking, but needs more research on when this occurs exactly
            if (r.content_type != TLS_Enum_Content_Type.RECORD_LAYER_CONTENT_TYPE_APPLICATION_DATA.value()) {
                throw new IllegalArgumentException("Expected an encrypted handshake pakcet");
            }

            byte[] ciphertext_raw = r.payload.getRawData();
            byte[] additionalData = generateAdditionalData(TLS_Enum_Content_Type.RECORD_LAYER_CONTENT_TYPE_APPLICATION_DATA.value(), computePlaintextMessageLen(ciphertext_raw), true);

            plaintexts.add(decryptAndVerifyClientBytes(ciphertext_raw, 0, additionalData));
        }
        return plaintexts;
    }

    public byte[] encryptDataAndWrap(byte[] toSend) {
        byte[] additionalData =  generateAdditionalData(TLS_Enum_Content_Type.RECORD_LAYER_CONTENT_TYPE_APPLICATION_DATA.value(), (short) toSend.length, false);
        byte[] cipher_raw = encryptServerBytes(toSend, additionalData);

        TLS_Encrypted_Application_Data.TLS_Encrypted_Application_Data_Builder e_b = new TLS_Encrypted_Application_Data.TLS_Encrypted_Application_Data_Builder(cipher_raw);
        TLS_Record_Layer.TLS_Record_Layer_Builder r_b = new TLS_Record_Layer.TLS_Record_Layer_Builder(TLS_Enum_Content_Type.RECORD_LAYER_CONTENT_TYPE_APPLICATION_DATA.value());
        r_b.payload_builder = (TLS_Builder) e_b;

        TLS_List_Builder<TLS_Record_Layer> list_b = new TLS_List_Builder<>();

        list_b.add(r_b);
        TLS_Packet.TLS_Packet_Builder packet_b = new TLS_Packet.TLS_Packet_Builder();
        packet_b.record_layer_list_builder = list_b;
        packet_b.fixSizeFields();

        try {
            return packet_b.build().getRawData();
        } catch (Exception e) {
            return null;
        }
    }

    public TLS_Server_State getCurrentState() {
        return this.currentState;
    }

    private short computePlaintextMessageLen(byte[] encrypted_payload) {
        return (short) (encrypted_payload.length - this.GCM_IV_EXPLICIT_SIZE - this.MAC_SIZE);
    }

    private byte[] generateAdditionalData(byte contentType, short msg_len, boolean isClient) {
        byte[] additionalData = Arrays.copyOf(this.GENERIC_ADDITIONAL_DATA, this.GENERIC_ADDITIONAL_DATA.length);
        if (isClient) {
            SSL_Helper.revert_to_bytes(this.client_seq_num, additionalData, 0, 8);
        } else {
            SSL_Helper.revert_to_bytes(this.server_seq_num, additionalData, 0, 8);
        }
        additionalData[8] = contentType;
        SSL_Helper.revert_to_bytes(SSL_Helper.castToInt(msg_len), additionalData, additionalData.length-2, 2);
        return additionalData;
    }

    public static byte[] HMAC(byte[] key, byte[] data) {
        //            Mac mac = Mac.getInstance("HmacSHA256"); mac.init(new SecretKeySpec(secret,""));

        HMac hmac = new HMac(new SHA256Digest());
        hmac.init(new KeyParameter(key));
        byte[] result = new byte[hmac.getMacSize()];

        hmac.update(data, 0, data.length);
        hmac.doFinal(result, 0);

        return result;
    }

    public static byte[] HMAC(byte[] key, byte[] data1, byte[] data2) {
        try {
            Mac hmac = Mac.getInstance("HmacSHA256");
            hmac.init(new SecretKeySpec(key, ""));

            hmac.update(data1, 0, data1.length);
            byte[] result = hmac.doFinal(data2);

            return result;
        }catch (Exception e) {
            return null;
        }
    }

    public static byte[] PRF(byte[] secret,  String label, byte[] seed, int len) throws Exception {
        byte[] label_bytes = label.getBytes();


        byte[] P_SHA256_seed = new byte[label_bytes.length + seed.length];
        System.arraycopy(label_bytes, 0, P_SHA256_seed, 0, label_bytes.length);
        System.arraycopy(seed, 0, P_SHA256_seed, label_bytes.length, seed.length);

        return P_SHA256(secret, P_SHA256_seed, len);
     }

    public static byte[] P_SHA256(byte[] secret, byte[] seed, int len) {
        int numRounds = (int) Math.ceil((float)len / 32);  //32 is how many bytes in a SHA256 digest
        byte[] returnBytes = new byte[len];
        int start = 0;
        byte[] A_i = Arrays.copyOf(seed, seed.length);   //A0
        byte[] A_i_plus_seed = new byte[A_i.length + seed.length];   //A(i) + seed
        byte[] A_i_plus_seed_hashed; //HMAC_hash(secret, A(i) + seed)

        //System.arraycopy(seed, 0, A_i_plus_seed, A_i.length, seed.length);
        for (int i = 0; i < numRounds; ++i) {
            A_i = HMAC(secret, A_i);  //A(i) = HMAC_hash(secret, A(i-1))
            System.arraycopy(A_i, 0, A_i_plus_seed, 0, A_i.length); //update A(i) + seed
            //A_i_plus_seed_hashed = HMAC(secret, A_i_plus_seed);
            A_i_plus_seed_hashed = HMAC(secret, A_i, seed); //compute HMAC_hash(secret, A(i) + seed)
            if (i == numRounds - 1) {
                System.arraycopy(A_i_plus_seed_hashed, 0, returnBytes, start, returnBytes.length - start);
                start = returnBytes.length;
            } else {
                System.arraycopy(A_i_plus_seed_hashed, 0, returnBytes, start, A_i_plus_seed_hashed.length);
                start += A_i_plus_seed_hashed.length;
            }
        }

        return returnBytes;
    }

    //Encrypted packet contains the explicit IV, it is as received
    public byte[] decryptAndVerifyClientBytes(byte[] encrypted_packet, int encr_packet_start, byte[] additionalData) {
        try {
            updateExplicitCLientIV(encrypted_packet, encr_packet_start);
            updateClientCipher(additionalData);

            // strip the initialization vector from the message
            byte[] messageToDecrypt = new byte[encrypted_packet.length -  this.GCM_IV_EXPLICIT_SIZE];
            System.arraycopy(encrypted_packet, this.GCM_IV_EXPLICIT_SIZE, messageToDecrypt, 0, messageToDecrypt.length);
            // create a temporary buffer to decode into (it'll include padding)
            this.buf_client = new byte[this.client_cipher.getOutputSize(messageToDecrypt.length)];
            int len_c = this.client_cipher.processBytes(messageToDecrypt, 0, messageToDecrypt.length, buf_client, 0);
            this.client_cipher.doFinal(buf_client, len_c);


            this.client_seq_num += 1;
            return this.buf_client;
        } catch (InvalidCipherTextException e) {
            e.printStackTrace();    //in case of MAC failure
            return null;
        }
    }

    //also includes the epxlicit IV in its return
    public byte[] encryptServerBytes(byte[] toEncrypt, byte[] additionalData_server) {
        try {
            updateServerCipher(additionalData_server);

            byte[] buf_server = new byte[this.GCM_IV_EXPLICIT_SIZE + this.server_cipher.getOutputSize(toEncrypt.length)];
            System.arraycopy(this.server_gcm_iv, this.GCM_IV_IMPLICIT_SIZE, buf_server, 0 , this.GCM_IV_EXPLICIT_SIZE);
            int len_s = this.server_cipher.processBytes(toEncrypt, 0, toEncrypt.length, buf_server, this.GCM_IV_EXPLICIT_SIZE);
            this.server_cipher.doFinal(buf_server, this.GCM_IV_EXPLICIT_SIZE + len_s);

            this.server_seq_num += 1;
            return buf_server;
        } catch (InvalidCipherTextException e) {
            e.printStackTrace();
            e.printStackTrace();    //in case of MAC failure
            return null;
        }
    }

    private void updateExplicitCLientIV(byte[] newExplicit, int start) {
        if (this.client_gcm_iv == null) {
            this.client_gcm_iv = new byte[this.GCM_IV_SIZE];
        }
        System.arraycopy(newExplicit, start, this.client_gcm_iv, this.GCM_IV_IMPLICIT_SIZE, this.GCM_IV_EXPLICIT_SIZE);
    }
    private void updateClientCipher(byte[] additionalData) {
        if (this.client_cipher == null) {
            this.client_cipher = new GCMBlockCipher(new AESEngine());
        }

        // setup cipher parameters with key and IV
        KeyParameter keyParam = new KeyParameter(this.client_aes_key);
        CipherParameters params = new AEADParameters(keyParam, 128, this.client_gcm_iv, additionalData);

        this.client_cipher.reset();
        this.client_cipher.init(false, params);
    }

    //automatically generates a secure explicit IV for the server
    private void updateServerCipher(byte[] additionalData) {
        //automatically generate the explicit part of the IV for server
        SecureRandom rng = new SecureRandom(this.master_secret);
        byte[] explicit_iv = new byte[this.GCM_IV_EXPLICIT_SIZE];
        rng.nextBytes(explicit_iv);
        System.arraycopy(explicit_iv, 0, this.server_gcm_iv, this.GCM_IV_IMPLICIT_SIZE, 8);

        if (this.server_cipher == null) {
            this.server_cipher = new GCMBlockCipher(new AESEngine());
        }

        // setup cipher parameters with key and IV
        KeyParameter keyParam = new KeyParameter(this.server_aes_key);
        CipherParameters params = new AEADParameters(keyParam, 128, this.server_gcm_iv, additionalData);

        this.server_cipher.reset();
        this.server_cipher.init(true, params);
    }

    private void generateEphemeralKeysFromMaster () throws Exception {
        //we have the master key
        //generate the key block
        byte[] randoms_for_key_block = Arrays.copyOf(server_random.s_getRawData(), 2*server_random.s_getTotalSize());
        System.arraycopy(client_random.s_getRawData(), 0, randoms_for_key_block, server_random.s_getTotalSize(), client_random.s_getTotalSize());
        this.key_block = PRF(this.master_secret, "key expansion", randoms_for_key_block, this.KEY_BLOCK_SIZE);

        //Client AES Key
        System.arraycopy(this.key_block, 0, this.client_aes_key, 0, this.AES_KEY_SIZE);
        //Server AES Key
        System.arraycopy(this.key_block, this.AES_KEY_SIZE, this.server_aes_key, 0, this.AES_KEY_SIZE);

        //Implicit client IV
        System.arraycopy(this.key_block, 2*this.AES_KEY_SIZE, this.client_gcm_iv, 0, this.GCM_IV_IMPLICIT_SIZE);
        //Implicit server IV
        System.arraycopy(this.key_block, 2*this.AES_KEY_SIZE + this.GCM_IV_IMPLICIT_SIZE, this.server_gcm_iv, 0, this.GCM_IV_IMPLICIT_SIZE);
    }

    public void do_ECDHE_key_exchange(TLS_Handshake_Client_Key_Exchange client_key_exchange) {
        this.currentState = TLS_Server_State.STATE_RECEIVED_CLIENT_KEY_EXCHANGE;
        KeyAgreement keyAgreement = null;
        try {
            byte[] client_ecdhe_key_bytes = client_key_exchange.client_params.pub_key;

            KeyFactory keyFactory = KeyFactory.getInstance("X25519");
            SubjectPublicKeyInfo subjectPublicKeyInfo = new SubjectPublicKeyInfo(new AlgorithmIdentifier(EdECObjectIdentifiers.id_X25519), client_ecdhe_key_bytes);
            X509EncodedKeySpec x509EncodedKeySpec = new X509EncodedKeySpec(subjectPublicKeyInfo.getEncoded());
            this.clientDHEPublicKey = keyFactory.generatePublic(x509EncodedKeySpec);

            keyAgreement = KeyAgreement.getInstance("X25519", BouncyCastleProvider.PROVIDER_NAME);
            keyAgreement.init( this.serverECDHEKeyPair.getPrivate());
            keyAgreement.doPhase(this.clientDHEPublicKey, true);

            byte[] pre_master = keyAgreement.generateSecret();    //pre master secret

            byte[] randoms = new byte[64];
            System.arraycopy(this.client_random.s_getRawData(), 0, randoms, 0, 32);
            System.arraycopy(this.server_random.s_getRawData(), 0, randoms, 32, 32);

            this.master_secret = PRF(pre_master, "master secret", randoms, this.MASTER_SECRET_LEN);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    //TODO: Actually generate the fake certificate
    public void generate_fake_certificate(TLS_Handshake_Extension_SERVER_NAME server_name) {
        try {
            LinkedList<byte[]> cert_chain = new LinkedList<>();
            String server_name_str = "myserverdomain.com";

            if (server_name != null) {
                server_name_str = new String(server_name.list.get(0).name_bytes);
            }
            KeyPairGenerator keyGen = KeyPairGenerator.getInstance("RSA");
            keyGen.initialize(2048);
            serverRSAKeyPair = keyGen.generateKeyPair();

            SSL_Certificate_Manager m = SSL_Certificate_Manager.getInstance();
            X509Certificate cert = m.createAndSignCertificate(server_name_str, serverRSAKeyPair);

            byte[] derCert = cert.getEncoded();

            cert_chain.add(derCert);
            //cert_chain.add(server_cert_bytes);
            //cert_chain.add(SSL_Certificate_Manager.ca_cert_raw_bytes);

            TLS_Handshake_Certificate.TLS_Handshake_Certificate_Builder certificate_builder = new TLS_Handshake_Certificate.TLS_Handshake_Certificate_Builder(cert_chain);

            ((TLS_Record_Layer.TLS_Record_Layer_Builder) (tlsServerHelloTemplateBuilder.record_layer_list_builder.get(1))).payload_builder = (TLS_Builder) certificate_builder;

        } catch (Exception e) {
            e.printStackTrace();
        }
    }
    public void generate_ecdhe_key_and_sign_server_key_exchange() {
        try {
            TLS_Builder server_hello_builder = ((TLS_Record_Layer.TLS_Record_Layer_Builder)(tlsServerHelloTemplateBuilder.record_layer_list_builder.get(0))).payload_builder;
            TLS_Handshake_Random_Component_Subclass.TLS_Handshake_Random_Component_Subclass_Builder server_random_builder =  new TLS_Handshake_Random_Component_Subclass.TLS_Handshake_Random_Component_Subclass_Builder(((TLS_Handshake_Server_Hello.TLS_Handshake_Server_Hello_Builder)server_hello_builder).random_component_builder.build());
            server_random = server_random_builder.build();

            //basically we only want to tell that the point formats are uncompressed, no other extensions to complicate things

            ((TLS_Handshake_Server_Hello.TLS_Handshake_Server_Hello_Builder) server_hello_builder).extensions = new TLS_List_Builder<TLS_Handshake_Extension_Abstract>();
            byte[] ec_point_formats_raw = new byte[] {(byte)0x00, (byte)0x0b, (byte)0x00, (byte)0x02, (byte)0x01, (byte)0x00};
            TLS_Handshake_Extension_EC_POINT_FORMATS.TLS_Handshake_Extension_EC_POINT_FORMATS_Builder ec_point_formats_builder = new TLS_Handshake_Extension_EC_POINT_FORMATS.TLS_Handshake_Extension_EC_POINT_FORMATS_Builder(ec_point_formats_raw, 0);
            ((TLS_Handshake_Server_Hello.TLS_Handshake_Server_Hello_Builder) server_hello_builder).extensions.add(ec_point_formats_builder);
            byte[] renegotiation_info_raw = new byte[] {(byte) 0xff, (byte)0x01,(byte) 0x00,(byte) 0x01, (byte)0x00};
            //secure renegotiation
            TLS_Handshake_Extension_Unknown.TLS_Handshake_Extension_Unknown_Builder extension_renegotiatiation_builder = new TLS_Handshake_Extension_Unknown.TLS_Handshake_Extension_Unknown_Builder(renegotiation_info_raw, 0);
            ((TLS_Handshake_Server_Hello.TLS_Handshake_Server_Hello_Builder) server_hello_builder).extensions.add(extension_renegotiatiation_builder);

            //extended master secret
//            byte[] extended_master_raw = new byte[] {(byte)0x00, (byte)0x17, (byte)0x00, (byte)0x00};
//            TLS_Handshake_Extension_Unknown.TLS_Handshake_Extension_Unknown_Builder extension_extended_master_builder = new TLS_Handshake_Extension_Unknown.TLS_Handshake_Extension_Unknown_Builder(extended_master_raw, 0);
//            ((TLS_Handshake_Server_Hello.TLS_Handshake_Server_Hello_Builder) server_hello_builder).extensions.add(extension_extended_master_builder);


            TLS_Builder server_key_exchange_builder = ((TLS_Record_Layer.TLS_Record_Layer_Builder)(tlsServerHelloTemplateBuilder.record_layer_list_builder.get(2))).payload_builder;
            TLS_Handshake_Server_Key_Exchange.ECDHE_Server_Params_Builder ecdhe_server_params_builder = new TLS_Handshake_Server_Key_Exchange.ECDHE_Server_Params_Builder(((TLS_Handshake_Server_Key_Exchange.TLS_Handshake_Server_Key_Exchange_Builder)server_key_exchange_builder).server_params.build());

            // Generate an ECDHE key pair
            KeyPairGenerator keyPairGenerator = KeyPairGenerator.getInstance("X25519", BouncyCastleProvider.PROVIDER_NAME);
            keyPairGenerator.initialize(new XDHParameterSpec(XDHParameterSpec.X25519));
            this.serverECDHEKeyPair = keyPairGenerator.generateKeyPair();

            byte[] public_key_bytes = serverECDHEKeyPair.getPublic().getEncoded();
            System.arraycopy(public_key_bytes, 12, ecdhe_server_params_builder.pub_key, 0, ecdhe_server_params_builder.pub_key.length);

            ecdhe_server_params = ecdhe_server_params_builder.build();

            byte[] client_random_bytes = client_random.s_getRawData();
            byte[] server_random_bytes = server_random.s_getRawData();
            byte[] ecdhe_server_params_bytes_for_signature = ecdhe_server_params.s_get_bytes_for_signature();

            // Generate the bytes which need to be signed in server key exchange
            byte[] bytes_to_sign = new byte[client_random_bytes.length + server_random_bytes.length + ecdhe_server_params_bytes_for_signature.length];
            System.arraycopy(client_random_bytes, 0, bytes_to_sign, 0, client_random_bytes.length);
            System.arraycopy(server_random_bytes, 0, bytes_to_sign, client_random_bytes.length, server_random_bytes.length);
            System.arraycopy(ecdhe_server_params_bytes_for_signature, 0, bytes_to_sign,  client_random_bytes.length + server_random_bytes.length, ecdhe_server_params_bytes_for_signature.length);

            // Sign the bytes
            Signature signatureSHA256Java = Signature.getInstance("SHA256withRSA/PSS");
            signatureSHA256Java.setParameter(new PSSParameterSpec("SHA-256", "MGF1", MGF1ParameterSpec.SHA256, 32, 1));

            signatureSHA256Java.initSign(serverRSAKeyPair.getPrivate());
            signatureSHA256Java.update(bytes_to_sign);
            byte[] signSHA256Java = signatureSHA256Java.sign();

            System.arraycopy(signSHA256Java, 0, ecdhe_server_params_builder.signature, 0, signSHA256Java.length);
            ((TLS_Handshake_Server_Key_Exchange.TLS_Handshake_Server_Key_Exchange_Builder) ((TLS_Builder) (((TLS_Record_Layer.TLS_Record_Layer_Builder) (tlsServerHelloTemplateBuilder.record_layer_list_builder.get(2))).payload_builder))).server_params = ecdhe_server_params_builder;
        }catch (Exception e) {
            e.printStackTrace();
        }
    }

    public RSAPublicKey readPublicKey(File file) throws Exception {
        KeyFactory factory = KeyFactory.getInstance("RSA");

        try (FileReader keyReader = new FileReader(file);
             PemReader pemReader = new PemReader(keyReader)) {

            PemObject pemObject = pemReader.readPemObject();
            byte[] content = pemObject.getContent();
            X509EncodedKeySpec pubKeySpec = new X509EncodedKeySpec(content);
            return (RSAPublicKey) factory.generatePublic(pubKeySpec);
        }
    }

    public RSAPrivateKey readPrivateKey(File file) throws Exception {
        KeyFactory factory = KeyFactory.getInstance("RSA");

        try (FileReader keyReader = new FileReader(file);
             PemReader pemReader = new PemReader(keyReader)) {

            PemObject pemObject = pemReader.readPemObject();
            byte[] content = pemObject.getContent();
            PKCS8EncodedKeySpec privKeySpec = new PKCS8EncodedKeySpec(content);
            return (RSAPrivateKey) factory.generatePrivate(privKeySpec);
        }
    }

    final public static byte[] close_alert = new byte[] {
        (byte)0x15, (byte)0x03, (byte)0x03, (byte)0x00, (byte)0x02, (byte)0x02, (byte)0x00
    };

    final byte[] server_cert_bytes = new byte[] {
            (byte)0x30, (byte)0x82, (byte)0x03, (byte)0x3b, (byte)0x30, (byte)0x82, (byte)0x02, (byte)0x23, (byte)0xa0, (byte)0x03, (byte)0x02, (byte)0x01, (byte)0x02, (byte)0x02, (byte)0x01, (byte)0x01,
            (byte)0x30, (byte)0x0d, (byte)0x06, (byte)0x09, (byte)0x2a, (byte)0x86, (byte)0x48, (byte)0x86, (byte)0xf7, (byte)0x0d, (byte)0x01, (byte)0x01, (byte)0x0b, (byte)0x05, (byte)0x00, (byte)0x30,
            (byte)0x19, (byte)0x31, (byte)0x17, (byte)0x30, (byte)0x15, (byte)0x06, (byte)0x03, (byte)0x55, (byte)0x04, (byte)0x03, (byte)0x0c, (byte)0x0e, (byte)0x6d, (byte)0x79, (byte)0x63, (byte)0x61,
            (byte)0x64, (byte)0x6f, (byte)0x6d, (byte)0x61, (byte)0x69, (byte)0x6e, (byte)0x2e, (byte)0x63, (byte)0x6f, (byte)0x6d, (byte)0x30, (byte)0x1e, (byte)0x17, (byte)0x0d, (byte)0x32, (byte)0x33,
            (byte)0x30, (byte)0x32, (byte)0x31, (byte)0x33, (byte)0x31, (byte)0x37, (byte)0x31, (byte)0x30, (byte)0x34, (byte)0x39, (byte)0x5a, (byte)0x17, (byte)0x0d, (byte)0x32, (byte)0x34, (byte)0x30,
            (byte)0x32, (byte)0x31, (byte)0x33, (byte)0x31, (byte)0x37, (byte)0x31, (byte)0x30, (byte)0x34, (byte)0x39, (byte)0x5a, (byte)0x30, (byte)0x1d, (byte)0x31, (byte)0x1b, (byte)0x30, (byte)0x19,
            (byte)0x06, (byte)0x03, (byte)0x55, (byte)0x04, (byte)0x03, (byte)0x0c, (byte)0x12, (byte)0x6d, (byte)0x79, (byte)0x73, (byte)0x65, (byte)0x72, (byte)0x76, (byte)0x65, (byte)0x72, (byte)0x64,
            (byte)0x6f, (byte)0x6d, (byte)0x61, (byte)0x69, (byte)0x6e, (byte)0x2e, (byte)0x63, (byte)0x6f, (byte)0x6d, (byte)0x30, (byte)0x82, (byte)0x01, (byte)0x22, (byte)0x30, (byte)0x0d, (byte)0x06,
            (byte)0x09, (byte)0x2a, (byte)0x86, (byte)0x48, (byte)0x86, (byte)0xf7, (byte)0x0d, (byte)0x01, (byte)0x01, (byte)0x01, (byte)0x05, (byte)0x00, (byte)0x03, (byte)0x82, (byte)0x01, (byte)0x0f,
            (byte)0x00, (byte)0x30, (byte)0x82, (byte)0x01, (byte)0x0a, (byte)0x02, (byte)0x82, (byte)0x01, (byte)0x01, (byte)0x00, (byte)0xbd, (byte)0x67, (byte)0xde, (byte)0xf8, (byte)0x7f, (byte)0xe1,
            (byte)0xc8, (byte)0x25, (byte)0x26, (byte)0x12, (byte)0x0c, (byte)0x80, (byte)0x6e, (byte)0xe9, (byte)0x64, (byte)0x5a, (byte)0x1f, (byte)0x82, (byte)0xd2, (byte)0x99, (byte)0x0a, (byte)0x62,
            (byte)0xa2, (byte)0x7a, (byte)0x9f, (byte)0x8a, (byte)0x91, (byte)0x8d, (byte)0x6e, (byte)0x91, (byte)0x4f, (byte)0xfc, (byte)0x9a, (byte)0x2e, (byte)0x11, (byte)0x26, (byte)0xbc, (byte)0xb9,
            (byte)0x58, (byte)0x18, (byte)0xb7, (byte)0xef, (byte)0x8d, (byte)0x92, (byte)0x7d, (byte)0xdd, (byte)0xef, (byte)0x70, (byte)0xdc, (byte)0x7f, (byte)0x06, (byte)0x3a, (byte)0x3e, (byte)0xe9,
            (byte)0x0b, (byte)0xa0, (byte)0xcd, (byte)0xe9, (byte)0x96, (byte)0x65, (byte)0xfe, (byte)0xb1, (byte)0x90, (byte)0xe9, (byte)0x35, (byte)0x2b, (byte)0xb7, (byte)0x13, (byte)0xd7, (byte)0xe9,
            (byte)0x10, (byte)0x5e, (byte)0xf6, (byte)0x84, (byte)0x8d, (byte)0x42, (byte)0x1a, (byte)0x82, (byte)0x45, (byte)0x99, (byte)0x4c, (byte)0xaa, (byte)0xfb, (byte)0x86, (byte)0x73, (byte)0x41,
            (byte)0xf1, (byte)0x02, (byte)0xae, (byte)0x08, (byte)0x13, (byte)0x39, (byte)0x82, (byte)0x03, (byte)0x53, (byte)0xc4, (byte)0x81, (byte)0x43, (byte)0x18, (byte)0xf4, (byte)0xff, (byte)0x1b,
            (byte)0x9c, (byte)0xb8, (byte)0x3e, (byte)0xda, (byte)0x28, (byte)0x12, (byte)0xf0, (byte)0x88, (byte)0xb7, (byte)0xaf, (byte)0x52, (byte)0x34, (byte)0x3a, (byte)0x7f, (byte)0xa9, (byte)0x1a,
            (byte)0xf5, (byte)0x13, (byte)0x4c, (byte)0xd5, (byte)0x14, (byte)0x43, (byte)0x21, (byte)0x5a, (byte)0xcc, (byte)0x97, (byte)0xc1, (byte)0x9e, (byte)0xcf, (byte)0x2d, (byte)0x16, (byte)0x86,
            (byte)0xcb, (byte)0x5b, (byte)0x1d, (byte)0x62, (byte)0xd9, (byte)0xea, (byte)0x3e, (byte)0xf1, (byte)0xfc, (byte)0x47, (byte)0x7e, (byte)0xb5, (byte)0x08, (byte)0xf4, (byte)0x23, (byte)0xcb,
            (byte)0x5e, (byte)0x1a, (byte)0xe7, (byte)0x51, (byte)0x52, (byte)0x2c, (byte)0x4d, (byte)0xbf, (byte)0xa9, (byte)0x79, (byte)0x62, (byte)0x0d, (byte)0xf8, (byte)0xe2, (byte)0x64, (byte)0x1f,
            (byte)0xd5, (byte)0x22, (byte)0x7f, (byte)0xc9, (byte)0x38, (byte)0xac, (byte)0x83, (byte)0x2c, (byte)0xe2, (byte)0xb7, (byte)0x7b, (byte)0x92, (byte)0x75, (byte)0x31, (byte)0x11, (byte)0xb3,
            (byte)0x8b, (byte)0xea, (byte)0x47, (byte)0xd7, (byte)0x4f, (byte)0x00, (byte)0xd6, (byte)0x6b, (byte)0xd4, (byte)0x9f, (byte)0x3f, (byte)0x64, (byte)0x76, (byte)0x7a, (byte)0xbc, (byte)0x1b,
            (byte)0x84, (byte)0xe1, (byte)0x5f, (byte)0xa6, (byte)0x33, (byte)0xad, (byte)0x1d, (byte)0x20, (byte)0xa2, (byte)0x8e, (byte)0xa1, (byte)0xac, (byte)0x66, (byte)0xb5, (byte)0x29, (byte)0x71,
            (byte)0x76, (byte)0xcd, (byte)0x1f, (byte)0xc9, (byte)0x87, (byte)0x40, (byte)0x33, (byte)0xca, (byte)0x57, (byte)0xb9, (byte)0x64, (byte)0xa7, (byte)0x4c, (byte)0xe7, (byte)0x06, (byte)0x13,
            (byte)0x2c, (byte)0x30, (byte)0x83, (byte)0xc7, (byte)0xf8, (byte)0xb2, (byte)0xe9, (byte)0x81, (byte)0x35, (byte)0x3e, (byte)0x50, (byte)0x3f, (byte)0x8b, (byte)0xcb, (byte)0x6e, (byte)0x83,
            (byte)0x07, (byte)0x39, (byte)0x5c, (byte)0x18, (byte)0x60, (byte)0xdf, (byte)0x83, (byte)0x9d, (byte)0xac, (byte)0xdb, (byte)0x02, (byte)0x03, (byte)0x01, (byte)0x00, (byte)0x01, (byte)0xa3,
            (byte)0x81, (byte)0x89, (byte)0x30, (byte)0x81, (byte)0x86, (byte)0x30, (byte)0x1d, (byte)0x06, (byte)0x03, (byte)0x55, (byte)0x1d, (byte)0x0e, (byte)0x04, (byte)0x16, (byte)0x04, (byte)0x14,
            (byte)0x80, (byte)0xc5, (byte)0x8e, (byte)0x97, (byte)0xcb, (byte)0x6d, (byte)0xb2, (byte)0x23, (byte)0x25, (byte)0x84, (byte)0xa2, (byte)0x55, (byte)0x77, (byte)0xb2, (byte)0x3a, (byte)0x86,
            (byte)0x08, (byte)0x4e, (byte)0x22, (byte)0x10, (byte)0x30, (byte)0x1f, (byte)0x06, (byte)0x03, (byte)0x55, (byte)0x1d, (byte)0x23, (byte)0x04, (byte)0x18, (byte)0x30, (byte)0x16, (byte)0x80,
            (byte)0x14, (byte)0x5a, (byte)0x4f, (byte)0xd7, (byte)0x6a, (byte)0x3c, (byte)0x84, (byte)0xda, (byte)0x79, (byte)0x53, (byte)0x4d, (byte)0x27, (byte)0xa5, (byte)0x56, (byte)0x83, (byte)0x63,
            (byte)0x4f, (byte)0x4e, (byte)0x1a, (byte)0xad, (byte)0x6f, (byte)0x30, (byte)0x09, (byte)0x06, (byte)0x03, (byte)0x55, (byte)0x1d, (byte)0x13, (byte)0x04, (byte)0x02, (byte)0x30, (byte)0x00,
            (byte)0x30, (byte)0x0b, (byte)0x06, (byte)0x03, (byte)0x55, (byte)0x1d, (byte)0x0f, (byte)0x04, (byte)0x04, (byte)0x03, (byte)0x02, (byte)0x05, (byte)0xa0, (byte)0x30, (byte)0x2c, (byte)0x06,
            (byte)0x09, (byte)0x60, (byte)0x86, (byte)0x48, (byte)0x01, (byte)0x86, (byte)0xf8, (byte)0x42, (byte)0x01, (byte)0x0d, (byte)0x04, (byte)0x1f, (byte)0x16, (byte)0x1d, (byte)0x4f, (byte)0x70,
            (byte)0x65, (byte)0x6e, (byte)0x53, (byte)0x53, (byte)0x4c, (byte)0x20, (byte)0x47, (byte)0x65, (byte)0x6e, (byte)0x65, (byte)0x72, (byte)0x61, (byte)0x74, (byte)0x65, (byte)0x64, (byte)0x20,
            (byte)0x43, (byte)0x65, (byte)0x72, (byte)0x74, (byte)0x69, (byte)0x66, (byte)0x69, (byte)0x63, (byte)0x61, (byte)0x74, (byte)0x65, (byte)0x30, (byte)0x0d, (byte)0x06, (byte)0x09, (byte)0x2a,
            (byte)0x86, (byte)0x48, (byte)0x86, (byte)0xf7, (byte)0x0d, (byte)0x01, (byte)0x01, (byte)0x0b, (byte)0x05, (byte)0x00, (byte)0x03, (byte)0x82, (byte)0x01, (byte)0x01, (byte)0x00, (byte)0x6a,
            (byte)0x07, (byte)0x31, (byte)0x8f, (byte)0x00, (byte)0x42, (byte)0x36, (byte)0xd3, (byte)0x5a, (byte)0xb7, (byte)0xab, (byte)0xbc, (byte)0x89, (byte)0x5f, (byte)0x53, (byte)0x5e, (byte)0xd4,
            (byte)0x87, (byte)0xf7, (byte)0xc4, (byte)0xd0, (byte)0x3d, (byte)0x8a, (byte)0x1e, (byte)0x54, (byte)0xb5, (byte)0x13, (byte)0xc5, (byte)0x31, (byte)0xdf, (byte)0x04, (byte)0xac, (byte)0xbb,
            (byte)0x2e, (byte)0x73, (byte)0x9d, (byte)0xfe, (byte)0x26, (byte)0x4d, (byte)0xdc, (byte)0xb3, (byte)0x58, (byte)0x1a, (byte)0x33, (byte)0xe9, (byte)0x45, (byte)0xad, (byte)0x3a, (byte)0x16,
            (byte)0x4f, (byte)0xdc, (byte)0x3e, (byte)0x95, (byte)0x4c, (byte)0x94, (byte)0xdf, (byte)0x92, (byte)0xa6, (byte)0x39, (byte)0x45, (byte)0xdd, (byte)0x02, (byte)0xf6, (byte)0x70, (byte)0x80,
            (byte)0xb9, (byte)0x1e, (byte)0xd0, (byte)0x2f, (byte)0xd5, (byte)0x66, (byte)0x9c, (byte)0x84, (byte)0x17, (byte)0xdd, (byte)0x1a, (byte)0x53, (byte)0xb8, (byte)0x78, (byte)0x18, (byte)0x13,
            (byte)0x58, (byte)0x4e, (byte)0x61, (byte)0x04, (byte)0x72, (byte)0xa2, (byte)0x89, (byte)0xe0, (byte)0xad, (byte)0x40, (byte)0x96, (byte)0x0e, (byte)0x27, (byte)0x04, (byte)0x7c, (byte)0xb4,
            (byte)0x73, (byte)0xe4, (byte)0xb9, (byte)0x31, (byte)0xc5, (byte)0xb5, (byte)0xea, (byte)0x11, (byte)0xe1, (byte)0xb6, (byte)0x77, (byte)0x8b, (byte)0xe4, (byte)0x8a, (byte)0x2e, (byte)0x24,
            (byte)0xcb, (byte)0x37, (byte)0x00, (byte)0x28, (byte)0x58, (byte)0xff, (byte)0x40, (byte)0x44, (byte)0x6b, (byte)0x0b, (byte)0xad, (byte)0x85, (byte)0x63, (byte)0x2c, (byte)0x38, (byte)0x60,
            (byte)0x8d, (byte)0xf0, (byte)0xaf, (byte)0xc0, (byte)0x9c, (byte)0x87, (byte)0x61, (byte)0xa0, (byte)0x6f, (byte)0xe5, (byte)0xc3, (byte)0xfd, (byte)0x15, (byte)0x89, (byte)0xbc, (byte)0xf6,
            (byte)0x60, (byte)0x56, (byte)0x67, (byte)0x82, (byte)0x77, (byte)0x63, (byte)0x7a, (byte)0xb6, (byte)0x9a, (byte)0x80, (byte)0xa0, (byte)0x46, (byte)0x09, (byte)0x75, (byte)0xca, (byte)0x80,
            (byte)0xa1, (byte)0x06, (byte)0xbc, (byte)0x23, (byte)0x79, (byte)0xb0, (byte)0xa3, (byte)0x3f, (byte)0x34, (byte)0x57, (byte)0xca, (byte)0xed, (byte)0x8b, (byte)0x7f, (byte)0xe3, (byte)0xe3,
            (byte)0x86, (byte)0xfd, (byte)0x87, (byte)0x4c, (byte)0xea, (byte)0x1d, (byte)0x39, (byte)0x62, (byte)0x21, (byte)0x6b, (byte)0xcd, (byte)0x79, (byte)0x23, (byte)0xa6, (byte)0x5d, (byte)0x6a,
            (byte)0x1e, (byte)0x54, (byte)0xbd, (byte)0x1c, (byte)0xfc, (byte)0xba, (byte)0xbc, (byte)0xbc, (byte)0xc2, (byte)0xa0, (byte)0xbe, (byte)0xb1, (byte)0x9b, (byte)0x3e, (byte)0x46, (byte)0x64,
            (byte)0x2c, (byte)0x05, (byte)0x73, (byte)0x23, (byte)0xe9, (byte)0x94, (byte)0x8e, (byte)0x9f, (byte)0x1c, (byte)0x40, (byte)0x25, (byte)0x60, (byte)0xce, (byte)0x51, (byte)0x99, (byte)0x72,
            (byte)0x5a, (byte)0x65, (byte)0x80, (byte)0x55, (byte)0xb5, (byte)0xb2, (byte)0xec, (byte)0xdd, (byte)0xb5, (byte)0x3c, (byte)0x4a, (byte)0x54, (byte)0x80, (byte)0xe0, (byte)0x7f, (byte)0x0b,
            (byte)0x4e, (byte)0x07, (byte)0x99, (byte)0x6c, (byte)0x1e, (byte)0x58, (byte)0x63, (byte)0x91, (byte)0x3c, (byte)0x3e, (byte)0xdb, (byte)0x2c, (byte)0x3f, (byte)0xff, (byte)0x7e
    };



    final byte[] server_hello_template_bytes = new byte[] {
            (byte)0x16, (byte)0x03, (byte)0x03, (byte)0x00, (byte)0x5d, (byte)0x02, (byte)0x00, (byte)0x00, (byte)0x59, (byte)0x03, (byte)0x03, (byte)0x43, (byte)0x57, (byte)0x8a, (byte)0x24, (byte)0x3b,
            (byte)0x66, (byte)0x8a, (byte)0x45, (byte)0x91, (byte)0x3f, (byte)0x8c, (byte)0x9e, (byte)0xdd, (byte)0xe6, (byte)0x38, (byte)0x66, (byte)0x3b, (byte)0x5b, (byte)0xc1, (byte)0xff, (byte)0x4d,
            (byte)0x2f, (byte)0x81, (byte)0x09, (byte)0x61, (byte)0x1e, (byte)0x30, (byte)0x07, (byte)0x5d, (byte)0x6d, (byte)0x6b, (byte)0xef, (byte)0x20, (byte)0xa4, (byte)0x38, (byte)0x9c, (byte)0x60,
            (byte)0x10, (byte)0x3f, (byte)0xcd, (byte)0xa0, (byte)0xbf, (byte)0xd4, (byte)0xb4, (byte)0xd7, (byte)0x90, (byte)0x6d, (byte)0x91, (byte)0x21, (byte)0xad, (byte)0x10, (byte)0xc8, (byte)0xac,
            (byte)0x7b, (byte)0xdd, (byte)0xd6, (byte)0xb8, (byte)0x72, (byte)0x06, (byte)0x22, (byte)0x85, (byte)0xa9, (byte)0x34, (byte)0x50, (byte)0xc9, (byte)0xc0, (byte)0x2f, (byte)0x00, (byte)0x00,
            (byte)0x11, (byte)0xff, (byte)0x01, (byte)0x00, (byte)0x01, (byte)0x00, (byte)0x00, (byte)0x0b, (byte)0x00, (byte)0x04, (byte)0x03, (byte)0x00, (byte)0x01, (byte)0x02, (byte)0x00, (byte)0x17,
            (byte)0x00, (byte)0x00, (byte)0x16, (byte)0x03, (byte)0x03, (byte)0x03, (byte)0x2d, (byte)0x0b, (byte)0x00, (byte)0x03, (byte)0x29, (byte)0x00, (byte)0x03, (byte)0x26, (byte)0x00, (byte)0x03,
            (byte)0x23, (byte)0x30, (byte)0x82, (byte)0x03, (byte)0x1f, (byte)0x30, (byte)0x82, (byte)0x02, (byte)0x07, (byte)0xa0, (byte)0x03, (byte)0x02, (byte)0x01, (byte)0x02, (byte)0x02, (byte)0x13,
            (byte)0x18, (byte)0x04, (byte)0x15, (byte)0xd6, (byte)0xef, (byte)0xa3, (byte)0x18, (byte)0x07, (byte)0xf8, (byte)0x66, (byte)0x76, (byte)0x34, (byte)0xb4, (byte)0x9b, (byte)0x42, (byte)0x6a,
            (byte)0x94, (byte)0xea, (byte)0xac, (byte)0x30, (byte)0x0d, (byte)0x06, (byte)0x09, (byte)0x2a, (byte)0x86, (byte)0x48, (byte)0x86, (byte)0xf7, (byte)0x0d, (byte)0x01, (byte)0x01, (byte)0x0b,
            (byte)0x05, (byte)0x00, (byte)0x30, (byte)0x19, (byte)0x31, (byte)0x17, (byte)0x30, (byte)0x15, (byte)0x06, (byte)0x03, (byte)0x55, (byte)0x04, (byte)0x03, (byte)0x0c, (byte)0x0e, (byte)0x6d,
            (byte)0x79, (byte)0x63, (byte)0x61, (byte)0x64, (byte)0x6f, (byte)0x6d, (byte)0x61, (byte)0x69, (byte)0x6e, (byte)0x2e, (byte)0x63, (byte)0x6f, (byte)0x6d, (byte)0x30, (byte)0x1e, (byte)0x17,
            (byte)0x0d, (byte)0x32, (byte)0x33, (byte)0x30, (byte)0x32, (byte)0x31, (byte)0x33, (byte)0x31, (byte)0x37, (byte)0x30, (byte)0x36, (byte)0x31, (byte)0x39, (byte)0x5a, (byte)0x17, (byte)0x0d,
            (byte)0x32, (byte)0x34, (byte)0x30, (byte)0x32, (byte)0x31, (byte)0x33, (byte)0x31, (byte)0x37, (byte)0x30, (byte)0x36, (byte)0x31, (byte)0x39, (byte)0x5a, (byte)0x30, (byte)0x19, (byte)0x31,
            (byte)0x17, (byte)0x30, (byte)0x15, (byte)0x06, (byte)0x03, (byte)0x55, (byte)0x04, (byte)0x03, (byte)0x0c, (byte)0x0e, (byte)0x6d, (byte)0x79, (byte)0x63, (byte)0x61, (byte)0x64, (byte)0x6f,
            (byte)0x6d, (byte)0x61, (byte)0x69, (byte)0x6e, (byte)0x2e, (byte)0x63, (byte)0x6f, (byte)0x6d, (byte)0x30, (byte)0x82, (byte)0x01, (byte)0x22, (byte)0x30, (byte)0x0d, (byte)0x06, (byte)0x09,
            (byte)0x2a, (byte)0x86, (byte)0x48, (byte)0x86, (byte)0xf7, (byte)0x0d, (byte)0x01, (byte)0x01, (byte)0x01, (byte)0x05, (byte)0x00, (byte)0x03, (byte)0x82, (byte)0x01, (byte)0x0f, (byte)0x00,
            (byte)0x30, (byte)0x82, (byte)0x01, (byte)0x0a, (byte)0x02, (byte)0x82, (byte)0x01, (byte)0x01, (byte)0x00, (byte)0xd1, (byte)0x45, (byte)0x45, (byte)0xdb, (byte)0x8c, (byte)0x19, (byte)0x07,
            (byte)0x05, (byte)0xf5, (byte)0x71, (byte)0xe3, (byte)0x8f, (byte)0x1c, (byte)0xfa, (byte)0x27, (byte)0x70, (byte)0x77, (byte)0xab, (byte)0x05, (byte)0x64, (byte)0x67, (byte)0x4a, (byte)0x5f,
            (byte)0xaa, (byte)0x20, (byte)0xd0, (byte)0xae, (byte)0xc4, (byte)0x41, (byte)0xb1, (byte)0xa5, (byte)0xd6, (byte)0xf3, (byte)0x90, (byte)0x35, (byte)0x1b, (byte)0xd7, (byte)0x1d, (byte)0x6f,
            (byte)0x5d, (byte)0x10, (byte)0xe2, (byte)0xbd, (byte)0x9f, (byte)0x83, (byte)0x06, (byte)0x96, (byte)0x71, (byte)0x67, (byte)0xff, (byte)0xbf, (byte)0x0f, (byte)0x9a, (byte)0x2f, (byte)0x5c,
            (byte)0x95, (byte)0x19, (byte)0x28, (byte)0x95, (byte)0xc0, (byte)0xe0, (byte)0x9b, (byte)0x59, (byte)0x72, (byte)0xb1, (byte)0x1a, (byte)0x00, (byte)0xd9, (byte)0x96, (byte)0x85, (byte)0x53,
            (byte)0x82, (byte)0x6b, (byte)0x4b, (byte)0xb7, (byte)0x23, (byte)0x8d, (byte)0x8f, (byte)0xd5, (byte)0x08, (byte)0xc2, (byte)0xff, (byte)0x17, (byte)0x25, (byte)0xc9, (byte)0x6e, (byte)0x0a,
            (byte)0x2a, (byte)0xca, (byte)0x18, (byte)0x87, (byte)0x8b, (byte)0x78, (byte)0xf9, (byte)0x8c, (byte)0x89, (byte)0x05, (byte)0x97, (byte)0x5c, (byte)0x5d, (byte)0xdf, (byte)0xd4, (byte)0xe6,
            (byte)0x88, (byte)0x3b, (byte)0xae, (byte)0xac, (byte)0xd6, (byte)0x7b, (byte)0x69, (byte)0x7a, (byte)0xc9, (byte)0x85, (byte)0x59, (byte)0x94, (byte)0xf1, (byte)0x08, (byte)0xcb, (byte)0xb0,
            (byte)0xb3, (byte)0xa7, (byte)0xfe, (byte)0x8b, (byte)0x96, (byte)0x35, (byte)0x0e, (byte)0xdb, (byte)0xa3, (byte)0x0a, (byte)0x71, (byte)0x70, (byte)0x15, (byte)0x63, (byte)0xd9, (byte)0xa5,
            (byte)0xf8, (byte)0xd5, (byte)0x4e, (byte)0x47, (byte)0xd9, (byte)0xb6, (byte)0x8b, (byte)0xcf, (byte)0x56, (byte)0x82, (byte)0x8a, (byte)0xee, (byte)0x50, (byte)0x45, (byte)0x30, (byte)0x09,
            (byte)0x26, (byte)0x37, (byte)0xda, (byte)0x73, (byte)0x37, (byte)0xf0, (byte)0xd4, (byte)0x3f, (byte)0x26, (byte)0x83, (byte)0x45, (byte)0xf5, (byte)0x7d, (byte)0xbe, (byte)0xe5, (byte)0x2f,
            (byte)0xbe, (byte)0x6b, (byte)0x60, (byte)0x8a, (byte)0x8e, (byte)0x0b, (byte)0x0e, (byte)0xe2, (byte)0x82, (byte)0x37, (byte)0x2e, (byte)0x3c, (byte)0x22, (byte)0x67, (byte)0x49, (byte)0x63,
            (byte)0x03, (byte)0x0b, (byte)0xda, (byte)0xf5, (byte)0x2d, (byte)0x78, (byte)0xbc, (byte)0x4c, (byte)0x8c, (byte)0x3c, (byte)0x3a, (byte)0xdd, (byte)0xf0, (byte)0x42, (byte)0x6b, (byte)0x78,
            (byte)0x87, (byte)0xdc, (byte)0xda, (byte)0xd8, (byte)0xb4, (byte)0x01, (byte)0xec, (byte)0x5d, (byte)0x5a, (byte)0x88, (byte)0x2f, (byte)0x82, (byte)0x02, (byte)0xfb, (byte)0x0f, (byte)0x21,
            (byte)0xe5, (byte)0xfd, (byte)0x36, (byte)0x10, (byte)0xf1, (byte)0x16, (byte)0xe9, (byte)0xee, (byte)0x6f, (byte)0x35, (byte)0xca, (byte)0x4e, (byte)0xc8, (byte)0xdf, (byte)0xa7, (byte)0x3b,
            (byte)0x24, (byte)0x45, (byte)0xe1, (byte)0xdf, (byte)0x0c, (byte)0xab, (byte)0x8b, (byte)0x26, (byte)0x94, (byte)0xc8, (byte)0x0a, (byte)0x6f, (byte)0xff, (byte)0x03, (byte)0xb7, (byte)0x77,
            (byte)0x7c, (byte)0x44, (byte)0x47, (byte)0xaf, (byte)0x4d, (byte)0x5c, (byte)0x95, (byte)0xfb, (byte)0x77, (byte)0x02, (byte)0x03, (byte)0x01, (byte)0x00, (byte)0x01, (byte)0xa3, (byte)0x60,
            (byte)0x30, (byte)0x5e, (byte)0x30, (byte)0x1d, (byte)0x06, (byte)0x03, (byte)0x55, (byte)0x1d, (byte)0x0e, (byte)0x04, (byte)0x16, (byte)0x04, (byte)0x14, (byte)0x5a, (byte)0x4f, (byte)0xd7,
            (byte)0x6a, (byte)0x3c, (byte)0x84, (byte)0xda, (byte)0x79, (byte)0x53, (byte)0x4d, (byte)0x27, (byte)0xa5, (byte)0x56, (byte)0x83, (byte)0x63, (byte)0x4f, (byte)0x4e, (byte)0x1a, (byte)0xad,
            (byte)0x6f, (byte)0x30, (byte)0x1f, (byte)0x06, (byte)0x03, (byte)0x55, (byte)0x1d, (byte)0x23, (byte)0x04, (byte)0x18, (byte)0x30, (byte)0x16, (byte)0x80, (byte)0x14, (byte)0x5a, (byte)0x4f,
            (byte)0xd7, (byte)0x6a, (byte)0x3c, (byte)0x84, (byte)0xda, (byte)0x79, (byte)0x53, (byte)0x4d, (byte)0x27, (byte)0xa5, (byte)0x56, (byte)0x83, (byte)0x63, (byte)0x4f, (byte)0x4e, (byte)0x1a,
            (byte)0xad, (byte)0x6f, (byte)0x30, (byte)0x0f, (byte)0x06, (byte)0x03, (byte)0x55, (byte)0x1d, (byte)0x13, (byte)0x01, (byte)0x01, (byte)0xff, (byte)0x04, (byte)0x05, (byte)0x30, (byte)0x03,
            (byte)0x01, (byte)0x01, (byte)0xff, (byte)0x30, (byte)0x0b, (byte)0x06, (byte)0x03, (byte)0x55, (byte)0x1d, (byte)0x0f, (byte)0x04, (byte)0x04, (byte)0x03, (byte)0x02, (byte)0x01, (byte)0x06,
            (byte)0x30, (byte)0x0d, (byte)0x06, (byte)0x09, (byte)0x2a, (byte)0x86, (byte)0x48, (byte)0x86, (byte)0xf7, (byte)0x0d, (byte)0x01, (byte)0x01, (byte)0x0b, (byte)0x05, (byte)0x00, (byte)0x03,
            (byte)0x82, (byte)0x01, (byte)0x01, (byte)0x00, (byte)0x1d, (byte)0x32, (byte)0xb0, (byte)0xbc, (byte)0x95, (byte)0xda, (byte)0x20, (byte)0xa1, (byte)0x8f, (byte)0x24, (byte)0x0c, (byte)0xbb,
            (byte)0x15, (byte)0x38, (byte)0x2a, (byte)0xca, (byte)0xe3, (byte)0xe3, (byte)0x42, (byte)0xe2, (byte)0xde, (byte)0x85, (byte)0x89, (byte)0xa5, (byte)0x8f, (byte)0x8e, (byte)0x00, (byte)0x0a,
            (byte)0x2e, (byte)0x80, (byte)0xfe, (byte)0xac, (byte)0xf4, (byte)0x27, (byte)0x74, (byte)0x11, (byte)0x09, (byte)0x43, (byte)0x4f, (byte)0x62, (byte)0xdf, (byte)0xf9, (byte)0x71, (byte)0xea,
            (byte)0x48, (byte)0xd6, (byte)0x13, (byte)0x9e, (byte)0x8b, (byte)0xc2, (byte)0xde, (byte)0x56, (byte)0x6a, (byte)0x6a, (byte)0x54, (byte)0xb5, (byte)0x6a, (byte)0x21, (byte)0xfc, (byte)0x60,
            (byte)0xfd, (byte)0xbc, (byte)0x23, (byte)0x0c, (byte)0x99, (byte)0xbc, (byte)0x13, (byte)0xd3, (byte)0xd1, (byte)0x92, (byte)0xe7, (byte)0x3a, (byte)0x17, (byte)0x2f, (byte)0x0e, (byte)0x41,
            (byte)0x15, (byte)0xcd, (byte)0x50, (byte)0x69, (byte)0xae, (byte)0x13, (byte)0xa0, (byte)0xee, (byte)0xca, (byte)0x73, (byte)0xfa, (byte)0x80, (byte)0xd0, (byte)0xc6, (byte)0x8f, (byte)0x14,
            (byte)0x04, (byte)0x73, (byte)0x1d, (byte)0x3a, (byte)0x51, (byte)0xcd, (byte)0x20, (byte)0x67, (byte)0xc7, (byte)0x54, (byte)0xb1, (byte)0x48, (byte)0x80, (byte)0xa4, (byte)0x65, (byte)0xf1,
            (byte)0x4c, (byte)0x86, (byte)0x4e, (byte)0x92, (byte)0x6e, (byte)0xfd, (byte)0xb4, (byte)0xa2, (byte)0xab, (byte)0x7d, (byte)0xc4, (byte)0xab, (byte)0x27, (byte)0xa6, (byte)0x2d, (byte)0xb7,
            (byte)0xae, (byte)0x31, (byte)0x4f, (byte)0x4f, (byte)0x58, (byte)0x00, (byte)0x5f, (byte)0x88, (byte)0xff, (byte)0xd8, (byte)0x74, (byte)0xd9, (byte)0xd3, (byte)0x1c, (byte)0x88, (byte)0x30,
            (byte)0x27, (byte)0x4e, (byte)0xdb, (byte)0x3d, (byte)0x00, (byte)0xa3, (byte)0x74, (byte)0x81, (byte)0xf8, (byte)0x2c, (byte)0x7c, (byte)0x09, (byte)0x2d, (byte)0x95, (byte)0xfe, (byte)0x77,
            (byte)0x3f, (byte)0x62, (byte)0x6c, (byte)0x00, (byte)0xc8, (byte)0xec, (byte)0xea, (byte)0x89, (byte)0x87, (byte)0x89, (byte)0x22, (byte)0xc1, (byte)0x94, (byte)0xe4, (byte)0x7c, (byte)0xa5,
            (byte)0x0e, (byte)0x87, (byte)0x93, (byte)0xd0, (byte)0x1a, (byte)0xbb, (byte)0xfb, (byte)0x7e, (byte)0xef, (byte)0xc8, (byte)0x33, (byte)0x4b, (byte)0x15, (byte)0x2e, (byte)0x6a, (byte)0xdf,
            (byte)0x39, (byte)0x49, (byte)0xbf, (byte)0xed, (byte)0xc3, (byte)0x09, (byte)0x7a, (byte)0x96, (byte)0x83, (byte)0x53, (byte)0x0d, (byte)0xdf, (byte)0x5b, (byte)0xa6, (byte)0x33, (byte)0x3c,
            (byte)0x8c, (byte)0xf5, (byte)0x67, (byte)0x40, (byte)0x1f, (byte)0x30, (byte)0xad, (byte)0x2c, (byte)0xab, (byte)0xd5, (byte)0x14, (byte)0x48, (byte)0xbf, (byte)0x4c, (byte)0x21, (byte)0xec,
            (byte)0x44, (byte)0xd6, (byte)0xaf, (byte)0xaa, (byte)0x04, (byte)0xed, (byte)0x40, (byte)0xc4, (byte)0x33, (byte)0xa9, (byte)0xcc, (byte)0x48, (byte)0xe0, (byte)0x63, (byte)0x13, (byte)0x7f,
            (byte)0x61, (byte)0x1f, (byte)0xd8, (byte)0xd7, (byte)0xc5, (byte)0xd6, (byte)0xae, (byte)0xd8, (byte)0x18, (byte)0xf6, (byte)0x44, (byte)0xe1, (byte)0x96, (byte)0x38, (byte)0x0d, (byte)0x45,
            (byte)0x2a, (byte)0xc9, (byte)0xeb, (byte)0xad, (byte)0x16, (byte)0x03, (byte)0x03, (byte)0x01, (byte)0x2c, (byte)0x0c, (byte)0x00, (byte)0x01, (byte)0x28, (byte)0x03, (byte)0x00, (byte)0x1d,
            (byte)0x20, (byte)0x41, (byte)0x1f, (byte)0x42, (byte)0x3a, (byte)0x23, (byte)0xc1, (byte)0x31, (byte)0x45, (byte)0xf9, (byte)0x20, (byte)0xf0, (byte)0x65, (byte)0x05, (byte)0x3d, (byte)0x15,
            (byte)0xad, (byte)0x69, (byte)0xbb, (byte)0xc9, (byte)0x51, (byte)0x78, (byte)0x40, (byte)0x62, (byte)0xd0, (byte)0x3c, (byte)0x7c, (byte)0x77, (byte)0xe9, (byte)0xbd, (byte)0x7a, (byte)0xda,
            (byte)0x36, (byte)0x08, (byte)0x04, (byte)0x01, (byte)0x00, (byte)0xb5, (byte)0x6d, (byte)0xfe, (byte)0xf0, (byte)0xa7, (byte)0x99, (byte)0x1f, (byte)0x50, (byte)0x67, (byte)0x5c, (byte)0xfd,
            (byte)0x38, (byte)0x55, (byte)0x38, (byte)0x94, (byte)0x02, (byte)0xa2, (byte)0x75, (byte)0x5e, (byte)0x98, (byte)0x35, (byte)0x35, (byte)0x15, (byte)0x23, (byte)0x19, (byte)0x9d, (byte)0x80,
            (byte)0x66, (byte)0x1f, (byte)0x7a, (byte)0x89, (byte)0x8a, (byte)0xcb, (byte)0x50, (byte)0x81, (byte)0xba, (byte)0x91, (byte)0x8d, (byte)0x73, (byte)0xe8, (byte)0xf7, (byte)0x20, (byte)0x93,
            (byte)0x71, (byte)0x7a, (byte)0x6f, (byte)0x21, (byte)0x80, (byte)0x79, (byte)0xf3, (byte)0xe1, (byte)0x12, (byte)0x15, (byte)0xc1, (byte)0x2e, (byte)0xef, (byte)0x12, (byte)0x6a, (byte)0xed,
            (byte)0x11, (byte)0x1a, (byte)0xaf, (byte)0x68, (byte)0x03, (byte)0xa3, (byte)0xd5, (byte)0xae, (byte)0xb4, (byte)0xc6, (byte)0xb8, (byte)0xce, (byte)0xdc, (byte)0x52, (byte)0xd5, (byte)0x99,
            (byte)0x6b, (byte)0x25, (byte)0xc0, (byte)0x25, (byte)0x46, (byte)0xae, (byte)0x9f, (byte)0x21, (byte)0x3e, (byte)0x6f, (byte)0x7f, (byte)0xaf, (byte)0x8b, (byte)0x60, (byte)0xba, (byte)0x0f,
            (byte)0xb8, (byte)0x9c, (byte)0x1e, (byte)0x83, (byte)0x21, (byte)0xec, (byte)0x06, (byte)0xe6, (byte)0xa6, (byte)0x9a, (byte)0xee, (byte)0x01, (byte)0x74, (byte)0x61, (byte)0xcd, (byte)0x0a,
            (byte)0xe8, (byte)0x8b, (byte)0x28, (byte)0xe4, (byte)0x13, (byte)0xac, (byte)0x34, (byte)0xf1, (byte)0xb5, (byte)0x57, (byte)0xab, (byte)0x3f, (byte)0x52, (byte)0xe5, (byte)0x01, (byte)0x47,
            (byte)0x8e, (byte)0x2d, (byte)0xc0, (byte)0x88, (byte)0x60, (byte)0xaf, (byte)0x16, (byte)0x04, (byte)0x3e, (byte)0xa0, (byte)0x14, (byte)0xe5, (byte)0x03, (byte)0x2a, (byte)0x34, (byte)0x93,
            (byte)0xca, (byte)0x2c, (byte)0x43, (byte)0x79, (byte)0xf2, (byte)0xf8, (byte)0xca, (byte)0xc2, (byte)0xd9, (byte)0x79, (byte)0xe4, (byte)0xf5, (byte)0xd5, (byte)0x3a, (byte)0x46, (byte)0xbd,
            (byte)0x14, (byte)0xd1, (byte)0x1f, (byte)0x6d, (byte)0x58, (byte)0x61, (byte)0x20, (byte)0x49, (byte)0x3d, (byte)0x41, (byte)0xa0, (byte)0xec, (byte)0xdf, (byte)0xf9, (byte)0x82, (byte)0x7e,
            (byte)0xfe, (byte)0xc7, (byte)0xfa, (byte)0xb7, (byte)0x87, (byte)0x72, (byte)0xdb, (byte)0x98, (byte)0x26, (byte)0x70, (byte)0xe7, (byte)0x7b, (byte)0xe1, (byte)0xdc, (byte)0x85, (byte)0x01,
            (byte)0x75, (byte)0xf0, (byte)0x39, (byte)0xd9, (byte)0xc1, (byte)0x94, (byte)0x0e, (byte)0x33, (byte)0xdd, (byte)0x70, (byte)0xbb, (byte)0x5b, (byte)0xc9, (byte)0xa4, (byte)0xa0, (byte)0xda,
            (byte)0xc8, (byte)0x0d, (byte)0x95, (byte)0x91, (byte)0x60, (byte)0x1e, (byte)0xa8, (byte)0xf5, (byte)0x3d, (byte)0x14, (byte)0xdc, (byte)0x0d, (byte)0xba, (byte)0x01, (byte)0x37, (byte)0x83,
            (byte)0xa6, (byte)0x86, (byte)0x44, (byte)0xc0, (byte)0x81, (byte)0x3e, (byte)0xf3, (byte)0x46, (byte)0x47, (byte)0x72, (byte)0x04, (byte)0xa0, (byte)0xb4, (byte)0xb1, (byte)0xee, (byte)0x8a,
            (byte)0x40, (byte)0x70, (byte)0x3d, (byte)0xc7, (byte)0x96, (byte)0x3f, (byte)0x49, (byte)0x75, (byte)0xd2, (byte)0x2c, (byte)0x81, (byte)0x6a, (byte)0x47, (byte)0x29, (byte)0x49, (byte)0xa4,
            (byte)0xd3, (byte)0x30, (byte)0x92, (byte)0xa4, (byte)0x37, (byte)0x16, (byte)0x03, (byte)0x03, (byte)0x00, (byte)0x04, (byte)0x0e, (byte)0x00, (byte)0x00, (byte)0x00
    };

        final byte[] private_cert_key_bytes = {
                (byte)0x2d, (byte)0x2d, (byte)0x2d, (byte)0x2d, (byte)0x2d, (byte)0x42, (byte)0x45, (byte)0x47, (byte)0x49, (byte)0x4e, (byte)0x20, (byte)0x50, (byte)0x52, (byte)0x49, (byte)0x56,
                (byte)0x41, (byte)0x54, (byte)0x45, (byte)0x20, (byte)0x4b, (byte)0x45, (byte)0x59, (byte)0x2d, (byte)0x2d, (byte)0x2d, (byte)0x2d, (byte)0x2d, (byte)0x0a, (byte)0x4d, (byte)0x49, (byte)0x49,
                (byte)0x45, (byte)0x76, (byte)0x51, (byte)0x49, (byte)0x42, (byte)0x41, (byte)0x44, (byte)0x41, (byte)0x4e, (byte)0x42, (byte)0x67, (byte)0x6b, (byte)0x71, (byte)0x68, (byte)0x6b, (byte)0x69,
                (byte)0x47, (byte)0x39, (byte)0x77, (byte)0x30, (byte)0x42, (byte)0x41, (byte)0x51, (byte)0x45, (byte)0x46, (byte)0x41, (byte)0x41, (byte)0x53, (byte)0x43, (byte)0x42, (byte)0x4b, (byte)0x63,
                (byte)0x77, (byte)0x67, (byte)0x67, (byte)0x53, (byte)0x6a, (byte)0x41, (byte)0x67, (byte)0x45, (byte)0x41, (byte)0x41, (byte)0x6f, (byte)0x49, (byte)0x42, (byte)0x41, (byte)0x51, (byte)0x43,
                (byte)0x39, (byte)0x5a, (byte)0x39, (byte)0x37, (byte)0x34, (byte)0x66, (byte)0x2b, (byte)0x48, (byte)0x49, (byte)0x4a, (byte)0x53, (byte)0x59, (byte)0x53, (byte)0x0a, (byte)0x44, (byte)0x49,
                (byte)0x42, (byte)0x75, (byte)0x36, (byte)0x57, (byte)0x52, (byte)0x61, (byte)0x48, (byte)0x34, (byte)0x4c, (byte)0x53, (byte)0x6d, (byte)0x51, (byte)0x70, (byte)0x69, (byte)0x6f, (byte)0x6e,
                (byte)0x71, (byte)0x66, (byte)0x69, (byte)0x70, (byte)0x47, (byte)0x4e, (byte)0x62, (byte)0x70, (byte)0x46, (byte)0x50, (byte)0x2f, (byte)0x4a, (byte)0x6f, (byte)0x75, (byte)0x45, (byte)0x53,
                (byte)0x61, (byte)0x38, (byte)0x75, (byte)0x56, (byte)0x67, (byte)0x59, (byte)0x74, (byte)0x2b, (byte)0x2b, (byte)0x4e, (byte)0x6b, (byte)0x6e, (byte)0x33, (byte)0x64, (byte)0x37, (byte)0x33,
                (byte)0x44, (byte)0x63, (byte)0x66, (byte)0x77, (byte)0x59, (byte)0x36, (byte)0x50, (byte)0x75, (byte)0x6b, (byte)0x4c, (byte)0x6f, (byte)0x4d, (byte)0x33, (byte)0x70, (byte)0x0a, (byte)0x6c,
                (byte)0x6d, (byte)0x58, (byte)0x2b, (byte)0x73, (byte)0x5a, (byte)0x44, (byte)0x70, (byte)0x4e, (byte)0x53, (byte)0x75, (byte)0x33, (byte)0x45, (byte)0x39, (byte)0x66, (byte)0x70, (byte)0x45,
                (byte)0x46, (byte)0x37, (byte)0x32, (byte)0x68, (byte)0x49, (byte)0x31, (byte)0x43, (byte)0x47, (byte)0x6f, (byte)0x4a, (byte)0x46, (byte)0x6d, (byte)0x55, (byte)0x79, (byte)0x71, (byte)0x2b,
                (byte)0x34, (byte)0x5a, (byte)0x7a, (byte)0x51, (byte)0x66, (byte)0x45, (byte)0x43, (byte)0x72, (byte)0x67, (byte)0x67, (byte)0x54, (byte)0x4f, (byte)0x59, (byte)0x49, (byte)0x44, (byte)0x55,
                (byte)0x38, (byte)0x53, (byte)0x42, (byte)0x51, (byte)0x78, (byte)0x6a, (byte)0x30, (byte)0x2f, (byte)0x78, (byte)0x75, (byte)0x63, (byte)0x75, (byte)0x44, (byte)0x37, (byte)0x61, (byte)0x0a,
                (byte)0x4b, (byte)0x42, (byte)0x4c, (byte)0x77, (byte)0x69, (byte)0x4c, (byte)0x65, (byte)0x76, (byte)0x55, (byte)0x6a, (byte)0x51, (byte)0x36, (byte)0x66, (byte)0x36, (byte)0x6b, (byte)0x61,
                (byte)0x39, (byte)0x52, (byte)0x4e, (byte)0x4d, (byte)0x31, (byte)0x52, (byte)0x52, (byte)0x44, (byte)0x49, (byte)0x56, (byte)0x72, (byte)0x4d, (byte)0x6c, (byte)0x38, (byte)0x47, (byte)0x65,
                (byte)0x7a, (byte)0x79, (byte)0x30, (byte)0x57, (byte)0x68, (byte)0x73, (byte)0x74, (byte)0x62, (byte)0x48, (byte)0x57, (byte)0x4c, (byte)0x5a, (byte)0x36, (byte)0x6a, (byte)0x37, (byte)0x78,
                (byte)0x2f, (byte)0x45, (byte)0x64, (byte)0x2b, (byte)0x74, (byte)0x51, (byte)0x6a, (byte)0x30, (byte)0x49, (byte)0x38, (byte)0x74, (byte)0x65, (byte)0x47, (byte)0x75, (byte)0x64, (byte)0x52,
                (byte)0x0a, (byte)0x55, (byte)0x69, (byte)0x78, (byte)0x4e, (byte)0x76, (byte)0x36, (byte)0x6c, (byte)0x35, (byte)0x59, (byte)0x67, (byte)0x33, (byte)0x34, (byte)0x34, (byte)0x6d, (byte)0x51,
                (byte)0x66, (byte)0x31, (byte)0x53, (byte)0x4a, (byte)0x2f, (byte)0x79, (byte)0x54, (byte)0x69, (byte)0x73, (byte)0x67, (byte)0x79, (byte)0x7a, (byte)0x69, (byte)0x74, (byte)0x33, (byte)0x75,
                (byte)0x53, (byte)0x64, (byte)0x54, (byte)0x45, (byte)0x52, (byte)0x73, (byte)0x34, (byte)0x76, (byte)0x71, (byte)0x52, (byte)0x39, (byte)0x64, (byte)0x50, (byte)0x41, (byte)0x4e, (byte)0x5a,
                (byte)0x72, (byte)0x31, (byte)0x4a, (byte)0x38, (byte)0x2f, (byte)0x5a, (byte)0x48, (byte)0x5a, (byte)0x36, (byte)0x76, (byte)0x42, (byte)0x75, (byte)0x45, (byte)0x34, (byte)0x56, (byte)0x2b,
                (byte)0x6d, (byte)0x0a, (byte)0x4d, (byte)0x36, (byte)0x30, (byte)0x64, (byte)0x49, (byte)0x4b, (byte)0x4b, (byte)0x4f, (byte)0x6f, (byte)0x61, (byte)0x78, (byte)0x6d, (byte)0x74, (byte)0x53,
                (byte)0x6c, (byte)0x78, (byte)0x64, (byte)0x73, (byte)0x30, (byte)0x66, (byte)0x79, (byte)0x59, (byte)0x64, (byte)0x41, (byte)0x4d, (byte)0x38, (byte)0x70, (byte)0x58, (byte)0x75, (byte)0x57,
                (byte)0x53, (byte)0x6e, (byte)0x54, (byte)0x4f, (byte)0x63, (byte)0x47, (byte)0x45, (byte)0x79, (byte)0x77, (byte)0x77, (byte)0x67, (byte)0x38, (byte)0x66, (byte)0x34, (byte)0x73, (byte)0x75,
                (byte)0x6d, (byte)0x42, (byte)0x4e, (byte)0x54, (byte)0x35, (byte)0x51, (byte)0x50, (byte)0x34, (byte)0x76, (byte)0x4c, (byte)0x62, (byte)0x6f, (byte)0x4d, (byte)0x48, (byte)0x4f, (byte)0x56,
                (byte)0x77, (byte)0x59, (byte)0x0a, (byte)0x59, (byte)0x4e, (byte)0x2b, (byte)0x44, (byte)0x6e, (byte)0x61, (byte)0x7a, (byte)0x62, (byte)0x41, (byte)0x67, (byte)0x4d, (byte)0x42, (byte)0x41,
                (byte)0x41, (byte)0x45, (byte)0x43, (byte)0x67, (byte)0x67, (byte)0x45, (byte)0x42, (byte)0x41, (byte)0x4b, (byte)0x34, (byte)0x79, (byte)0x41, (byte)0x37, (byte)0x31, (byte)0x6b, (byte)0x54,
                (byte)0x39, (byte)0x44, (byte)0x31, (byte)0x61, (byte)0x59, (byte)0x67, (byte)0x62, (byte)0x2b, (byte)0x31, (byte)0x73, (byte)0x47, (byte)0x53, (byte)0x41, (byte)0x63, (byte)0x6a, (byte)0x42,
                (byte)0x2f, (byte)0x79, (byte)0x4d, (byte)0x53, (byte)0x43, (byte)0x4c, (byte)0x69, (byte)0x38, (byte)0x41, (byte)0x67, (byte)0x66, (byte)0x4e, (byte)0x7a, (byte)0x4a, (byte)0x75, (byte)0x53,
                (byte)0x44, (byte)0x4c, (byte)0x61, (byte)0x0a, (byte)0x54, (byte)0x53, (byte)0x71, (byte)0x6c, (byte)0x46, (byte)0x34, (byte)0x31, (byte)0x6d, (byte)0x6f, (byte)0x51, (byte)0x46, (byte)0x4b,
                (byte)0x32, (byte)0x59, (byte)0x54, (byte)0x49, (byte)0x2f, (byte)0x2f, (byte)0x33, (byte)0x48, (byte)0x6f, (byte)0x33, (byte)0x4d, (byte)0x48, (byte)0x57, (byte)0x45, (byte)0x44, (byte)0x6c,
                (byte)0x62, (byte)0x77, (byte)0x77, (byte)0x45, (byte)0x76, (byte)0x70, (byte)0x78, (byte)0x67, (byte)0x65, (byte)0x44, (byte)0x6e, (byte)0x41, (byte)0x4d, (byte)0x56, (byte)0x68, (byte)0x71,
                (byte)0x55, (byte)0x47, (byte)0x51, (byte)0x32, (byte)0x57, (byte)0x64, (byte)0x4b, (byte)0x39, (byte)0x2f, (byte)0x62, (byte)0x69, (byte)0x58, (byte)0x76, (byte)0x42, (byte)0x72, (byte)0x5a,
                (byte)0x71, (byte)0x6f, (byte)0x6f, (byte)0x38, (byte)0x0a, (byte)0x4c, (byte)0x57, (byte)0x6a, (byte)0x72, (byte)0x6e, (byte)0x34, (byte)0x49, (byte)0x4e, (byte)0x57, (byte)0x59, (byte)0x31,
                (byte)0x76, (byte)0x5a, (byte)0x58, (byte)0x6e, (byte)0x59, (byte)0x34, (byte)0x42, (byte)0x77, (byte)0x54, (byte)0x52, (byte)0x4a, (byte)0x7a, (byte)0x6c, (byte)0x69, (byte)0x75, (byte)0x52,
                (byte)0x34, (byte)0x59, (byte)0x33, (byte)0x72, (byte)0x46, (byte)0x63, (byte)0x50, (byte)0x6a, (byte)0x4f, (byte)0x39, (byte)0x6d, (byte)0x78, (byte)0x47, (byte)0x4d, (byte)0x54, (byte)0x6d,
                (byte)0x4c, (byte)0x54, (byte)0x77, (byte)0x47, (byte)0x6d, (byte)0x4f, (byte)0x32, (byte)0x37, (byte)0x52, (byte)0x69, (byte)0x77, (byte)0x69, (byte)0x55, (byte)0x78, (byte)0x2b, (byte)0x66,
                (byte)0x64, (byte)0x6d, (byte)0x31, (byte)0x61, (byte)0x68, (byte)0x0a, (byte)0x66, (byte)0x70, (byte)0x36, (byte)0x6e, (byte)0x58, (byte)0x47, (byte)0x4f, (byte)0x57, (byte)0x4d, (byte)0x73,
                (byte)0x74, (byte)0x67, (byte)0x4e, (byte)0x75, (byte)0x72, (byte)0x32, (byte)0x37, (byte)0x6f, (byte)0x46, (byte)0x51, (byte)0x33, (byte)0x64, (byte)0x67, (byte)0x2f, (byte)0x2b, (byte)0x4f,
                (byte)0x55, (byte)0x73, (byte)0x61, (byte)0x72, (byte)0x4c, (byte)0x47, (byte)0x53, (byte)0x68, (byte)0x71, (byte)0x6c, (byte)0x36, (byte)0x45, (byte)0x78, (byte)0x4a, (byte)0x38, (byte)0x64,
                (byte)0x70, (byte)0x73, (byte)0x59, (byte)0x64, (byte)0x2b, (byte)0x73, (byte)0x2b, (byte)0x72, (byte)0x37, (byte)0x52, (byte)0x6b, (byte)0x50, (byte)0x6b, (byte)0x6e, (byte)0x66, (byte)0x33,
                (byte)0x49, (byte)0x41, (byte)0x30, (byte)0x4d, (byte)0x48, (byte)0x4f, (byte)0x0a, (byte)0x51, (byte)0x4a, (byte)0x45, (byte)0x63, (byte)0x35, (byte)0x4a, (byte)0x6f, (byte)0x41, (byte)0x4d,
                (byte)0x4e, (byte)0x6b, (byte)0x35, (byte)0x77, (byte)0x4b, (byte)0x61, (byte)0x55, (byte)0x39, (byte)0x33, (byte)0x52, (byte)0x53, (byte)0x47, (byte)0x47, (byte)0x35, (byte)0x67, (byte)0x78,
                (byte)0x54, (byte)0x6c, (byte)0x6f, (byte)0x75, (byte)0x33, (byte)0x51, (byte)0x67, (byte)0x32, (byte)0x6c, (byte)0x2f, (byte)0x42, (byte)0x32, (byte)0x34, (byte)0x32, (byte)0x4d, (byte)0x68,
                (byte)0x6b, (byte)0x48, (byte)0x6f, (byte)0x69, (byte)0x37, (byte)0x36, (byte)0x50, (byte)0x41, (byte)0x63, (byte)0x34, (byte)0x50, (byte)0x64, (byte)0x63, (byte)0x39, (byte)0x2f, (byte)0x4d,
                (byte)0x4c, (byte)0x46, (byte)0x57, (byte)0x2b, (byte)0x4b, (byte)0x57, (byte)0x59, (byte)0x0a, (byte)0x63, (byte)0x41, (byte)0x48, (byte)0x57, (byte)0x2f, (byte)0x56, (byte)0x5a, (byte)0x70,
                (byte)0x48, (byte)0x4f, (byte)0x63, (byte)0x4e, (byte)0x71, (byte)0x37, (byte)0x58, (byte)0x63, (byte)0x42, (byte)0x51, (byte)0x6a, (byte)0x75, (byte)0x39, (byte)0x66, (byte)0x4c, (byte)0x53,
                (byte)0x67, (byte)0x67, (byte)0x4a, (byte)0x4e, (byte)0x67, (byte)0x49, (byte)0x6b, (byte)0x33, (byte)0x72, (byte)0x50, (byte)0x39, (byte)0x57, (byte)0x53, (byte)0x42, (byte)0x6a, (byte)0x54,
                (byte)0x41, (byte)0x4d, (byte)0x45, (byte)0x43, (byte)0x67, (byte)0x59, (byte)0x45, (byte)0x41, (byte)0x33, (byte)0x73, (byte)0x49, (byte)0x6d, (byte)0x54, (byte)0x66, (byte)0x4d, (byte)0x34,
                (byte)0x2b, (byte)0x61, (byte)0x4d, (byte)0x35, (byte)0x66, (byte)0x35, (byte)0x5a, (byte)0x42, (byte)0x0a, (byte)0x77, (byte)0x68, (byte)0x36, (byte)0x30, (byte)0x67, (byte)0x2f, (byte)0x42,
                (byte)0x79, (byte)0x72, (byte)0x33, (byte)0x67, (byte)0x56, (byte)0x58, (byte)0x56, (byte)0x49, (byte)0x4f, (byte)0x73, (byte)0x35, (byte)0x69, (byte)0x56, (byte)0x32, (byte)0x50, (byte)0x74,
                (byte)0x4a, (byte)0x5a, (byte)0x32, (byte)0x4e, (byte)0x72, (byte)0x70, (byte)0x4d, (byte)0x4a, (byte)0x57, (byte)0x51, (byte)0x65, (byte)0x79, (byte)0x41, (byte)0x75, (byte)0x6e, (byte)0x66,
                (byte)0x74, (byte)0x5a, (byte)0x30, (byte)0x72, (byte)0x6f, (byte)0x53, (byte)0x5a, (byte)0x4a, (byte)0x54, (byte)0x73, (byte)0x65, (byte)0x6b, (byte)0x35, (byte)0x4c, (byte)0x31, (byte)0x73,
                (byte)0x50, (byte)0x43, (byte)0x7a, (byte)0x2b, (byte)0x68, (byte)0x30, (byte)0x65, (byte)0x31, (byte)0x54, (byte)0x0a, (byte)0x4a, (byte)0x64, (byte)0x37, (byte)0x6d, (byte)0x42, (byte)0x49,
                (byte)0x63, (byte)0x71, (byte)0x77, (byte)0x39, (byte)0x6c, (byte)0x74, (byte)0x53, (byte)0x4d, (byte)0x6a, (byte)0x57, (byte)0x50, (byte)0x55, (byte)0x39, (byte)0x46, (byte)0x4e, (byte)0x59,
                (byte)0x57, (byte)0x45, (byte)0x6a, (byte)0x49, (byte)0x64, (byte)0x32, (byte)0x64, (byte)0x6e, (byte)0x71, (byte)0x78, (byte)0x32, (byte)0x49, (byte)0x59, (byte)0x56, (byte)0x2f, (byte)0x70,
                (byte)0x53, (byte)0x50, (byte)0x4c, (byte)0x75, (byte)0x34, (byte)0x36, (byte)0x37, (byte)0x75, (byte)0x6e, (byte)0x46, (byte)0x5a, (byte)0x66, (byte)0x72, (byte)0x4b, (byte)0x44, (byte)0x65,
                (byte)0x66, (byte)0x59, (byte)0x49, (byte)0x79, (byte)0x2b, (byte)0x4c, (byte)0x64, (byte)0x65, (byte)0x6d, (byte)0x6c, (byte)0x0a, (byte)0x48, (byte)0x42, (byte)0x6b, (byte)0x49, (byte)0x73,
                (byte)0x41, (byte)0x45, (byte)0x53, (byte)0x75, (byte)0x34, (byte)0x66, (byte)0x58, (byte)0x76, (byte)0x73, (byte)0x62, (byte)0x39, (byte)0x6f, (byte)0x59, (byte)0x33, (byte)0x79, (byte)0x52,
                (byte)0x42, (byte)0x36, (byte)0x6d, (byte)0x66, (byte)0x66, (byte)0x45, (byte)0x43, (byte)0x67, (byte)0x59, (byte)0x45, (byte)0x41, (byte)0x32, (byte)0x61, (byte)0x75, (byte)0x55, (byte)0x43,
                (byte)0x65, (byte)0x44, (byte)0x66, (byte)0x79, (byte)0x61, (byte)0x30, (byte)0x42, (byte)0x48, (byte)0x2b, (byte)0x31, (byte)0x6a, (byte)0x59, (byte)0x77, (byte)0x4a, (byte)0x6b, (byte)0x79,
                (byte)0x78, (byte)0x58, (byte)0x71, (byte)0x38, (byte)0x2b, (byte)0x31, (byte)0x76, (byte)0x6d, (byte)0x58, (byte)0x64, (byte)0x53, (byte)0x0a, (byte)0x4c, (byte)0x31, (byte)0x46, (byte)0x46,
                (byte)0x70, (byte)0x78, (byte)0x67, (byte)0x75, (byte)0x43, (byte)0x36, (byte)0x4a, (byte)0x62, (byte)0x2f, (byte)0x4d, (byte)0x6e, (byte)0x62, (byte)0x4b, (byte)0x6a, (byte)0x37, (byte)0x6c,
                (byte)0x34, (byte)0x46, (byte)0x50, (byte)0x4d, (byte)0x51, (byte)0x4e, (byte)0x74, (byte)0x31, (byte)0x52, (byte)0x34, (byte)0x57, (byte)0x41, (byte)0x65, (byte)0x78, (byte)0x42, (byte)0x54,
                (byte)0x6b, (byte)0x6c, (byte)0x38, (byte)0x6e, (byte)0x61, (byte)0x49, (byte)0x74, (byte)0x70, (byte)0x57, (byte)0x6f, (byte)0x45, (byte)0x43, (byte)0x43, (byte)0x6c, (byte)0x43, (byte)0x37,
                (byte)0x72, (byte)0x31, (byte)0x56, (byte)0x34, (byte)0x6f, (byte)0x72, (byte)0x47, (byte)0x4f, (byte)0x34, (byte)0x71, (byte)0x5a, (byte)0x51, (byte)0x0a, (byte)0x4b, (byte)0x73, (byte)0x6e,
                (byte)0x2f, (byte)0x59, (byte)0x66, (byte)0x6e, (byte)0x6b, (byte)0x4a, (byte)0x35, (byte)0x46, (byte)0x41, (byte)0x72, (byte)0x59, (byte)0x55, (byte)0x33, (byte)0x6f, (byte)0x65, (byte)0x62,
                (byte)0x68, (byte)0x45, (byte)0x51, (byte)0x4a, (byte)0x7a, (byte)0x2b, (byte)0x73, (byte)0x73, (byte)0x46, (byte)0x54, (byte)0x32, (byte)0x6c, (byte)0x63, (byte)0x68, (byte)0x38, (byte)0x51,
                (byte)0x66, (byte)0x4b, (byte)0x4f, (byte)0x41, (byte)0x36, (byte)0x48, (byte)0x75, (byte)0x53, (byte)0x39, (byte)0x66, (byte)0x68, (byte)0x72, (byte)0x68, (byte)0x4a, (byte)0x49, (byte)0x78,
                (byte)0x69, (byte)0x35, (byte)0x45, (byte)0x69, (byte)0x51, (byte)0x65, (byte)0x66, (byte)0x6b, (byte)0x59, (byte)0x43, (byte)0x48, (byte)0x45, (byte)0x70, (byte)0x0a, (byte)0x4e, (byte)0x61,
                (byte)0x78, (byte)0x65, (byte)0x37, (byte)0x62, (byte)0x46, (byte)0x38, (byte)0x2b, (byte)0x34, (byte)0x73, (byte)0x43, (byte)0x67, (byte)0x59, (byte)0x41, (byte)0x4b, (byte)0x75, (byte)0x55,
                (byte)0x67, (byte)0x45, (byte)0x79, (byte)0x71, (byte)0x65, (byte)0x38, (byte)0x68, (byte)0x4e, (byte)0x6f, (byte)0x59, (byte)0x78, (byte)0x56, (byte)0x41, (byte)0x48, (byte)0x47, (byte)0x31,
                (byte)0x6c, (byte)0x50, (byte)0x58, (byte)0x4c, (byte)0x7a, (byte)0x34, (byte)0x71, (byte)0x34, (byte)0x78, (byte)0x44, (byte)0x77, (byte)0x4e, (byte)0x61, (byte)0x4b, (byte)0x57, (byte)0x4e,
                (byte)0x63, (byte)0x6c, (byte)0x36, (byte)0x66, (byte)0x72, (byte)0x2f, (byte)0x71, (byte)0x69, (byte)0x36, (byte)0x64, (byte)0x63, (byte)0x4f, (byte)0x6d, (byte)0x63, (byte)0x0a, (byte)0x4a,
                (byte)0x4e, (byte)0x6d, (byte)0x52, (byte)0x6f, (byte)0x49, (byte)0x4a, (byte)0x69, (byte)0x54, (byte)0x5a, (byte)0x33, (byte)0x66, (byte)0x4a, (byte)0x4e, (byte)0x63, (byte)0x34, (byte)0x6f,
                (byte)0x6d, (byte)0x53, (byte)0x70, (byte)0x58, (byte)0x66, (byte)0x58, (byte)0x63, (byte)0x63, (byte)0x63, (byte)0x54, (byte)0x46, (byte)0x69, (byte)0x57, (byte)0x70, (byte)0x4f, (byte)0x55,
                (byte)0x57, (byte)0x75, (byte)0x37, (byte)0x58, (byte)0x68, (byte)0x4a, (byte)0x31, (byte)0x2b, (byte)0x54, (byte)0x71, (byte)0x6b, (byte)0x6e, (byte)0x56, (byte)0x36, (byte)0x4b, (byte)0x39,
                (byte)0x77, (byte)0x7a, (byte)0x36, (byte)0x35, (byte)0x61, (byte)0x47, (byte)0x61, (byte)0x37, (byte)0x6b, (byte)0x36, (byte)0x4b, (byte)0x70, (byte)0x35, (byte)0x39, (byte)0x32, (byte)0x0a,
                (byte)0x4d, (byte)0x41, (byte)0x5a, (byte)0x52, (byte)0x43, (byte)0x69, (byte)0x2f, (byte)0x57, (byte)0x62, (byte)0x34, (byte)0x75, (byte)0x51, (byte)0x77, (byte)0x74, (byte)0x70, (byte)0x49,
                (byte)0x37, (byte)0x59, (byte)0x6a, (byte)0x61, (byte)0x6f, (byte)0x49, (byte)0x57, (byte)0x73, (byte)0x4d, (byte)0x2f, (byte)0x61, (byte)0x2f, (byte)0x78, (byte)0x32, (byte)0x36, (byte)0x32,
                (byte)0x61, (byte)0x70, (byte)0x49, (byte)0x62, (byte)0x66, (byte)0x77, (byte)0x47, (byte)0x36, (byte)0x32, (byte)0x33, (byte)0x42, (byte)0x4d, (byte)0x63, (byte)0x2b, (byte)0x48, (byte)0x7a,
                (byte)0x37, (byte)0x42, (byte)0x52, (byte)0x30, (byte)0x71, (byte)0x52, (byte)0x45, (byte)0x55, (byte)0x30, (byte)0x51, (byte)0x4b, (byte)0x42, (byte)0x67, (byte)0x44, (byte)0x43, (byte)0x32,
                (byte)0x0a, (byte)0x79, (byte)0x49, (byte)0x50, (byte)0x4a, (byte)0x63, (byte)0x75, (byte)0x77, (byte)0x45, (byte)0x72, (byte)0x75, (byte)0x72, (byte)0x42, (byte)0x53, (byte)0x75, (byte)0x76,
                (byte)0x65, (byte)0x6e, (byte)0x6b, (byte)0x59, (byte)0x68, (byte)0x47, (byte)0x70, (byte)0x73, (byte)0x4b, (byte)0x75, (byte)0x56, (byte)0x34, (byte)0x6b, (byte)0x4b, (byte)0x52, (byte)0x31,
                (byte)0x7a, (byte)0x31, (byte)0x39, (byte)0x33, (byte)0x45, (byte)0x65, (byte)0x4a, (byte)0x47, (byte)0x62, (byte)0x2f, (byte)0x58, (byte)0x41, (byte)0x2b, (byte)0x6b, (byte)0x4c, (byte)0x45,
                (byte)0x4d, (byte)0x73, (byte)0x49, (byte)0x2b, (byte)0x61, (byte)0x30, (byte)0x57, (byte)0x78, (byte)0x32, (byte)0x63, (byte)0x50, (byte)0x48, (byte)0x58, (byte)0x75, (byte)0x2b, (byte)0x42,
                (byte)0x4f, (byte)0x0a, (byte)0x30, (byte)0x42, (byte)0x56, (byte)0x43, (byte)0x2f, (byte)0x47, (byte)0x50, (byte)0x57, (byte)0x52, (byte)0x33, (byte)0x61, (byte)0x38, (byte)0x78, (byte)0x6f,
                (byte)0x4d, (byte)0x65, (byte)0x66, (byte)0x46, (byte)0x4b, (byte)0x4b, (byte)0x45, (byte)0x31, (byte)0x78, (byte)0x39, (byte)0x31, (byte)0x6c, (byte)0x64, (byte)0x39, (byte)0x46, (byte)0x78,
                (byte)0x58, (byte)0x73, (byte)0x62, (byte)0x52, (byte)0x34, (byte)0x31, (byte)0x4f, (byte)0x63, (byte)0x31, (byte)0x74, (byte)0x49, (byte)0x59, (byte)0x64, (byte)0x7a, (byte)0x39, (byte)0x2f,
                (byte)0x5a, (byte)0x67, (byte)0x34, (byte)0x46, (byte)0x2f, (byte)0x5a, (byte)0x4a, (byte)0x37, (byte)0x47, (byte)0x70, (byte)0x2b, (byte)0x2f, (byte)0x39, (byte)0x41, (byte)0x75, (byte)0x6a,
                (byte)0x65, (byte)0x63, (byte)0x0a, (byte)0x4f, (byte)0x63, (byte)0x6b, (byte)0x42, (byte)0x48, (byte)0x61, (byte)0x44, (byte)0x4b, (byte)0x57, (byte)0x6d, (byte)0x6f, (byte)0x39, (byte)0x67,
                (byte)0x38, (byte)0x4f, (byte)0x56, (byte)0x4d, (byte)0x35, (byte)0x43, (byte)0x5a, (byte)0x65, (byte)0x4e, (byte)0x51, (byte)0x44, (byte)0x58, (byte)0x6c, (byte)0x43, (byte)0x4e, (byte)0x57,
                (byte)0x73, (byte)0x61, (byte)0x72, (byte)0x55, (byte)0x2f, (byte)0x64, (byte)0x6f, (byte)0x4d, (byte)0x47, (byte)0x74, (byte)0x76, (byte)0x41, (byte)0x6f, (byte)0x47, (byte)0x41, (byte)0x43,
                (byte)0x73, (byte)0x31, (byte)0x6e, (byte)0x6d, (byte)0x6c, (byte)0x33, (byte)0x61, (byte)0x39, (byte)0x4a, (byte)0x79, (byte)0x4d, (byte)0x59, (byte)0x7a, (byte)0x47, (byte)0x44, (byte)0x4a,
                (byte)0x4f, (byte)0x32, (byte)0x4f, (byte)0x0a, (byte)0x31, (byte)0x51, (byte)0x7a, (byte)0x78, (byte)0x30, (byte)0x53, (byte)0x47, (byte)0x37, (byte)0x6c, (byte)0x30, (byte)0x32, (byte)0x68,
                (byte)0x45, (byte)0x53, (byte)0x66, (byte)0x33, (byte)0x77, (byte)0x77, (byte)0x59, (byte)0x32, (byte)0x4e, (byte)0x4d, (byte)0x6b, (byte)0x63, (byte)0x4a, (byte)0x70, (byte)0x52, (byte)0x59,
                (byte)0x32, (byte)0x49, (byte)0x33, (byte)0x2f, (byte)0x6b, (byte)0x76, (byte)0x4d, (byte)0x47, (byte)0x75, (byte)0x4c, (byte)0x56, (byte)0x76, (byte)0x31, (byte)0x55, (byte)0x71, (byte)0x4c,
                (byte)0x4c, (byte)0x4a, (byte)0x5a, (byte)0x38, (byte)0x4b, (byte)0x59, (byte)0x4f, (byte)0x32, (byte)0x38, (byte)0x78, (byte)0x56, (byte)0x64, (byte)0x42, (byte)0x65, (byte)0x70, (byte)0x61,
                (byte)0x53, (byte)0x5a, (byte)0x4a, (byte)0x4a, (byte)0x0a, (byte)0x30, (byte)0x42, (byte)0x53, (byte)0x6c, (byte)0x6f, (byte)0x79, (byte)0x43, (byte)0x37, (byte)0x4b, (byte)0x54, (byte)0x7a,
                (byte)0x45, (byte)0x37, (byte)0x4a, (byte)0x6b, (byte)0x69, (byte)0x55, (byte)0x2b, (byte)0x59, (byte)0x6d, (byte)0x65, (byte)0x50, (byte)0x75, (byte)0x63, (byte)0x51, (byte)0x48, (byte)0x53,
                (byte)0x61, (byte)0x33, (byte)0x66, (byte)0x47, (byte)0x72, (byte)0x6a, (byte)0x6d, (byte)0x33, (byte)0x53, (byte)0x4b, (byte)0x61, (byte)0x4d, (byte)0x4c, (byte)0x44, (byte)0x2b, (byte)0x6e,
                (byte)0x55, (byte)0x74, (byte)0x68, (byte)0x63, (byte)0x37, (byte)0x61, (byte)0x47, (byte)0x4a, (byte)0x75, (byte)0x41, (byte)0x57, (byte)0x77, (byte)0x68, (byte)0x66, (byte)0x54, (byte)0x62,
                (byte)0x53, (byte)0x43, (byte)0x56, (byte)0x6d, (byte)0x2b, (byte)0x0a, (byte)0x41, (byte)0x62, (byte)0x48, (byte)0x5a, (byte)0x32, (byte)0x4b, (byte)0x69, (byte)0x41, (byte)0x48, (byte)0x73,
                (byte)0x45, (byte)0x72, (byte)0x56, (byte)0x49, (byte)0x37, (byte)0x66, (byte)0x55, (byte)0x6c, (byte)0x72, (byte)0x37, (byte)0x56, (byte)0x56, (byte)0x63, (byte)0x3d, (byte)0x0a, (byte)0x2d,
                (byte)0x2d, (byte)0x2d, (byte)0x2d, (byte)0x2d, (byte)0x45, (byte)0x4e, (byte)0x44, (byte)0x20, (byte)0x50, (byte)0x52, (byte)0x49, (byte)0x56, (byte)0x41, (byte)0x54, (byte)0x45, (byte)0x20,
                (byte)0x4b, (byte)0x45, (byte)0x59, (byte)0x2d, (byte)0x2d, (byte)0x2d, (byte)0x2d, (byte)0x2d, (byte)0x0a
        };

        public TLS_Packet.TLS_Packet_Builder tlsServerHelloTemplateBuilder = new TLS_Packet.TLS_Packet_Builder(new TLS_Packet(server_hello_template_bytes));
}
