package edu.wpi.appjudicator.ssl.tls_packet_parsing;

import java.util.LinkedList;
import java.util.List;

import edu.wpi.appjudicator.ssl.SSL_Helper;

public class TLS_Handshake_Certificate extends TLS_Handshake_Sublayer {
    byte handshake_type;
    byte[] len;
    byte[] certificates_length;
    List<TLS_Handshake_Certificate_Subclass> certs;

    @Override
    public int getTotalSize() {
        return 1 + len.length + SSL_Helper.convert_to_int(len, 0, len.length);
    }


    TLS_Handshake_Certificate(TLS_Handshake_Certificate_Builder b) {

        this.handshake_type = b.handshake_type;
        this.len = new byte[b.len.length];
        this.certificates_length = new byte[b.certificates_length.length];

        System.arraycopy(b.len, 0, this.len, 0, this.len.length);
        System.arraycopy(b.certificates_length, 0, this.certificates_length, 0, this.certificates_length.length);

        try {
            this.certs = new LinkedList<>(b.certs.build());
        } catch (Exception e) {

        }
    }

    TLS_Handshake_Certificate(byte[] raw_data, int start) {
        handshake_type = raw_data[start];
        len = new byte[3];
        certificates_length = new byte[3];
        System.arraycopy(raw_data, start + 1, len, 0, 3);
        System.arraycopy(raw_data, start + 4, certificates_length, 0, 3);

        int certs_bytes = SSL_Helper.convert_to_int(certificates_length, 0, 3);

        if (certs_bytes > 0) {
            certs = new LinkedList<TLS_Handshake_Certificate_Subclass>();
            int certs_bytes_constructed =0;
            while (certs_bytes_constructed < certs_bytes) {
                TLS_Handshake_Certificate_Subclass newCert = new TLS_Handshake_Certificate_Subclass(raw_data, start + 7 + certs_bytes_constructed);
                certs.add(newCert);
                certs_bytes_constructed += newCert.getTotalSize();
            }
        }
    }


    @Override
    protected byte[] buildRawData() {
        byte[] rd = new byte[getTotalSize()];
        rd[0] = handshake_type;
        System.arraycopy(len, 0, rd, 1, 3);
        System.arraycopy(certificates_length, 0, rd, 4, 3);
        int idx = 7;
        for (TLS_Handshake_Certificate_Subclass c : certs) {
            byte[] s_rd = c.getRawData();
            System.arraycopy(s_rd, 0, rd, idx, s_rd.length);
            idx += s_rd.length;
        }

        return rd;
    }

    public static class TLS_Handshake_Certificate_Builder implements TLS_Builder<TLS_Handshake_Certificate> {

        byte handshake_type;
        byte[] len;
        byte[] certificates_length;
        TLS_List_Builder<TLS_Handshake_Certificate_Subclass> certs;

        TLS_Handshake_Certificate_Builder(TLS_Handshake_Certificate o) {
            this.handshake_type = o.handshake_type;
            this.len = new byte[o.len.length];
            this.certificates_length = new byte[o.certificates_length.length];

            System.arraycopy(o.len, 0, this.len, 0, this.len.length);
            System.arraycopy(o.certificates_length, 0, this.certificates_length, 0, this.certificates_length.length);

            this.certs = new TLS_List_Builder<TLS_Handshake_Certificate_Subclass>();

            for (TLS_Handshake_Certificate_Subclass s : o.certs) {
                this.certs.add(new TLS_Handshake_Certificate_Subclass.TLS_Handshake_Certificate_Subclass_Builder(s));
            }
        }

        public TLS_Handshake_Certificate_Builder(List<byte[]> raw_certs) {
            this.handshake_type = TLS_Enum_Handshake_Type.SUBLAYER_HANDSHAKE_TYPE_CERTIFICATE.value();
            this.len = new byte[3];
            this.certificates_length = new byte[3];

            int certificates_len_int = 0;
            this.certs = new TLS_List_Builder<TLS_Handshake_Certificate_Subclass>();
            for (byte[] c : raw_certs) {
                certificates_len_int += c.length;
                this.certs.add(new TLS_Handshake_Certificate_Subclass.TLS_Handshake_Certificate_Subclass_Builder(c));
            }
            SSL_Helper.revert_to_bytes(certificates_len_int, certificates_length, 0, 3);
            SSL_Helper.revert_to_bytes(certificates_len_int + 3, len, 0, 3);
        }

        @Override
        public TLS_Handshake_Certificate build() throws Exception {
            return new TLS_Handshake_Certificate(this);
        }

        @Override
        public void fixSizeFields() {
            int totalSize = certs.fixLengthsAndGetTotalSize();
            if (this.certificates_length == null) {
                this.certificates_length = new byte[3];
            }
            SSL_Helper.revert_to_bytes(totalSize, this.certificates_length, 0, this.certificates_length.length);

            if (this.len == null) {
                this.len = new byte[3];
            }
            SSL_Helper.revert_to_bytes(totalSize + this.certificates_length.length, this.len, 0, this.len.length);

        }

        @Override
        public int fixAndGetTotalSize() {
            fixSizeFields();
            return SSL_Helper.convert_to_int(this.len, 0, this.len.length) + this.len.length + 1;

        }

    }

}
