package edu.wpi.appjudicator.ssl.tls_packet_parsing;

import android.util.Log;

import org.pcap4j.util.LazyValue;

import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;

import edu.wpi.appjudicator.ssl.SSL_Helper;

public class TLS_Handshake_Certificate_Subclass {
    protected final LazyValue<byte[]> rawDataCache;

    byte[] length;
    byte[] data;    //for our purposes this should be enough

    TLS_Handshake_Certificate_Subclass(TLS_Handshake_Certificate_Subclass_Builder b) {
        rawDataCache = new LazyValue<byte[]>(
                new LazyValue.BuildValueCommand<byte[]>() {
                    @Override
                    public byte[] buildValue() {
                        return buildRawData();
                    }
                });;

        this.length = new byte[b.length.length];
        this.data = new byte[b.data.length];

        System.arraycopy(b.length, 0, this.length, 0, this.length.length);
        System.arraycopy(b.data, 0, this.data, 0, this.data.length);
    }

    TLS_Handshake_Certificate_Subclass(byte[] rawData, int start) {
        length = new byte[3];
        System.arraycopy(rawData, start, length, 0, 3);
        int len = SSL_Helper.convert_to_int(length, 0, 3);
        if (len > 0) {
            data = new byte[len];
            System.arraycopy(rawData, start + 3, data, 0,len);
        }

        rawDataCache = new LazyValue<byte[]>(
                new LazyValue.BuildValueCommand<byte[]>() {
                    @Override
                    public byte[] buildValue() {
                        return buildRawData();
                    }
                });;
    }

    TLS_Handshake_Certificate_Subclass(String certificate_path) {
        Path path = Paths.get(certificate_path);
        try {
            byte[] data = Files.readAllBytes(path);
            System.arraycopy(data, 0, length, 0, 3);
            System.arraycopy(data, 3, data, 0, SSL_Helper.convert_to_int(length, 0, 3));
        } catch (Exception e) {
            Log.e("TLS Handshake Certificate Open File", "Exception while trying to read filepath! " + e.getMessage());
        }

        rawDataCache = new LazyValue<byte[]>(
                new LazyValue.BuildValueCommand<byte[]>() {
                    @Override
                    public byte[] buildValue() {
                        return buildRawData();
                    }
                });;
    }

    private byte[] buildRawData() {
        byte[] rd = new byte[getTotalSize()];
        System.arraycopy(length, 0, rd, 0, 3);
        if (SSL_Helper.convert_to_int(length, 0, 3) > 0) {
            System.arraycopy(data, 0, rd, 3, data.length);
        }
        return  rd;
    }

    public byte[] getRawData() {

        byte[] rawData = rawDataCache.getValue();

        byte[] copy = new byte[rawData.length];
        System.arraycopy(rawData, 0, copy, 0, copy.length);
        return copy;
    }

    public int getTotalSize() {
        return length.length + data.length;
    }

    public static class TLS_Handshake_Certificate_Subclass_Builder implements TLS_Builder<TLS_Handshake_Certificate_Subclass> {
        byte[] length;
        byte[] data;    //for our purposes this should be enough

        TLS_Handshake_Certificate_Subclass_Builder(byte[] raw_cert) {
            length = new byte[3];
            data = Arrays.copyOf(raw_cert,  raw_cert.length);
            SSL_Helper.revert_to_bytes(raw_cert.length, length,0, 3);
        }
        TLS_Handshake_Certificate_Subclass_Builder(TLS_Handshake_Certificate_Subclass o) {
            this.length = new byte[o.length.length];
            this.data = new byte[o.data.length];

            System.arraycopy(o.length, 0, this.length, 0, this.length.length);
            System.arraycopy(o.data, 0, this.data, 0, this.data.length);
        }

        @Override
        public TLS_Handshake_Certificate_Subclass build() throws Exception {
            return new TLS_Handshake_Certificate_Subclass(this);
        }

        @Override
        public void fixSizeFields() {
            if (this.length == null) {
                this.length = new byte[3];
            }
            SSL_Helper.revert_to_bytes(this.data.length, length,0, 3);
        }

        @Override
        public int fixAndGetTotalSize() {
            fixSizeFields();
            return SSL_Helper.convert_to_int(this.length, 0, this.length.length) + this.length.length;
        }

    }
}
