package edu.wpi.appjudicator.ssl.tls_packet_parsing;

import org.pcap4j.util.LazyValue;

import java.io.Serializable;
import java.util.LinkedList;
import java.util.List;

import edu.wpi.appjudicator.ssl.SSL_Helper;

public class TLS_Handshake_Client_Hello extends TLS_Handshake_Sublayer {
    byte handshake_type;
    byte[] length;
    byte version_major;
    byte version_minor;
    public TLS_Handshake_Random_Component_Subclass random_component;
    TLS_Handshake_SessionID_Subclass session_ID;
    TLS_Cypher_Suites cypher_suites;
    TLS_Compression_Methods compression_methods;
    short extensions_len;
    List<TLS_Handshake_Extension_Abstract> extensions;

    @Override
    public int getTotalSize() {
        return 1 + length.length + 2 + random_component.s_getTotalSize() +
                session_ID.s_getTotalSize() + cypher_suites.s_getTotalSize() + compression_methods.s_getTotalSize() + 2 + SSL_Helper.castToInt(extensions_len);
    }


    public TLS_Handshake_Extension_SERVER_NAME getServerNameExtension() {
        TLS_Handshake_Extension_SERVER_NAME ext = null;
        for (TLS_Handshake_Extension_Abstract e : extensions) {
            if (e.getClass() == TLS_Handshake_Extension_SERVER_NAME.class) {
                ext = (TLS_Handshake_Extension_SERVER_NAME)e;
                break;
            }
        }
        return ext;
    }

    public class TLS_Cypher_Suites implements Serializable {
        private final LazyValue<byte[]> s_rawDataCache;

        short cypher_suites_len;
        short[] cypher_suites;

        TLS_Cypher_Suites(byte[] raw_data, int start) {
            cypher_suites_len = (short) (SSL_Helper.convert_to_int(raw_data, start, 2));
            if (cypher_suites_len > 0) {
                int num_suites = cypher_suites_len / 2;
                cypher_suites = new short[num_suites];
                for (int suite = 0; suite < num_suites; ++suite) {
                    cypher_suites[suite] = (short) (SSL_Helper.convert_to_int(raw_data, start + 2 + 2*suite, 2));
                }
            }

            s_rawDataCache = new LazyValue<byte[]>(
                    new LazyValue.BuildValueCommand<byte[]>() {
                        @Override
                        public byte[] buildValue() {
                            return s_buildRawData();
                        }
                    }
            );;
        }


        byte[] s_buildRawData() {
            byte[] rd = new byte[s_getTotalSize()];
            SSL_Helper.revert_to_bytes((int) cypher_suites_len, rd, 0, 2);

            int dstPos = 2;
            int total_num_suites = cypher_suites_len / 2;
            for (int suite = 0; suite < total_num_suites; ++suite) {
                SSL_Helper.revert_to_bytes((int) cypher_suites[suite], rd, dstPos, 2);
                dstPos += 2;
            }

            return  rd;
        }

        public byte[] getRawData() {
            byte[] rawData = s_rawDataCache.getValue();

            byte[] copy = new byte[rawData.length];
            System.arraycopy(rawData, 0, copy, 0, copy.length);
            return copy;
        }

        int s_getTotalSize() {
            return  2 + SSL_Helper.castToInt(cypher_suites_len);
        }
    }

    public static class TLS_Compression_Methods implements Serializable {
        byte compression_methods_len;
        byte[] compression_methods;

        TLS_Compression_Methods(byte[] raw_data, int start) {
            compression_methods_len = raw_data[start];
            if (compression_methods_len > 0) {
                compression_methods = new byte[compression_methods_len];
                for (int m = 0; m < compression_methods_len; ++m) {
                    compression_methods[m] = raw_data[start + 1 + m];
                }
            }
        }

        int s_getTotalSize() {
            return  1+ compression_methods_len;
        }
    }

    TLS_Handshake_Client_Hello(byte[] raw_data, int start) {
        handshake_type = raw_data[start];
        length = new byte[3];
        System.arraycopy(raw_data, start + 1, length, 0, 3);

        version_major = raw_data[start + 4];
        version_minor = raw_data[start + 5];

        random_component = new TLS_Handshake_Random_Component_Subclass(raw_data, start+6);

        int size_of_constant_sized_components = 38;

        session_ID = new TLS_Handshake_SessionID_Subclass(raw_data, start + size_of_constant_sized_components);

        cypher_suites = new TLS_Cypher_Suites(raw_data, start + size_of_constant_sized_components + session_ID.s_getTotalSize());

        compression_methods = new TLS_Compression_Methods(raw_data, start + size_of_constant_sized_components + session_ID.s_getTotalSize() + cypher_suites.s_getTotalSize());

        int extensions_start_idx = size_of_constant_sized_components + session_ID.s_getTotalSize() + cypher_suites.s_getTotalSize() + compression_methods.s_getTotalSize();
        extensions_len = (short ) SSL_Helper.convert_to_int(raw_data, start + extensions_start_idx, 2);
        int totalExtensionsBytesToConstruct = SSL_Helper.castToInt(extensions_len);
        if (totalExtensionsBytesToConstruct > 0) {
            extensions = new LinkedList<TLS_Handshake_Extension_Abstract>();
            int totalExtensionsBytesConstructed = 0;
            TLS_Handshake_Extension_Factory factory = new TLS_Handshake_Extension_Factory();
            while (totalExtensionsBytesConstructed < totalExtensionsBytesToConstruct) {
                TLS_Handshake_Extension_Abstract newExtension = factory.build(raw_data, start + extensions_start_idx + 2 + totalExtensionsBytesConstructed);
                extensions.add(newExtension);
                totalExtensionsBytesConstructed += newExtension.getTotalSize();
            }
        }
    }

    protected byte[] buildRawData() {
        byte[] rd = new byte[getTotalSize()];
        rd[0] = handshake_type;
        System.arraycopy(length, 0, rd, 1, 3);
        rd[4] = version_major;
        rd[5] = version_minor;

        System.arraycopy(random_component.GMT_time, 0, rd, 6, 4);
        System.arraycopy(random_component.random_bytes, 0, rd, 10, 28);

        int dst_pos = 38;
        rd[dst_pos] = session_ID.session_ID_len;
        dst_pos += 1;
        if (session_ID.session_ID_len > 0) {
            System.arraycopy(session_ID.sessionID, 0, rd, dst_pos, session_ID.session_ID_len);
            dst_pos += session_ID.session_ID_len;
        }

        byte[] suites_data = this.cypher_suites.s_rawDataCache.getValue();
        System.arraycopy(suites_data, 0, rd , dst_pos, this.cypher_suites.s_getTotalSize());
        dst_pos += this.cypher_suites.s_getTotalSize();
        rd[dst_pos] = this.compression_methods.compression_methods_len;
        dst_pos += 1;
        System.arraycopy(this.compression_methods.compression_methods, 0, rd , dst_pos, SSL_Helper.castToInt(this.compression_methods.compression_methods_len));
        dst_pos += SSL_Helper.castToInt(this.compression_methods.compression_methods_len);

        SSL_Helper.revert_to_bytes((int)extensions_len, rd, dst_pos, 2);
        dst_pos += 2;

        for (TLS_Handshake_Extension_Abstract e : extensions) {
            System.arraycopy(e.getRawData(), 0, rd, dst_pos, e.getTotalSize());
            dst_pos += e.getTotalSize();
        }

        return rd;
    }


}
