package edu.wpi.appjudicator.ssl.tls_packet_parsing;

import org.pcap4j.util.LazyValue;

import edu.wpi.appjudicator.ssl.SSL_Helper;

public class TLS_Handshake_Client_Key_Exchange extends TLS_Handshake_Sublayer {
    byte handshake_type;
    byte[] length;
    public ECDHE_Client_Params client_params;

    @Override
    public int getTotalSize() {
        return 1 + length.length + client_params.totalLength;
    }

    TLS_Handshake_Client_Key_Exchange(byte[] raw_data, int start) {
        handshake_type = raw_data[start];
        length = new byte[3];
        System.arraycopy(raw_data, start+1, length, 0, 3);
        client_params = new ECDHE_Client_Params(raw_data, start+4, SSL_Helper.convert_to_int(length, 0, 3));
    }

    public static class  ECDHE_Client_Params {
        int totalLength; //not in the actual structure

        byte pub_key_len;
        public byte[] pub_key;

        LazyValue<byte[]> rawDataCache;

        public ECDHE_Client_Params(byte[] raw_data, int start, int len) {
            totalLength = len;

            pub_key_len = raw_data[start];
            int pub_key_len_actual = SSL_Helper.castToInt(pub_key_len);
            if (pub_key_len_actual > 0) {
                pub_key = new byte[pub_key_len_actual];
                System.arraycopy(raw_data, start+1, pub_key, 0, pub_key_len_actual);
            }

            rawDataCache = new LazyValue<byte[]>(
                    new LazyValue.BuildValueCommand<byte[]>() {
                        @Override
                        public byte[] buildValue() {
                            return s_buildRawData();
                        }
                    });;

        }

        public byte[] s_buildRawData() {
            byte[] rd = new byte[totalLength];
            rd[0] = pub_key_len;
            int pub_key_len_actual = SSL_Helper.castToInt(pub_key_len);
            if (pub_key_len_actual > 0){
                System.arraycopy(pub_key, 0, rd, 1, pub_key_len_actual);
            }

            return rd;
        }

        public byte[] getRawData() {
            byte[] rawData = rawDataCache.getValue();

            byte[] copy = new byte[rawData.length];
            System.arraycopy(rawData, 0, copy, 0, copy.length);
            return copy;
        }

        public int getTotalSize() {
            return totalLength;
        }
    }


    @Override
    protected byte[] buildRawData() {
        byte[] rd = new byte[getTotalSize()];
        rd[0] = handshake_type;
        System.arraycopy(length, 0, rd, 1, 3);
        byte[] serverParamsRawData = client_params.getRawData();
        System.arraycopy(serverParamsRawData, 0, rd, 4, serverParamsRawData.length);
        return rd;
    }

}
