package edu.wpi.appjudicator.ssl.tls_packet_parsing;

import org.pcap4j.util.LazyValue;

import edu.wpi.appjudicator.ssl.SSL_Helper;

public abstract class TLS_Handshake_Extension_Abstract {
    protected final LazyValue<byte[]> rawDataCache;

    protected short type;
    protected short length;

    TLS_Handshake_Extension_Abstract(TLS_Handshake_Extension_Abstract_Builder b) {
        this.type = b.type;
        this.length = b.length;

        rawDataCache = new LazyValue<byte[]>(
                new LazyValue.BuildValueCommand<byte[]>() {
                    @Override
                    public byte[] buildValue() {
                        return buildRawData();
                    }
                });;
    }

    TLS_Handshake_Extension_Abstract(byte[] raw_data, int start) {
        type = (short) SSL_Helper.convert_to_int(raw_data, start, 2);
        length = (short) SSL_Helper.convert_to_int(raw_data, start + 2, 2);

        rawDataCache = new LazyValue<byte[]>(
                new LazyValue.BuildValueCommand<byte[]>() {
                    @Override
                    public byte[] buildValue() {
                        return buildRawData();
                    }
                });;
    }

    abstract byte[] buildRawData();

    public final byte[] getRawData() {
        byte[] rawData = rawDataCache.getValue();

        byte[] copy = new byte[rawData.length];
        System.arraycopy(rawData, 0, copy, 0, copy.length);
        return copy;
    }

    public final int getTotalSize() {
        return 4 + SSL_Helper.castToInt(length);
    }

    public static abstract class TLS_Handshake_Extension_Abstract_Builder implements TLS_Builder<TLS_Handshake_Extension_Abstract> {
        protected short type;
        protected short length;

        TLS_Handshake_Extension_Abstract_Builder(TLS_Handshake_Extension_Abstract o) {
            this.type = o.type;
            this.length = o.length;
        }

        TLS_Handshake_Extension_Abstract_Builder(byte[] rd, int start) {

            type = (short) SSL_Helper.convert_to_int(rd, start, 2);
            length = (short) SSL_Helper.convert_to_int(rd, start + 2, 2);
        }

        public abstract TLS_Handshake_Extension_Abstract build();

        public final int fixAndGetTotalSize() {
            fixSizeFields();
            return  SSL_Helper.castToInt(this.length) + 4;
        }
    }
}
