package edu.wpi.appjudicator.ssl.tls_packet_parsing;

import java.util.LinkedList;
import java.util.List;

import edu.wpi.appjudicator.ssl.SSL_Helper;

public class TLS_Handshake_Extension_SERVER_NAME extends TLS_Handshake_Extension_Abstract{
    short list_length;
    public List<Server_Name_Subclass> list;

    TLS_Handshake_Extension_SERVER_NAME(TLS_Handshake_Extension_SERVER_NAME_Builder b) {
        super(b);

        this.list_length = b.list_length;
        try{
            this.list = new LinkedList<Server_Name_Subclass>(b.list.build());
        } catch (Exception e) {}
    }

    TLS_Handshake_Extension_SERVER_NAME(byte[] raw_data, int start) {
        super(raw_data, start);

        list_length = (short) SSL_Helper.convert_to_int(raw_data, start + 4, 2);
        int totalBytesToConstruct = SSL_Helper.castToInt(list_length);

        if (totalBytesToConstruct > 0) {
            list = new LinkedList<Server_Name_Subclass>();
            int totalBytesConstructed = 0;
            while (totalBytesConstructed < totalBytesToConstruct) {
                Server_Name_Subclass newName = new Server_Name_Subclass(raw_data, start + 6 + totalBytesConstructed);
                list.add(newName);
                totalBytesConstructed += newName.s_getTotalSize();
            }
        }
    }

    public static class Server_Name_Subclass {
        public byte name_type;
        public short name_length;
        public byte[] name_bytes;

        Server_Name_Subclass(Server_Name_Subclass_Builder b) {
            this.name_type = b.name_type;
            this.name_length = b.name_length;
            this.name_bytes = new byte[b.name_bytes.length];
            System.arraycopy(b.name_bytes, 0, this.name_bytes, 0, this.name_bytes.length);
        }
        
        Server_Name_Subclass(byte[] raw_data, int start) {
            name_type = raw_data[start];
            name_length = (short) SSL_Helper.convert_to_int(raw_data, start + 1, 2);
            name_bytes = new byte[SSL_Helper.castToInt(name_length)];
            System.arraycopy(raw_data, start + 3, name_bytes, 0, name_bytes.length);
        }

        byte[] s_convertToBytes() {
            byte[] rd = new byte[s_getTotalSize()];
            rd[0] = name_type;
            SSL_Helper.revert_to_bytes(SSL_Helper.castToInt(name_length), rd, 1, 2);
            System.arraycopy(name_bytes, 0, rd, 3, SSL_Helper.castToInt(name_length));
            return rd;
        }

        int s_getTotalSize() {
            return 3 + SSL_Helper.castToInt(name_length);
        }
    }

    public static class Server_Name_Subclass_Builder implements TLS_Builder<Server_Name_Subclass> {
        byte name_type;
        short name_length;
        byte[] name_bytes;

        Server_Name_Subclass_Builder(Server_Name_Subclass o) {
            this.name_type = o.name_type;
            this.name_length = o.name_length;
            this.name_bytes = new byte[name_bytes.length];
            System.arraycopy(o.name_bytes, 0, this.name_bytes, 0, this.name_bytes.length);
        }

        @Override
        public Server_Name_Subclass build() throws Exception {
            return new Server_Name_Subclass(this);
        }

        @Override
        public void fixSizeFields() {
            this.name_length = (short) name_bytes.length;
        }

        @Override
        public int fixAndGetTotalSize() {
            return SSL_Helper.castToInt(this.name_length) + 3;
        }
    }

    @Override
    byte[] buildRawData() {
        byte[] rd = new byte[getTotalSize()];
        SSL_Helper.revert_to_bytes(SSL_Helper.castToInt(super.type), rd, 0, 2);
        SSL_Helper.revert_to_bytes(SSL_Helper.castToInt(super.length), rd, 2, 2);
        SSL_Helper.revert_to_bytes(SSL_Helper.castToInt(list_length), rd, 4, 2);
        int idx = 6;
        for (Server_Name_Subclass s : list) {
            byte[] cp = s.s_convertToBytes();
            System.arraycopy(cp, 0, rd, idx, s.s_getTotalSize());
            idx += s.s_getTotalSize();
        }

        return rd;
    }

    public static class TLS_Handshake_Extension_SERVER_NAME_Builder extends TLS_Handshake_Extension_Abstract_Builder{
        short list_length;
        TLS_List_Builder<Server_Name_Subclass> list;

        TLS_Handshake_Extension_SERVER_NAME_Builder(TLS_Handshake_Extension_SERVER_NAME o) {
            super(o);

            this.list_length = o.list_length;
            this.list = new TLS_List_Builder<Server_Name_Subclass>();
            for (TLS_Handshake_Extension_SERVER_NAME.Server_Name_Subclass s : o.list) {
                this.list.add(new Server_Name_Subclass_Builder(s));
            }
        }

        @Override
        public TLS_Handshake_Extension_SERVER_NAME build() {
            return new TLS_Handshake_Extension_SERVER_NAME(this);
        }

        @Override
        public void fixSizeFields() {
            int list_size = this.list.fixLengthsAndGetTotalSize();
            this.list_length = (short) list_size;
            super.length = (short) (list_size + 2);
        }

    }
}
