package edu.wpi.appjudicator.ssl.tls_packet_parsing;

import java.util.LinkedList;
import java.util.List;

import edu.wpi.appjudicator.ssl.SSL_Helper;

public class TLS_Handshake_Server_Hello extends TLS_Handshake_Sublayer {
    byte handshake_type;
    byte[] length;
    byte version_major;
    byte version_minor;
    TLS_Handshake_Random_Component_Subclass random_component;
    TLS_Handshake_SessionID_Subclass session_ID;
    short cypher_suite;
    byte compression_method;
    short extensions_len;
    List<TLS_Handshake_Extension_Abstract> extensions;

    public int getTotalSize() {
        return  1 + length.length + SSL_Helper.convert_to_int(length, 0, 3);
    }

    TLS_Handshake_Server_Hello(TLS_Handshake_Server_Hello_Builder b) {
        this.handshake_type=b.handshake_type;
        this.version_major=b.version_major;
        this.version_minor=b.version_minor;
        this.cypher_suite=b.cypher_suite;
        this.compression_method=b.compression_method;
        this.extensions_len=b.extensions_len;

        this.length=new byte[b.length.length];
        System.arraycopy(b.length,0,this.length,0,this.length.length);

        try {
            this.random_component = b.random_component_builder.build();
            this.session_ID = b.session_ID_builder.build();
            this.extensions = new LinkedList<TLS_Handshake_Extension_Abstract>(b.extensions.build());
        } catch (Exception e) {

        }
    }

    TLS_Handshake_Server_Hello(byte[] raw_data, int start) {
        handshake_type = raw_data[start];
        length = new byte[3];
        System.arraycopy(raw_data, start + 1, length, 0, 3);

        version_major = raw_data[start + 4];
        version_minor = raw_data[start + 5];

        random_component = new TLS_Handshake_Random_Component_Subclass(raw_data, start+6);

        int size_of_constant_sized_components = 38;

        session_ID = new TLS_Handshake_SessionID_Subclass(raw_data, start + size_of_constant_sized_components);

        cypher_suite = (short) SSL_Helper.convert_to_int(raw_data, start + size_of_constant_sized_components + session_ID.s_getTotalSize(), 2);

        compression_method = raw_data[start + size_of_constant_sized_components + session_ID.s_getTotalSize() + 2];

        int extensions_start_idx = size_of_constant_sized_components + session_ID.s_getTotalSize() + 3;
        extensions_len = (short )(raw_data[start + extensions_start_idx] * 256 + raw_data[start + extensions_start_idx + 1]);
        int totalExtensionsBytesToConstruct = SSL_Helper.castToInt(extensions_len);
        if (totalExtensionsBytesToConstruct > 0) {
            extensions = new LinkedList<TLS_Handshake_Extension_Abstract>();
            int totalExtensionsBytesConstructed = 0;
            TLS_Handshake_Extension_Factory factory = new TLS_Handshake_Extension_Factory();
            while (totalExtensionsBytesConstructed < totalExtensionsBytesToConstruct) {
                TLS_Handshake_Extension_Abstract newExtension = factory.build(raw_data, start + extensions_start_idx + 2 + totalExtensionsBytesConstructed);
                extensions.add(newExtension);
                totalExtensionsBytesConstructed += newExtension.getTotalSize();
            }
        }

    }

    protected byte[] buildRawData() {
        byte[] rd = new byte[getTotalSize()];
        rd[0] = handshake_type;
        System.arraycopy(length, 0, rd, 1, 3);
        rd[4] = version_major;
        rd[5] = version_minor;

        System.arraycopy(random_component.GMT_time, 0, rd, 6, 4);
        System.arraycopy(random_component.random_bytes, 0, rd, 10, 28);

        int dst_pos = 38;
        rd[dst_pos] = session_ID.session_ID_len;
        dst_pos += 1;
        if (SSL_Helper.castToInt(session_ID.session_ID_len) > 0) {
            System.arraycopy(session_ID.sessionID, 0, rd, dst_pos, session_ID.sessionID.length);
            dst_pos += SSL_Helper.castToInt(session_ID.session_ID_len);
        }

        SSL_Helper.revert_to_bytes(SSL_Helper.castToInt(cypher_suite), rd, dst_pos, 2);
        rd[dst_pos + 2] = compression_method;

        SSL_Helper.revert_to_bytes(SSL_Helper.castToInt(extensions_len), rd, dst_pos + 3, 2);

        dst_pos += 5;
        for (TLS_Handshake_Extension_Abstract e : extensions) {
            System.arraycopy(e.getRawData(), 0, rd, dst_pos, e.getTotalSize());
            dst_pos += e.getTotalSize();
        }

        return rd;
    }

    public static class TLS_Handshake_Server_Hello_Builder  implements TLS_Builder<TLS_Handshake_Server_Hello> {
        byte handshake_type;
        byte[] length;
        byte version_major;
        byte version_minor;
        public TLS_Handshake_Random_Component_Subclass.TLS_Handshake_Random_Component_Subclass_Builder random_component_builder;
        public TLS_Handshake_SessionID_Subclass.TLS_Handshake_SessionID_Subclass_Builder session_ID_builder;
        short cypher_suite;
        byte compression_method;
        short extensions_len;
        public TLS_List_Builder<TLS_Handshake_Extension_Abstract> extensions;

        TLS_Handshake_Server_Hello_Builder(TLS_Handshake_Server_Hello o){
            this.handshake_type=o.handshake_type;
            this.version_major=o.version_major;
            this.version_minor=o.version_minor;
            this.cypher_suite=o.cypher_suite;
            this.compression_method=o.compression_method;
            this.extensions_len=o.extensions_len;

            this.length=new byte[o.length.length];
            System.arraycopy(o.length,0,this.length,0,this.length.length);

            this.random_component_builder = new TLS_Handshake_Random_Component_Subclass.TLS_Handshake_Random_Component_Subclass_Builder(o.random_component);
            this.session_ID_builder = new TLS_Handshake_SessionID_Subclass.TLS_Handshake_SessionID_Subclass_Builder(o.session_ID);
            this.extensions = new TLS_List_Builder<TLS_Handshake_Extension_Abstract>();

            TLS_Handshake_Extension_Builder_Factory extension_builder_factory = new TLS_Handshake_Extension_Builder_Factory();
            for (TLS_Handshake_Extension_Abstract e : o.extensions) {
                TLS_Handshake_Extension_Abstract.TLS_Handshake_Extension_Abstract_Builder b = extension_builder_factory.build(e);
                this.extensions.add(b);
            }

        }

        public TLS_Handshake_Server_Hello build() {
            return new TLS_Handshake_Server_Hello(this);
        }

        @Override
        public void fixSizeFields() {
            int ext_len = this.extensions.fixLengthsAndGetTotalSize();
            extensions_len = (short) ext_len;
            int session_id_len = this.session_ID_builder.fixAndGetTotalSize();
            int random_len = this.random_component_builder.fixAndGetTotalSize();

            int totalLen = ext_len + session_id_len + random_len + 7;
            if (this.length == null) {
                this.length = new byte[3];
            }

            SSL_Helper.revert_to_bytes(totalLen, this.length, 0, this.length.length);
        }

        @Override
        public int fixAndGetTotalSize() {
            fixSizeFields();
            return SSL_Helper.convert_to_int(this.length, 0, this.length.length) + this.length.length + 1;
        }

    }

}
