package edu.wpi.appjudicator.ssl.tls_packet_parsing;

import org.pcap4j.util.LazyValue;

import edu.wpi.appjudicator.ssl.SSL_Helper;

public class TLS_Handshake_Server_Key_Exchange extends TLS_Handshake_Sublayer {
    byte handshake_type;
    byte[] length;
    ECDHE_Server_Params server_params;

    @Override
    public int getTotalSize() {
        return 1 + length.length + SSL_Helper.convert_to_int(length, 0, length.length);
    }

    TLS_Handshake_Server_Key_Exchange(TLS_Handshake_Server_Key_Exchange_Builder b) {
        this.handshake_type = b.handshake_type;
        this.length = new byte[b.length.length];

        System.arraycopy(b.length, 0, this.length, 0, this.length.length);
        try{
            this.server_params = b.server_params.build();
        } catch (Exception e) {

        }
    }

    TLS_Handshake_Server_Key_Exchange(byte[] raw_data, int start) {
        handshake_type = raw_data[start];
        length = new byte[3];
        System.arraycopy(raw_data, start+1, length, 0, 3);
        server_params = new ECDHE_Server_Params(raw_data, start+4, SSL_Helper.convert_to_int(length, 0, 3));
    }

    public static class  ECDHE_Server_Params {
        byte curve_type;
        short named_curve;
        byte pub_key_len;
        byte[] pub_key;
        short signature_alg;
        short signature_len;
        byte[] signature;

        LazyValue<byte[]> rawDataCache;

        ECDHE_Server_Params(ECDHE_Server_Params_Builder b) {
            this.curve_type = b.curve_type;
            this.named_curve = b.named_curve;
            this.pub_key_len = b.pub_key_len;
            this.signature_alg = b.signature_alg;
            this.signature_len = b.signature_len;

            this.pub_key = new byte[b.pub_key.length];
            this.signature = new byte[b.signature.length];

            System.arraycopy(b.pub_key, 0, this.pub_key, 0, this.pub_key.length);
            System.arraycopy(b.signature, 0, this.signature, 0, this.signature.length);

            rawDataCache = new LazyValue<byte[]>(
                    new LazyValue.BuildValueCommand<byte[]>() {
                        @Override
                        public byte[] buildValue() {
                            return s_buildRawData();
                        }
                    });;
        }

        public ECDHE_Server_Params(byte[] raw_data, int start, int len) {
            curve_type = raw_data[start];
            named_curve = (short) SSL_Helper.convert_to_int(raw_data, start+1, 2);
            pub_key_len = raw_data[start+3];
            int pub_key_len_actual = SSL_Helper.castToInt(pub_key_len);
            if (pub_key_len_actual > 0) {
                pub_key = new byte[pub_key_len_actual];
                System.arraycopy(raw_data, start+4, pub_key, 0, pub_key_len_actual);
            }
            signature_alg = (short) SSL_Helper.convert_to_int(raw_data, start+4 + pub_key_len_actual, 2);
            signature_len = (short) SSL_Helper.convert_to_int(raw_data, start+6 + pub_key_len_actual, 2);
            if (SSL_Helper.castToInt(signature_len) > 0) {
                signature = new byte[SSL_Helper.castToInt(signature_len)];
                System.arraycopy(raw_data, start + 8 + pub_key_len_actual, signature, 0, SSL_Helper.castToInt(signature_len));
            }

            rawDataCache = new LazyValue<byte[]>(
                    new LazyValue.BuildValueCommand<byte[]>() {
                        @Override
                        public byte[] buildValue() {
                            return s_buildRawData();
                        }
                    });;

        }

        public byte[] s_get_bytes_for_signature() {
            byte[] rd = new byte[4 + pub_key.length];
            rd[0] = this.curve_type;
            SSL_Helper.revert_to_bytes((int) this.named_curve, rd, 1, 2);
            rd[3] = this.pub_key_len;
            System.arraycopy(this.pub_key, 0, rd, 4, this.pub_key.length);
            return rd;
        }

        public byte[] s_buildRawData() {
            byte[] rd = new byte[getTotalSize()];
            rd[0] = curve_type;
            SSL_Helper.revert_to_bytes(SSL_Helper.castToInt(named_curve),  rd, 1, 2);
            rd[3] = pub_key_len;
            int pub_key_len_actual = SSL_Helper.castToInt(pub_key_len);
            if (pub_key_len_actual > 0){
                System.arraycopy(pub_key, 0, rd, 4, pub_key_len_actual);
            }
            SSL_Helper.revert_to_bytes(SSL_Helper.castToInt(signature_alg),  rd, 4 + pub_key_len_actual, 2);
            SSL_Helper.revert_to_bytes(SSL_Helper.castToInt(signature_len),  rd, 6 + pub_key_len_actual, 2);
            if (SSL_Helper.castToInt(signature_len) > 0) {
                System.arraycopy(signature, 0, rd, 8+pub_key_len_actual, signature.length);
            }

            return rd;
        }

        public byte[] getRawData() {
            byte[] rawData = rawDataCache.getValue();

            byte[] copy = new byte[rawData.length];
            System.arraycopy(rawData, 0, copy, 0, copy.length);
            return copy;
        }

        public int getTotalSize() {
            return  1 + 2 + 1 + SSL_Helper.castToInt(pub_key_len) + 2 + 2 +SSL_Helper.castToInt(signature_len);
        }
    }

    public static class ECDHE_Server_Params_Builder implements TLS_Builder<ECDHE_Server_Params> {
        int totalLength; //not in the actual structure

        byte curve_type;
        short named_curve;
        byte pub_key_len;
        public byte[] pub_key;
        short signature_alg;
        short signature_len;
        public byte[] signature;

        public ECDHE_Server_Params_Builder(ECDHE_Server_Params o) {
            this.curve_type = o.curve_type;
            this.named_curve = o.named_curve;
            this.pub_key_len = o.pub_key_len;
            this.signature_alg = o.signature_alg;
            this.signature_len = o.signature_len;

            this.pub_key = new byte[o.pub_key.length];
            this.signature = new byte[o.signature.length];

            System.arraycopy(o.pub_key, 0, this.pub_key, 0, this.pub_key.length);
            System.arraycopy(o.signature, 0, this.signature, 0, this.signature.length);

        }

        @Override
        public ECDHE_Server_Params build() throws Exception {
            return new ECDHE_Server_Params(this);
        }

        @Override
        public void fixSizeFields() {
            this.signature_len = (short) signature.length;
            this.pub_key_len = (byte) pub_key.length;

        }

        @Override
        public int fixAndGetTotalSize() {
            fixSizeFields();
            return  1 + 2 + 1 + SSL_Helper.castToInt(pub_key_len) + 2 + 2 +SSL_Helper.castToInt(signature_len);
        }
    }


    @Override
    protected byte[] buildRawData() {
        byte[] rd = new byte[getTotalSize()];
        rd[0] = handshake_type;
        System.arraycopy(length, 0, rd, 1, 3);
        byte[] serverParamsRawData = server_params.getRawData();
        System.arraycopy(serverParamsRawData, 0, rd, 4, serverParamsRawData.length);
        return rd;
    }


    public static class TLS_Handshake_Server_Key_Exchange_Builder implements TLS_Builder<TLS_Handshake_Server_Key_Exchange>{
        byte handshake_type;
        byte[] length;
        public ECDHE_Server_Params_Builder server_params;

        TLS_Handshake_Server_Key_Exchange_Builder(TLS_Handshake_Server_Key_Exchange o) {
            this.handshake_type = o.handshake_type;
            this.length = new byte[o.length.length];

            System.arraycopy(o.length, 0, this.length, 0, this.length.length);
            this.server_params = new ECDHE_Server_Params_Builder(o.server_params);
        }

        @Override
        public TLS_Handshake_Server_Key_Exchange build() throws Exception {
            return new TLS_Handshake_Server_Key_Exchange(this);
        }

        @Override
        public void fixSizeFields() {
            int server_params_len = this.server_params.fixAndGetTotalSize();
            if (this.length == null) {
                this.length = new byte[3];
            }
            SSL_Helper.revert_to_bytes(server_params_len, this.length, 0, this.length.length);
        }

        @Override
        public int fixAndGetTotalSize() {
            fixSizeFields();
            return SSL_Helper.convert_to_int(this.length, 0, this.length.length) + this.length.length + 1;
        }
    }
}
