package edu.wpi.appjudicator.ssl.tls_packet_parsing;

import org.pcap4j.util.LazyValue;

import java.io.Serializable;
import java.util.LinkedList;

public class TLS_Packet implements Serializable {
    int length;
    private final LazyValue<byte[]> rawDataCache;


    public LinkedList<TLS_Record_Layer> record_layers;


    protected TLS_Packet(TLS_Packet_Builder b) throws Exception{
        this.length = b.length;
        try {
            this.record_layers = new LinkedList<TLS_Record_Layer>(b.record_layer_list_builder.build());
        } catch (Exception e) {}


        rawDataCache = new LazyValue<byte[]>(
                new LazyValue.BuildValueCommand<byte[]>() {
                    @Override
                    public byte[] buildValue() {
                        return buildRawData();
                    }
                }
        );;
    }
    public TLS_Packet(byte[] raw_bytes) {
        this(raw_bytes, raw_bytes.length);
    }

    public TLS_Packet(byte[] raw_bytes, int len) {
        length = len;

        int start = 0;
        int end = length - 1;

        record_layers = new LinkedList<TLS_Record_Layer>();

        while(end-start > 0) {
            try {
                TLS_Record_Layer newLayer = new TLS_Record_Layer(raw_bytes, start, end);
                record_layers.add(newLayer);
                start += newLayer.getTotalSize();
            } catch (IllegalStateException e) {
                //TODO: Send a decode error or unexpected message error
                e.printStackTrace();
                break;
            }
        }

        rawDataCache = new LazyValue<byte[]>(
                new LazyValue.BuildValueCommand<byte[]>() {
                    @Override
                    public byte[] buildValue() {
                        return buildRawData();
                    }
                }
        );;
    }

    byte[] buildRawData() {
        byte[] rd = new byte[length];
        int dstPos = 0;

        for (TLS_Record_Layer layer : record_layers) {
            System.arraycopy(layer.getRawData(), 0, rd, dstPos, layer.getTotalSize());
            dstPos += layer.getTotalSize();
        }
        return rd;
    }

    public byte[] getRawData() {
        byte[] rawData = rawDataCache.getValue();

        byte[] copy = new byte[rawData.length];
        System.arraycopy(rawData, 0, copy, 0, copy.length);
        return copy;
    }

    public static class TLS_Packet_Builder implements TLS_Builder<TLS_Packet>{
        int length;
        public TLS_List_Builder<TLS_Record_Layer> record_layer_list_builder;

        public TLS_Packet_Builder() {
            this.length = 0;
            this.record_layer_list_builder = null;
        }
        public TLS_Packet_Builder(TLS_Packet p) {
            this.length = p.length;
            this.record_layer_list_builder = new TLS_List_Builder<TLS_Record_Layer>();
            for (TLS_Record_Layer r : p.record_layers) {
                this.record_layer_list_builder.add(new TLS_Record_Layer.TLS_Record_Layer_Builder(r));
            }
        }

        public TLS_Packet build() throws Exception {
            if (this.length == 0 || this.record_layer_list_builder == null) {
                throw new Exception("Lenght or Record_Layers not initialized");
            }
            return new TLS_Packet(this);
        }

        @Override
        public void fixSizeFields() {
           this.length = record_layer_list_builder.fixLengthsAndGetTotalSize();
        }

        @Override
        public int fixAndGetTotalSize() {
            fixSizeFields();
            return this.length;
        }

    }
}








